// This may look like C code, but it is really -*- C++ -*-
/* 
Copyright (C) 1988 Free Software Foundation
    written by Doug Lea (dl@rocky.oswego.edu)

This file is part of GNU CC.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY.  No author or distributor
accepts responsibility to anyone for the consequences of using it
or for whether it serves any particular purpose or works at all,
unless he says so in writing.  Refer to the GNU CC General Public
License for full details.

Everyone is granted permission to copy, modify and redistribute
GNU CC, but only under the conditions described in the
GNU CC General Public License.   A copy of this license is
supposed to have been given to you along with GNU CC so you
can know your rights and responsibilities.  It should be in a
file named COPYING.  Among other things, the copyright notice
and this notice must be preserved on all copies.  
*/


#ifndef _ArgStringpSLList_h
#ifdef __GNUG__
#pragma once
#pragma interface
#endif
#define _ArgStringpSLList_h 1

#include <Pix.h>
#include "ArgStringp.defs.h"

#ifndef _ArgStringpSLListNode_h
#define _ArgStringpSLListNode_h 1

struct ArgStringpSLListNode
{
  ArgStringpSLListNode*         tl;
  ArgStringp                    hd;
                         ArgStringpSLListNode();
                         ArgStringpSLListNode(ArgStringp  h, ArgStringpSLListNode* t = 0);
                         ~ArgStringpSLListNode();
};

inline ArgStringpSLListNode::ArgStringpSLListNode() {}

inline ArgStringpSLListNode::ArgStringpSLListNode(ArgStringp  h, ArgStringpSLListNode* t) :hd(h), tl(t) {}

inline ArgStringpSLListNode::~ArgStringpSLListNode() {}

typedef ArgStringpSLListNode* ArgStringpSLListNodePtr;

#endif


class ArgStringpSLList
{
protected:
  ArgStringpSLListNode*        last;

public:
                        ArgStringpSLList();
                        ArgStringpSLList(ArgStringpSLList& a);
                        ~ArgStringpSLList();

  ArgStringpSLList&            operator = (ArgStringpSLList& a);

  int                   empty();
  int                   length();

  void                  clear();

  Pix                   prepend(ArgStringp  item);
  Pix                   append(ArgStringp  item);

  void                  join(ArgStringpSLList&);

  Pix                   prepend(ArgStringpSLListNode*);
  Pix                   append(ArgStringpSLListNode*);

  ArgStringp&                  operator () (Pix p);
  Pix                   first();
  void                  next(Pix& p);
  int                   owns(Pix p);
  Pix                   ins_after(Pix p, ArgStringp  item);
  void                  del_after(Pix p);

  ArgStringp&                  front();
  ArgStringp&                  rear();
  ArgStringp                   remove_front();
  int                   remove_front(ArgStringp& x);
  void                  del_front();

  void                  error(const char* msg);
  int                   OK();
};


inline ArgStringpSLList::~ArgStringpSLList()
{
  clear();
}

inline ArgStringpSLList::ArgStringpSLList()
{
  last = 0;
}

inline int ArgStringpSLList::empty()
{
  return last == 0;
}


inline Pix ArgStringpSLList::first()
{
  return (last == 0)? 0 : Pix(last->tl);
}

inline void ArgStringpSLList::next(Pix& p)
{
  p = (p == 0 || p == last)? 0 : Pix(((ArgStringpSLListNode*)(p))->tl);
}

inline ArgStringp& ArgStringpSLList::operator () (Pix p)
{
  if (p == 0) error("null Pix");
  return ((ArgStringpSLListNode*)(p))->hd;
}

inline ArgStringp& ArgStringpSLList::front()
{
  if (last == 0) error("front: empty list");
  return last->tl->hd;
}

inline ArgStringp& ArgStringpSLList::rear()
{
  if (last == 0) error("rear: empty list");
  return last->hd;
}


#endif
