// -*- C++ -*-
//------------------
//--- Builtin.cc ---
//------------------
/* 
 * Synopsis: 
 * Implementation Notes:
 * ToDo:
 * 
 * $Id: Builtin.cc,v 1.1 92/03/29 09:16:22 rvillari Exp Locker: rvillari $
 * $Log:	Builtin.cc,v $
Revision 1.1  92/03/29  09:16:22  rvillari
Initial revision

 *
 *
*/

#include "Builtin.h"
#include "support.h"
#include "InternalEnv.h"
#include "Glob.h"
#include <math.h>
#include "RexxStream.h"
#include "stringsupport.h"
#include <sys/types.h>
#include <sys/time.h>
#include <time.h>

extern Glob glob;

// some support macros to help in writing builtin-functions

#define EnsureExists(number) {						      \
    if (ArgsOmitted(number)) {						      \
	SetError(EIncorrectCallToRoutine);				      \
	return;								      \
    }									      \
}

#define EnsureNotExists(number) {				      \
    if (!ArgsOmitted(number)) {						      \
	SetError(EIncorrectCallToRoutine);				      \
	return;								      \
    }									      \
}

#define EnsureNumber(number) {					      \
    EnsureExists(number);						      \
    if (!Args(number).IsANumber()) {					      \
	SetError(EIncorrectCallToRoutine);				      \
	return;								      \
    }									      \
}

#define EnsureWhole(number) {					      \
    EnsureExists(number);						      \
    if (!Args(number).IsAWhole()) {					      \
	SetError(EIncorrectCallToRoutine);				      \
	return;								      \
    }									      \
}

#define EnsurePositiveWhole(number) {					      \
    EnsureExists(number);						      \
    if (!Args(number).IsAWhole() || Args(number).Integer() < 1) {     \
	SetError(EIncorrectCallToRoutine);				      \
	return;								      \
    }									      \
}

#define EnsureNonNegativeWhole(number) {			      \
    EnsureExists(number);						      \
    if (!Args(number).IsAWhole() || Args(number).Integer() < 0) {     \
	SetError(EIncorrectCallToRoutine);				      \
	return;								      \
    }									      \
}

#define EnsureMaxArgs(number) { 				      \
    if (GetNumArguments() > number) {					      \
	SetError(EIncorrectCallToRoutine);				      \
	return;								      \
    }									      \
}

#define EnsureMinArgs(number) {						      \
    for (int i = 0; i < number; i++)					      \
	if (ArgsOmitted(i)) {						      \
	    SetError(EIncorrectCallToRoutine);				      \
	    return;							      \
	}								      \
}

#define EnsureExactArgs(number) {   \
    if (GetNumArguments() != number) {					      \
	SetError(EIncorrectCallToRoutine);				      \
	return;								      \
    }									      \
    EnsureMinArgs(number);  \
}

// assumes it exists
char GetOption(int number) {
    char* cp = Args(number).String();
    if (cp[0] >= 'a' && cp[0] <= 'z') return ((cp[0]-'a')+'A'); // convert to upper
    return cp[0];
}

// assumes it exists
char GetPad(int number) {
    char* cp = Args(number).String();
    return cp[0];
}

void ABBREV() {
    // (information,info[,length])
    EnsureMinArgs(2);
    EnsureMaxArgs(3);
    if (GetNumArguments() == 2) {
	String cp = common_prefix(String(Args(0).String()), String(Args(1).String()));
	if (cp.length() == Args(1).Length()) {
	    SetReturn(ArgString(1.0));
	} else {
	    SetReturn(ArgString(0.0));
	}
    } else {
	// three arguments
	EnsurePositiveWhole(2);
	int length = Args(2).Integer();
	String cp = common_prefix(String(Args(0).String()), String(Args(1).String()));
	if (cp.length() > length) {
	    SetReturn(ArgString(1.0));
	} else {
	    SetReturn(ArgString(0.0));
	}
    }

}

void ABS() {
    // (number)
    EnsureExactArgs(1); // make sure we were passed one argument
    EnsureNumber(0);
    RexxNumber number = Args(0).Number();
    if (number.sign() < 0) {
	number.negate();
	SetReturn(Form(number, GetDigits()));
    } else {
	SetReturn(Form(number, GetDigits()));
    }
}

void ADDRESS() {
    // ()
    EnsureExactArgs(0);
    SetReturn(GetCurrentAddress());
}

void ARG() {
    // ([n[,option]])
    ArgString result;
    EnsureMaxArgs(2);
    if (GetNumArguments() == 0) {
	InternalData* temp = glob.Environments.PopEnv();
	result = ArgString((int)GetNumArguments());
	glob.Environments.PushEnv(temp);
	SetReturn(result);
    } else if (GetNumArguments() == 1) {
/*
	if (!number.IsAWhole()) {
	    SetError(EIncorrectCallToRoutine);
	    return;
	}
*/
	EnsureWhole(0);

	int number = Args(0).Integer();
	if (number < 1) {
	    SetError(EIncorrectCallToRoutine);
	    return;
	}

	InternalData* temp = glob.Environments.PopEnv();
	if (ArgsOmitted(number - 1)) {
	    result = ArgString("");
	} else {
	    result = Args(number - 1);
	}

	glob.Environments.PushEnv(temp);
	SetReturn(result);
    } else if (GetNumArguments() == 2) {
	EnsureExists(0);
	EnsureWhole(0);

	int number = Args(0).Integer();

	if (number < 1) {
	    SetError(EIncorrectCallToRoutine);
	    return;
	}

	char option = GetOption(1);
	if (option != 'O' && option != 'E') {
	    SetError(EIncorrectCallToRoutine);
	    return;
	}
	InternalData* temp = glob.Environments.PopEnv();
	if (option == 'O') {
	    result = ArgString((int)ArgsOmitted(number - 1));
	} else {
	    result = ArgString((int)!ArgsOmitted(number - 1));
	}
	glob.Environments.PushEnv(temp);
	SetReturn(result);
    } else {
	// should set error here
	SetError(EIncorrectCallToRoutine);
	return;
    }
}

void BITAND() {
    // (string1[,[string2][,pad]])
    EnsureMinArgs(1);
    EnsureMaxArgs(3);

    ArgString string2("");
    char pad = '\0';
    
    ArgString string1 = Args(0);
    if (!ArgsOmitted(1)) string2 = Args(1);

    ArgString temp;
    // make sure string1 is the longer of the two
    if (string2.Length() > string1.Length()) {
	// swap
	temp = string1;
	string1 = string2;
	string2 = temp;
    }

    if (!ArgsOmitted(2)) {
	pad = GetPad(2);
	// pad shorter string to length of longer string
	string2 = padonright(string2, string1.Length(), pad);
    }

    char* cp1 = string1.String();
    char* cp2 = string2.String();
    ArgString result(string1.String(), string1.Length());
    char* cpr = result.String();
    for (int i = 0; i < string2.Length(); i++) {
	cpr[i] = cp1[i] & cp2[i];
    }
    SetReturn(result);
}

void BITOR() {
    // (string1[,[string2][,pad]])
    EnsureMinArgs(1);
    EnsureMaxArgs(3);

    ArgString string2("");
    char pad = '\0';
    
    ArgString string1 = Args(0);
    if (!ArgsOmitted(1)) string2 = Args(1);

    ArgString temp;
    // make sure string1 is the longer of the two
    if (string2.Length() > string1.Length()) {
	// swap
	temp = string1;
	string1 = string2;
	string2 = temp;
    }

    if (!ArgsOmitted(2)) {
	pad = GetPad(2);
	// pad shorter string to length of longer string
	string2 = padonright(string2, string1.Length(), pad);
    }

    char* cp1 = string1.String();
    char* cp2 = string2.String();
    ArgString result(string1.String(), string1.Length());
    char* cpr = result.String();
    for (int i = 0; i < string2.Length(); i++) {
	cpr[i] = cp1[i] | cp2[i];
    }
    SetReturn(result);
}

void BITXOR() {
    // (string1[,[string2][,pad]])
    EnsureMinArgs(1);
    EnsureMaxArgs(3);

    ArgString string2("");
    char pad = '\0';
    
    ArgString string1 = Args(0);
    if (!ArgsOmitted(1)) string2 = Args(1);

    ArgString temp;
    // make sure string1 is the longer of the two
    if (string2.Length() > string1.Length()) {
	// swap
	temp = string1;
	string1 = string2;
	string2 = temp;
    }

    if (!ArgsOmitted(2)) {
	pad = GetPad(2);
	// pad shorter string to length of longer string
	string2 = padonright(string2, string1.Length(), pad);
    }

    char* cp1 = string1.String();
    char* cp2 = string2.String();
    ArgString result(string1.String(), string1.Length());
    char* cpr = result.String();
    for (int i = 0; i < string2.Length(); i++) {
	cpr[i] = cp1[i] ^ cp2[i];
    }
    SetReturn(result);
}

void B2X() {
    EnsureExactArgs(1);
    
    if (!validbin(Args(0))) {
	SetError(EInvalidHexadecimalOrBinaryString);
	return;
    }

    // we have a valid binary
    String temp = Args(0).String();
    temp.gsub(RXwhite, "");
    if ((temp.length() % 4) != 0) {
	temp = replicate("0", 4 - (temp.length() % 4)) + temp; // add those extra 0's
    }

    ArgString result("", temp.length()/4);
    char* cp = result.String();
    int index = 0;
    for (int i = 0; i < temp.length()/4; i++) {
	cp[i] = bintohex(temp[index], temp[index+1], temp[index+2], temp[index+3]);
	index += 4;
    }
    SetReturn(result);
}

void CENTER() {
    // (string,length[,pad])
    EnsureMinArgs(2);
    EnsureMaxArgs(3);
    EnsurePositiveWhole(1);

    int desiredlength = Args(1).Integer();
    char pad = ' ';

    if (!ArgsOmitted(2)) {
	pad = GetPad(2);
    }
    int length = Args(0).Length();

    ArgString result("", desiredlength);
    char* cp = result.String();

    if (desiredlength > length) {
	// put pad characers
	int frontpad = (desiredlength - length) / 2;
	int rearpad = desiredlength - length - frontpad;
	for (int i = 0; i < frontpad; i++) {
	    *cp = pad;
	    cp++;
	}
	char* string = Args(0).String();
	for (i = 0; i < length; i++) {
	    *cp = string[i];
	    cp++;
	}
	for (i = 0; i < rearpad; i++) {
	    *cp = pad;
	    cp++;
	}
    } else {
	// strip off some characters
	int frontstrip = (length - desiredlength) / 2;
	char* string = Args(0).String();
	for (int i = 0; i < desiredlength; i++) {
	    cp[i] = string[frontstrip + i];
	}
    }
    SetReturn(result);
}

void CHARIN() {
    // ([name][,[start][,length]])
    // stream function
    int length = 1; // default
    int charnumber = -1; // default
    EnsureMaxArgs(3);

    if (!ArgsOmitted(1)) {
	EnsurePositiveWhole(1);
	charnumber = Args(1).Integer() - 1;
    }

    if (!ArgsOmitted(2)) {
	EnsurePositiveWhole(2);
	length = Args(2).Integer();
    }

    String filename = "stdin";
    if (!ArgsOmitted(0)) {
	filename = Args(0).String();
    }

    ArgString ret;
    if (length < 0 || charnumber < -1) {
	SetError(EIncorrectCallToRoutine);
	return;
    }
    // make sure we know about it
    if (!glob.Streams.contains(filename)) {
	glob.Streams[filename] = new RexxStream(filename);
    }
    ret = ArgString("");
    glob.Streams[filename]->Charin(ret, length, charnumber);
    // possibly an error
    if (glob.Streams[filename]->Error() != SNoError) {
	// should raise NOTREADY trap
	if (SignalOn(SigNotReady)) {
	    if (ActionIsSignal(SigNotReady)) {
		SetError(ENotReady);
		SetSignalDescription(SigNotReady, ArgString((char*)filename));
		return;
	    } else {
		SetSignalPending(SigNotReady);
		SetSignalDelayed(SigNotReady);
		SetSignalDescription(SigNotReady, ArgString((char*)filename));
	    }
	}
    }
    SetReturn(ret);
}

void CHAROUT() {
    // ([name][,[string][,start]])

    // stream function
    int charnumber = -1; // default
    ArgString string(""); // default
    EnsureMaxArgs(3);

    if (!ArgsOmitted(2)) {
	EnsurePositiveWhole(2);
	charnumber = Args(2).Integer() - 1;
    }

    if (!ArgsOmitted(1)) {
	string = Args(1);
    }

    String filename = "stdout";
    if (!ArgsOmitted(0)) {
	filename = Args(0).String();
    }

    if (charnumber < -1) {
	SetError(EIncorrectCallToRoutine);
	return;
    }
    // make sure we know about it
    if (!glob.Streams.contains(filename)) {
	glob.Streams[filename] = new RexxStream(filename);
    }

    int ret = glob.Streams[filename]->Charout(string, charnumber);
    // possibly an error
    if (glob.Streams[filename]->Error() != SNoError) {
	// should raise NOTREADY trap
	if (SignalOn(SigNotReady)) {
	    if (ActionIsSignal(SigNotReady)) {
		SetError(ENotReady);
		SetSignalDescription(SigNotReady, ArgString((char*)filename));
		return;
	    } else {
		SetSignalPending(SigNotReady);
		SetSignalDelayed(SigNotReady);
		SetSignalDescription(SigNotReady, ArgString((char*)filename));
	    }
	}
    }
    SetReturn(ArgString(ret));
}

void CHARS() {
    // ([name])
    EnsureMaxArgs(1);
    String filename = "stdin";
    if (!ArgsOmitted(0)) {
	filename = Args(0).String();
    }

    // make sure we know about it
    if (!glob.Streams.contains(filename)) {
	glob.Streams[filename] = new RexxStream(filename);
    }

    SetReturn(ArgString(glob.Streams[filename]->Chars()));
}

void COMPARE() {
    // (string1,string2[,pad])
    EnsureMinArgs(2);
    EnsureMaxArgs(3);
    
    char pad = ' ';

    ArgString string1;
    ArgString string2;
    if (Args(0).Length() > Args(1).Length()) {
	string1 = Args(0);
	string2 = Args(1);
    } else {
	string1 = Args(1);
	string2 = Args(0);
    }
    // string1 is longer of the two
    if (!ArgsOmitted(2)) {
	pad = GetPad(2);
    }
    string2 = padonright(string2, string1.Length(), pad);
    // string1 and string2 have the same length now
    char* cp1 = string1.String();
    char* cp2 = string2.String();
    for (int i = 0; i < string1.Length(); i++) {
	if (*cp1 != *cp2) {
	    SetReturn(ArgString(i + 1));
	    return;
	}
	cp1++;
	cp2++;
    }
    SetReturn(ArgString(0.0));
}

void CONDITION() {
    EnsureMaxArgs(1);

    char option = 'I';

    if (!ArgsOmitted(0)) {
	option = GetOption(0);
    }

    if (option != 'C' && option != 'D' && option != 'I' && option != 'S') {
	SetError(EIncorrectCallToRoutine);
	return;
    }

    if (GetSignalBeingProcessed() == SigSignal) {
	// this is considered as not processing a 'real' signal, so we return null
	SetReturn(ArgString(""));
	return;
    }

    switch (option) {
      case 'C' :
	SetReturn(GetSignalLabel(GetSignalBeingProcessed()));
	break;
      case 'D' :
	SetReturn(GetSignalDescription(GetSignalBeingProcessed()));
	break;
      case 'I' :
	if (ActionIsSignal(GetSignalBeingProcessed())) {
	    SetReturn(ArgString("SIGNAL"));
	} else {
	    SetReturn(ArgString("CALL"));
	}
	break;
      case 'S' :
	if (SignalOn(GetSignalBeingProcessed())) {
	    SetReturn(ArgString("ON"));
	} else if (SignalOff(GetSignalBeingProcessed())) {
	    SetReturn(ArgString("OFF"));
	} else {
	    SetReturn(ArgString("DELAYED"));
	}
	break;
    };
}

void COPIES() {
    EnsureMinArgs(2);
    EnsureNonNegativeWhole(1);
    ArgString result("", Args(1).Integer() * Args(0).Length());
    int n = result.Length();
    char* string = Args(0).String();
    int stringlength = Args(0).Length();
    char* cp = result.String();
    for (int i = 0; i < n; i += stringlength) {
	for (int j = 0; j < stringlength; j++) {
	    cp[i + j] = string[j];
	}
    }
    SetReturn(result);
}

void C2D() {
    EnsureMinArgs(1);
    EnsureMaxArgs(2);
    if (!ArgsOmitted(1)) {
	// do signed conversion
	EnsureNonNegativeWhole(1);
	int n = Args(1).Integer();
	SetReturn(c2d(Args(0), n));
    } else {
	// do unsigned conversion
	SetReturn(uc2d(Args(0)));
    }
}

void C2X() {
    EnsureExactArgs(1);
    ArgString result("", Args(0).Length() * 2);
    char* string = Args(0).String();
    char* cp = result.String();
    int index = 0;
    for (int i = 0; i < result.Length(); i += 2, index++) {
	cp[i] = chartohex(string[index] >> 4);
	cp[i+1] = chartohex(string[index]);
    }
    SetReturn(result);
}

void DATATYPE() {
    EnsureMinArgs(1);

    if (ArgsOmitted(1)) {
	if (Args(0).IsANumber()) SetReturn(ArgString("NUM"));
	else SetReturn(ArgString("CHAR"));
    } else {
	char option = GetOption(1);
	int result;
	switch (option) {
	  case 'A' :
	    result = isalphanumeric(Args(0));
	    break;
	  case 'B' :
	    result = validbin(Args(0));
	    break;
	  case 'L' :
	    result = islower(Args(0));
	    break;
	  case 'M' :
	    result = ismixed(Args(0));
	    break;
	  case 'N' :
	    result = Args(0).IsANumber();
	    break;
	  case 'S' :
	    result = issymbol(Args(0));
	    break;
	  case 'U' :
	    result = isupper(Args(0));
	    break;
	  case 'W' :
	    result = Args(0).IsAWhole();
	    break;
	  case 'X' :
	    result = validhex(Args(0));
	    break;
	  default:
	    SetError(EIncorrectCallToRoutine);
	    return;
	};
	SetReturn(ArgString(result));
    }
}

void DATE() {
    EnsureMaxArgs(1);
    char option = 'N';

    if (!ArgsOmitted(0)) {
	option = GetOption(0);
    }

    time_t clock;
    time(&clock);
    struct tm* currenttime;
    currenttime = localtime(&clock);
    char buf[128];

    switch (option) {
      case 'B' :
	// unimplemented
	strcpy(buf, "Unimplemented option for date()");
	break;
      case 'D' :
	strftime(buf, 128, "%j", currenttime);
	break;
      case 'E' :
	strftime(buf, 128, "%d/%m/%y", currenttime);
	break;
      case 'M' :
	strftime(buf, 128, "%B", currenttime);
	break;
      case 'N' :
	strftime(buf, 128, "%d %b %y", currenttime);
	break;
      case 'O' :
	strftime(buf, 128, "%y/%m/%d", currenttime);
	break;
      case 'S' :
	strftime(buf, 128, "%Y%m%d", currenttime);
	break;
      case 'U' :
	strftime(buf, 128, "%d/%m/%y", currenttime);
	break;
      case 'W' :
	strftime(buf, 128, "%A", currenttime);
	break;
      default:
	SetError(EIncorrectCallToRoutine);
	return;
    };
    SetReturn(ArgString(buf));
}

void DELSTR() {
    EnsureMinArgs(2);
    EnsureMaxArgs(3);
    
    EnsurePositiveWhole(1);
    
    int length = Args(0).Length();

    int n = Args(1).Integer() - 1;
    if (!ArgsOmitted(2)) {
	EnsureNonNegativeWhole(2);
	length = Args(2).Integer();
    }

    if (n > Args(0).Length()) {
	SetReturn(Args(0));  // remains unchanged
    } else {
	ArgString result("");
	if ((n + length) > Args(0).Length()) {
	    // remove to end of string
	    result += substring(Args(0), 0, n);
	} else {
	    result += substring(Args(0), 0, n);
	    result += substring(Args(0), n + length, Args(0).Length());
	}
	SetReturn(result);
    }
}

void DELWORD() {
    // (string,n[,length])
    EnsureMinArgs(2);
    EnsureMaxArgs(3);
    EnsurePositiveWhole(1);
    
    int n = Args(1).Integer();
    int length = Args(0).Length();  // guaranteed > maxwords in this string
    if (!ArgsOmitted(2)) {
	EnsureNonNegativeWhole(2);
	length = Args(2).Integer();
    }
    ArgString string = Args(0);
    ArgString result("");
    int index = 0;
    int wordnumber = 0;
    int readingspaces = 1;
    for (int i = 0; i < string.Length(); i++) {
	if (readingspaces && string[i] == ' ') {
	    // we're reading spaces and we got a space
	    // do nothing
	} else if (readingspaces && string[i] != ' ') {
	    // we're reading spaces and we got a non-space
	    wordnumber++;
	    readingspaces = 0;
	} else if (!readingspaces && string[i] != ' ') {
	    // we're reading words and we got a non-space
	    // readingspaces = 1;
	} else if (!readingspaces && string[i] == ' ') {
	    // we're reading words and we got a space
	    readingspaces = 1;
	}

	if (wordnumber >= n && wordnumber < n + length) {
	} else {
	    result += ArgString(string[i]);
	}
    }
    SetReturn(result);
}

void DIGITS() {
    EnsureExactArgs(0);
    SetReturn((int)GetDigits());
}

void D2C() {
    EnsureMinArgs(1);
    EnsureMaxArgs(2);
    
    EnsureWhole(0);
    if (!ArgsOmitted(1)) {
	EnsureNonNegativeWhole(1);
	int n = Args(1).Integer();
	ArgString result;
	if (Args(0).Integer() < 0) {
	    // number is negative
	    ArgString positive(-Args(0).Integer());
	    result = ud2c(positive);
	    SetReturn(padonleft(twocomplement(result), n, (char)0xff));
	} else {
	    // number is positive
	    result = ud2c(Args(0).Integer());
	    SetReturn(padonleft(result, n, '\0'));
	}
	// 
    } else {
	EnsureNonNegativeWhole(0);
	SetReturn(ud2c(Args(0)));
    }
}

void D2X() {
    EnsureMinArgs(1);
    EnsureMaxArgs(2);
    
    if (!ArgsOmitted(1)) {
	EnsureNonNegativeWhole(1);
	int n = Args(1).Integer();
	ArgString result;
	if (Args(0).Integer() < 0) {
	    // number is negative
	    ArgString positive(-Args(0).Integer());
	    result = ud2x(positive);
	    SetReturn(padonleft(xtwocomplement(result), n, 'F'));
	} else {
	    // number is positive
	    result = ud2x(Args(0));
	    SetReturn(padonleft(result, n, '0'));
	}
	// 
    } else {
	EnsureNonNegativeWhole(0);
	SetReturn(ud2x(Args(0)));
    }
}

void ERRORTEXT() {
    EnsureExactArgs(1);

    EnsureNonNegativeWhole(0);
    int n = Args(0).Integer();
    if (n < 0 || n > 99) {
	SetError(EIncorrectCallToRoutine);
	return;
    }

    SetReturn(ArgString(GetErrorText((Errors) n)));
}

void FORM() {
    EnsureExactArgs(0);
    if (GetForm()) {
	SetReturn("ENGINEERING");
    } else {
	SetReturn("SCIENTIFIC");
    }
}

void FORMAT() {
    EnsureMinArgs(1);
    int before = -1;
    int after = -1;
    int expp = -1;
    int expt = -1;
    
    if (!Args(0).IsANumber()) {
	SetError(EBadArithmeticConversion);
	return;
    }
    if (!ArgsOmitted(1)) {
	EnsureNonNegativeWhole(1);
	before = Args(1).Integer();
    }

    if (!ArgsOmitted(2)) {
	EnsureNonNegativeWhole(2);
	after = Args(2).Integer();
    }
    
    if (!ArgsOmitted(3)) {
	EnsureNonNegativeWhole(3);
	expp = Args(3).Integer();
    }

    if (!ArgsOmitted(4)) {
	EnsureNonNegativeWhole(4);
	expt = Args(4).Integer();
    } else {
	expt = GetDigits();
    }

    SetReturn(Format(Args(0).Number(), GetDigits(), 0, before, after, expp, expt));
}

void FUZZ() {
    EnsureExactArgs(0);
    SetReturn((int)GetFuzz());
}

void INSERT() {
    // (new,target[,[n][,[length][,pad]]])
    EnsureMinArgs(2);
    EnsureMaxArgs(5);

    char pad = ' ';
    int n = 0;
    int length = Args(0).Length(); // default is new length
    
    if (!ArgsOmitted(2)) {
	EnsureNonNegativeWhole(2);
	n = Args(2).Integer();
    }

    if (!ArgsOmitted(3)) {
	EnsureNonNegativeWhole(3);
	length = Args(3).Integer();
    }

    if (!ArgsOmitted(4)) {
	pad = GetPad(4);
    }

    ArgString newstring = padonright(Args(0), length, pad);
    ArgString target = Args(1);
    ArgString result;
    if (n == 0) {
	// insert before string
	result = newstring;
	result += target;
    } else if ( n > target.Length()) {
	// insert after string
	result = target;
	result += padonleft(newstring, newstring.Length() + n - target.Length(), pad);
    } else {
	//
	result = substring(target, 0, n);
	result += newstring;
	result += substring(target, n, target.Length());
    }
    SetReturn(result);
}

void LASTPOS() {
    // (needle,haystack[,start])
    EnsureMinArgs(2);
    EnsureMaxArgs(3);

    ArgString needle = Args(0);
    ArgString haystack = Args(1);

    int start = haystack.Length();//default
    
    if (!ArgsOmitted(2)) {
	EnsurePositiveWhole(2);
	start = Args(2).Integer();
	if (start > haystack.Length()) 
	    start = haystack.Length();
    }

    if (needle.Length() == 0) {
	SetReturn((int)0);
	return;
    }

    start = start - needle.Length(); //
    for (int i = start; i >= 0; i--) {
	if (substring(haystack, i, i + needle.Length()) ==
	    needle) {
	    SetReturn((int)i + 1);
	    return;
	}
    }
    SetReturn(0);
}

void LEFT() {
    // (string,length[,pad])
    EnsureMinArgs(2);
    EnsureMaxArgs(3);
    EnsureNonNegativeWhole(1);

    char pad = ' ';
    
    int length = Args(1).Integer();
    if (!ArgsOmitted(2)) {
	pad = GetPad(2);
    }
    SetReturn(padonright(Args(0), length, pad));
}

void LENGTH() {
    EnsureExactArgs(1);
    SetReturn(Args(0).Length());
}

void LINEIN() {
    // ([name][,[line][,count]])
    // stream function
    int count = 1; // default
    int linenumber = -1; // default
    EnsureMaxArgs(3);

    if (!ArgsOmitted(1)) {
	EnsurePositiveWhole(1);
	linenumber = Args(1).Integer() - 1;
    }

    if (!ArgsOmitted(2)) {
	EnsurePositiveWhole(2);
	count = Args(2).Integer();
    }

    String filename = "stdin";
    if (!ArgsOmitted(0)) {
	filename = Args(0).String();
    }

    ArgString ret;
    if (count < 0 || count > 1 || linenumber < -1) {
	SetError(EIncorrectCallToRoutine);
	return;
    }
    // make sure we know about it
    if (!glob.Streams.contains(filename)) {
	glob.Streams[filename] = new RexxStream(filename);
    }
    ret = ArgString("");
    glob.Streams[filename]->Linein(ret, count, linenumber);
    // possibly an error
    if (glob.Streams[filename]->Error() != SNoError) {
	// should raise NOTREADY trap
	if (SignalOn(SigNotReady)) {
	    if (ActionIsSignal(SigNotReady)) {
		SetError(ENotReady);
		SetSignalDescription(SigNotReady, ArgString((char*)filename));
		return;
	    } else {
		SetSignalPending(SigNotReady);
		SetSignalDelayed(SigNotReady);
		SetSignalDescription(SigNotReady, ArgString((char*)filename));
	    }
	}
    }
    SetReturn(ret);
}

void LINEOUT() {
    // ([name][,[string][,line]])

    // stream function
    int linenumber = -1; // default
    ArgString string(""); // default
    EnsureMaxArgs(3);

    if (!ArgsOmitted(2)) {
	EnsurePositiveWhole(2);
	linenumber = Args(2).Integer() - 1;
    }

    if (!ArgsOmitted(1)) {
	string = Args(1);
    }

    String filename = "stdout";
    if (!ArgsOmitted(0)) {
	filename = Args(0).String();
    }

    if (linenumber < -1) {
	SetError(EIncorrectCallToRoutine);
	return;
    }
    // make sure we know about it
    if (!glob.Streams.contains(filename)) {
	glob.Streams[filename] = new RexxStream(filename);
    }

    int ret = glob.Streams[filename]->Lineout(string, linenumber);
    // possibly an error
    if (glob.Streams[filename]->Error() != SNoError) {
	// should raise NOTREADY trap
	if (SignalOn(SigNotReady)) {
	    if (ActionIsSignal(SigNotReady)) {
		SetError(ENotReady);
		SetSignalDescription(SigNotReady, ArgString((char*)filename));
		return;
	    } else {
		SetSignalPending(SigNotReady);
		SetSignalDelayed(SigNotReady);
		SetSignalDescription(SigNotReady, ArgString((char*)filename));
	    }
	}
    }
    SetReturn(ArgString(ret));
}

void LINES() {
    // ([name])
    EnsureMaxArgs(1);
    String filename = "stdin";
    if (!ArgsOmitted(0)) {
	filename = Args(0).String();
    }

    // make sure we know about it
    if (!glob.Streams.contains(filename)) {
	glob.Streams[filename] = new RexxStream(filename);
    }

    SetReturn(ArgString(glob.Streams[filename]->Lines()));
}

void MAX() {
    EnsureMinArgs(1);
    int num = GetNumArguments();
    EnsureMinArgs(num);

    RexxNumber max;

    if (!Args(0).IsANumber()) {
	SetError(EBadArithmeticConversion);
	return;
    }

    
    max = Args(0).Number();
    
    for (int i = 1; i < num; i++) {
	if (!Args(i).IsANumber()) {
	    SetError(EBadArithmeticConversion);
	    return;
	}
	RexxNumber n1 = max;
	RexxNumber n2 = Args(i).Number();
	n2.negate();
	RexxNumber comp;
	add(n1, n2, GetDigits() - GetFuzz(), comp);
	if (comp.sign() < 0) max = Args(i).Number();
    }
    SetReturn(Form(max, GetDigits()));
}

void MIN() {
    EnsureMinArgs(1);
    int num = GetNumArguments();
    EnsureMinArgs(num);

    RexxNumber min;

    if (!Args(0).IsANumber()) {
	SetError(EBadArithmeticConversion);
	return;
    }

    min = Args(0).Number();
    
    for (int i = 1; i < num; i++) {
	if (!Args(i).IsANumber()) {
	    SetError(EBadArithmeticConversion);
	    return;
	}
	RexxNumber n1 = min;
	RexxNumber n2 = Args(i).Number();
	n2.negate();
	RexxNumber comp;
	add(n1, n2, GetDigits() - GetFuzz(), comp);
	if (comp.sign() > 0) min = Args(i).Number();
    }
    SetReturn(Form(min, GetDigits()));
}

void OVERLAY() {
    // (new,target[,[n][,[length][,pad]]])
    EnsureMinArgs(2);
    EnsureMaxArgs(5);

    char pad = ' ';
    int n = 0;
    int length = Args(0).Length(); // default is new length
    
    if (!ArgsOmitted(2)) {
	EnsurePositiveWhole(2);
	n = Args(2).Integer() - 1;
    }

    if (!ArgsOmitted(3)) {
	EnsureNonNegativeWhole(3);
	length = Args(3).Integer();
    }

    if (!ArgsOmitted(4)) {
	pad = GetPad(4);
    }

    ArgString newstring = padonright(Args(0), length, pad);
    ArgString target = Args(1);
    ArgString result;
    if ( n > target.Length()) {
	// overlay after string
	result = target;
	result += padonleft(newstring, newstring.Length() + n - target.Length(), pad);
    } else {
	// overlay before string
	result = substring(target, 0, n); // get before part
	result += newstring; // add newstring
	result += substring(target, n + newstring.Length(), target.Length()); // get after newstring
    }
    SetReturn(result);
}

void POS() {
    // (needle,haystack[,start])
    EnsureMinArgs(2);
    EnsureMaxArgs(3);

    ArgString needle = Args(0);
    ArgString haystack = Args(1);

    int start = 0;
    
    if (!ArgsOmitted(2)) {
	EnsurePositiveWhole(2);
	start = Args(2).Integer() - 1;
    }

    if (needle.Length() == 0) {
	SetReturn((int)0);
	return;
    }

    int end = haystack.Length() - needle.Length() + 1;
    for (int i = start; i < end; i++) {
	if (substring(haystack, i, i + needle.Length()) ==
	    needle) {
	    SetReturn((int)i + 1);
	    return;
	}
    }
    SetReturn(0);
}

void QUEUED() {
    EnsureExactArgs(0);
    SetReturn((int)glob.GlobalQueue.Enqueued());
}

void RANDOM() {
    EnsureMaxArgs(3);
    int maximum = 999;
    int minimum = 0;
    int seed;
    if (GetNumArguments() == 1) {
	// (max)
	EnsureNonNegativeWhole(1);
	maximum = Args(0).Integer();
	// normalized is between zero and one
    } else {
	// ([min][,[max][,seed]])
	if (!ArgsOmitted(0)) {
	    // specified minimum
	    EnsureWhole(0);
	    minimum = Args(0).Integer();
	}
	if (!ArgsOmitted(1)) {
	    // specified maximum
	    EnsureWhole(0);
	    maximum = Args(1).Integer();
	    if (maximum < minimum) {
		SetError(EIncorrectCallToRoutine);
		return;
	    }
	}
	if (!ArgsOmitted(2)) {
	    // specified seed
	    EnsureWhole(1);
	    seed = Args(1).Integer();
	    srandom(seed);
	}
    }
    if (maximum - minimum > 10000) {
	SetError(EIncorrectCallToRoutine);
	return;
    }
    long rvalue = random();
    double normalized = ((double)rvalue / (double)MAXLONG);
    int result = (int)(normalized * (maximum - minimum + 1) + minimum);
    SetReturn((int)result);
}

void REVERSE() {
    EnsureExactArgs(1);
    SetReturn(reverse(Args(0)));
}

void RIGHT() {
    // (string,length[,pad])
    EnsureMinArgs(2);
    EnsureMaxArgs(3);
    EnsureNonNegativeWhole(1);

    char pad = ' ';
    
    int length = Args(1).Integer();
    if (!ArgsOmitted(2)) {
	pad = GetPad(2);
    }
    SetReturn(padonleft(Args(0), length, pad));
}

void SIGN() {
    EnsureExactArgs(1);
    if (!Args(0).IsANumber()) {
	SetError(EBadArithmeticConversion);
	return;
    }

    SetReturn(Args(0).Number().sign());
}

void SOURCELINE() {
    EnsureMaxArgs(1);
    if (ArgsOmitted(0)) {
	SetReturn(ArgString(GetSourcecodeLines()));
    } else {
	EnsurePositiveWhole(0);
	int linenumber = Args(0).Integer() - 1;
	if (linenumber < 0) {
	    SetError(EIncorrectCallToRoutine);
	    return;
	}
	SetReturn(ArgString(GetSourcecode(linenumber)));
    }
}

void SPACE() {
    // (string[,[n][,pad]])
    EnsureMinArgs(1);
    EnsureMaxArgs(3);
    int n = 1;
    char pad = ' ';
    if (!ArgsOmitted(1)) {
	EnsureNonNegativeWhole(1);
	n = Args(1).Integer();
    }
    if (!ArgsOmitted(2)) {
	pad = GetPad(2);
    }

    ArgString string = stripfront(striprear(Args(0)));
    ArgString result("");
    ArgString padstring("", n);
    for (int i = 0; i < n; i++) padstring[i] = pad;

    int readingspaces = 0;
    for (i = 0; i < string.Length(); i++) {
	if (readingspaces && string[i] == ' ') {
	    // we're reading spaces and we got a space
	    // do nothing
	} else if (readingspaces && string[i] != ' ') {
	    // we're reading spaces and we got a non-space
	    readingspaces = 0;
	    result += padstring;
	} else if (!readingspaces && string[i] != ' ') {
	    // we're reading words and we got a non-space
	} else if (!readingspaces && string[i] == ' ') {
	    // we're reading words and we got a space
	    readingspaces = 1;
	}

	if (!readingspaces) {
	    result += ArgString(string[i]);
	}
    }
    SetReturn(result);
}

void STREAM() {
    // (name[,operation[,streamcommand]])
    // stream commands supported -- 'W'=open write, 'R'=open read,
    // 'C'=close, 'E'=exists, 'B'=both read/write, 'N'=no error(clear error)

    EnsureMinArgs(1);
    EnsureMaxArgs(3);
    String name = Args(0).String(); // get name of the stream

    // make sure we know about it
    if (!glob.Streams.contains(name)) glob.Streams[name] = new RexxStream(name);

    char option = 'S'; // our default
    char command;
    if (!ArgsOmitted(1)) option = GetOption(1); // get operation
    if (option != 'C' && option != 'D' && option != 'S') {
	// invalid operation
	SetError(EIncorrectCallToRoutine);
	return;
    }
    int ret; 
   switch (option) {
      case 'C' :
	EnsureExists(2);
	command = GetOption(2);
	if (command != 'W' && command != 'R' && command != 'C' && command != 'E' && command != 'B') {
	    SetError(EIncorrectCallToRoutine);
	    return;
	}
	switch (command) {
	  case 'B' :
	    ret = glob.Streams[name]->ReadWriteAccess();
	    SetReturn(ArgString(ret));
	    break;
	  case 'N' :
	    glob.Streams[name]->ClearError();
	    SetReturn(ArgString(1)); // always succeeds
	    break;
	  case 'W' :
	    ret = glob.Streams[name]->WriteAccess();
	    SetReturn(ArgString(ret));
	    break;
	  case 'R' :
	    ret = glob.Streams[name]->ReadAccess();
	    SetReturn(ArgString(ret));
	    break;
	  case 'C' :
	    ret = glob.Streams[name]->Close();
	    SetReturn(""); // close doesn't return anything
	    break;
	  case 'E' :
	    ret = glob.Streams[name]->Exists();
	    SetReturn(ArgString(ret)); // return 1 or 0
	    break;
	};
	// do command here
	break;
      case 'D' :
	EnsureNotExists(2);
	// do description here
	SetReturn(glob.Streams[name]->Description());
	break;
      case 'S' : 
	EnsureNotExists(2);
	// do state here
	SetReturn(glob.Streams[name]->State());
	break;
    };
}

void STRIP() {
    // (string[,[option][,char]])
    EnsureMinArgs(1);
    EnsureMaxArgs(3);

    char option = 'B';
    char removethis = ' ';

    if (!ArgsOmitted(1)) {
	option = GetOption(1);
    }
    if (!ArgsOmitted(2)) {
	removethis = GetPad(2);
    }

    switch (option) {
      case 'B' :
	SetReturn(stripfront(striprear(Args(0), removethis), removethis));
	break;
      case 'L' :
	SetReturn(stripfront(Args(0),removethis));
	break;
      case 'T' :
	SetReturn(striprear(Args(0),removethis));
	break;
      default:
	SetError(EIncorrectCallToRoutine);
	return;
    };
}

void SUBSTR() {
    // (string,n[,[length][,pad]])
    EnsureMinArgs(2);
    EnsureMaxArgs(4);
    EnsurePositiveWhole(1);

    char pad = ' ';
    int length = Args(0).Length();
    int n = Args(1).Integer();
    if (!ArgsOmitted(3)) {
	pad = GetPad(3);
    }

    if (!ArgsOmitted(2)) {
	EnsureNonNegativeWhole(2);
	length = Args(2).Integer();
	ArgString string = substring(Args(0), n-1, n + length);
	SetReturn(padonright(string, length, pad));
    } else {
	ArgString string = substring(Args(0), n-1, n + length);
	SetReturn(string);
    }
}

void SUBWORD() {
    // (string,n[,length])
    EnsureMinArgs(2);
    EnsureMaxArgs(3);
    EnsurePositiveWhole(1);
    
    int n = Args(1).Integer();
    int length = Args(0).Length();  // guaranteed > maxwords in this string
    if (!ArgsOmitted(2)) {
	EnsureNonNegativeWhole(2);
	length = Args(2).Integer();
    }
    ArgString string = Args(0);
    ArgString result("");
    int index = 0;
    int wordnumber = 0;
    int readingspaces = 1;
    for (int i = 0; i < string.Length(); i++) {
	if (readingspaces && string[i] == ' ') {
	    // we're reading spaces and we got a space
	    // do nothing
	} else if (readingspaces && string[i] != ' ') {
	    // we're reading spaces and we got a non-space
	    wordnumber++;
	    readingspaces = 0;
	} else if (!readingspaces && string[i] != ' ') {
	    // we're reading words and we got a non-space
	    // readingspaces = 1;
	} else if (!readingspaces && string[i] == ' ') {
	    // we're reading words and we got a space
	    readingspaces = 1;
	}

	if (wordnumber >= n && wordnumber < n + length) {
	    result += ArgString(string[i]);
	} else {
	}
    }
    SetReturn(stripfront(striprear(result)));
}

void SYMBOL() {
    EnsureExactArgs(1);
    ArgString symbol = Args(0);

    if (!issymbol(symbol)) {
	SetReturn("BAD");
    } else {
	if (symbol[0] == '.' || (symbol[0] >= '0' && symbol[0] <= '9')) {
	    // could be a literal
	    SetReturn("LIT");
	} else {
	    // or a symbol
	    String ssymbol = symbol;
	    ssymbol.upcase();
	    ArgString result = glob.SymbolTables.ReadAny(ssymbol);
	    if (glob.UninitializedAccess) {
		SetReturn("LIT");
	    } else {
		SetReturn("VAR");
	    }
	}
    }
}

void TIME() {
    EnsureMaxArgs(1);
    char option = 'N';

    if (!ArgsOmitted(0)) {
	option = GetOption(0);
    }

    time_t clock;
    time(&clock);
    struct tm* currenttime;
    currenttime = localtime(&clock);
    char buf[128];

    switch (option) {
      case 'C' :
	strftime(buf, 128, "%I:%M%p", currenttime);
	break;
      case 'E' :
	long sec, micro;
	GetElapsedTime(sec, micro);
	if (sec == 0) {
	    InternalData* temp = glob.Environments.PopEnv();
	    SetElapsedTime((long)clock, 0);
	    glob.Environments.PushEnv(temp);
	    sprintf(buf, "0.000000");
	} else {
	    // micro not implemented
	    sprintf(buf, "%ld.000000", (long)clock - sec);
	}
	break;
      case 'H' :
	strftime(buf, 128, "%H", currenttime);
	break;
      case 'L' :
	strftime(buf, 128, "%H:%M:%S.000000", currenttime);
	break;
      case 'M' :
	strcpy(buf, "Unimplemented option for time()");
//	strftime(buf, 128, "%M", currenttime);
	break;
      case 'N' :
	strftime(buf, 128, "%H:%M:%S", currenttime);
	break;
      case 'R' :
	GetElapsedTime(sec, micro);
	if (sec == 0) {
	    InternalData* temp = glob.Environments.PopEnv();
	    SetElapsedTime((long)clock, 0);
	    glob.Environments.PushEnv(temp);
	    sprintf(buf, "0.000000");
	} else {
	    InternalData* temp = glob.Environments.PopEnv();
	    SetElapsedTime((long)clock, 0);
	    glob.Environments.PushEnv(temp);
	    // micro not implemented
	    sprintf(buf, "%ld.000000", (long)clock - sec);
	}
	break;
      case 'S' :
	strcpy(buf, "Unimplemented option for time()");
//	strftime(buf, 128, "%S", currenttime);
	break;
      default:
	SetError(EIncorrectCallToRoutine);
	return;
    };

    SetReturn(lower(ArgString(buf)));
}

void TRACE() {
    EnsureMaxArgs(1);

    SetReturn("O");
}

void TRANSLATE() {
    // (string[,[tableo][,[tablei][,pad]]])

    EnsureMinArgs(1);
    EnsureMaxArgs(4);

    ArgString tableo;
    ArgString tablei;
    int tableilength;
    char pad = ' ';

    if (ArgsOmitted(1) && ArgsOmitted(2)) {
	// conver to upper case
	SetReturn(upper(Args(0)));
	return;
    }

    if (!ArgsOmitted(3)) {
	pad = GetPad(3);
    }

    if (!ArgsOmitted(2)) {
	tablei = Args(2);
    } else {
	tablei = xrange(0x00, 0xff);
    }
    tableilength = tablei.Length();

    if (!ArgsOmitted(1)) {
	tableo = padonright(Args(1), tableilength, pad);
    } else {
	tableo = padonright("", tableilength, pad);
    }

    ArgString string = Args(0);
    int length = string.Length();
    for (int i = 0; i < length; i++) {
	for (int j = 0; j < tableilength; j++) {
	    if (string[i] == tablei[j]) {
		string[i] = tableo[j];
		break;
	    }
	}
    }
    SetReturn(string);
}

void TRUNC() {
    // (number[,n])
    EnsureMinArgs(1);
    EnsureMaxArgs(2);
    int n = 0;
    if (!Args(0).IsANumber()) {
	SetError(EBadArithmeticConversion);
	return;
    }
    if (!ArgsOmitted(1)) {
	EnsureNonNegativeWhole(1);
	n = Args(1).Integer();
    }

    SetReturn(Truncate(Args(0).Number(), GetDigits(), n));
}

void VALUE() {
    // (name[,[newvalue]][,selector])

    EnsureMinArgs(1);
    EnsureMaxArgs(3);
    
    ArgString environment = "SYMBOLTABLE";
    ArgString result;
    ArgString symbol = Args(0);
    
    if (!ArgsOmitted(2)) {
	environment = Args(2);
    }
    
    if (environment == "SYMBOLTABLE") {
	if (!issymbol(symbol)) {
	    SetError(ESymbolExpected);
	} else {
	    String ssymbol = symbol;
	    ssymbol.upcase();
	    if (symbol[0] == '.' || (symbol[0] >= '0' && symbol[0] <= '9')) {
		// could be a literal
		SetReturn(ArgString((char*)ssymbol));
	    } else {
		// or a symbol
		result = glob.SymbolTables.ReadAny(ssymbol);
		SetReturn(result);
		if (!ArgsOmitted(1)) {
		    // we will write out the value
		    glob.SymbolTables.WriteAny(ssymbol, Args(1));
		}
	    }
	}
    } else if (environment == "ENVIRONMENT") {
	// should return environment variable here
	char* env = getenv(symbol);
	if (env) {
	    result = env;
	} else {
	    result = "";
	}
	SetReturn(result);
	if (!ArgsOmitted(1)) {
	    // write out the new value
	    ArgString avalue = symbol;
	    avalue += "=";
	    avalue += Args(1);
	    char* value = new char[strlen(avalue) + 1];
	    strcpy(value, avalue);
	    putenv(value);
	}
    } else {
	SetError(EUnknownAddress);
    }
}

void VERIFY() {
    // (string,reference[,[option][,start]])
    EnsureMinArgs(2);
    EnsureMaxArgs(4);
    char option = 'N';
    int start = 0;

    if (!ArgsOmitted(2)) {
	option = GetOption(2);
    }
    
    if (!ArgsOmitted(3)) {
	EnsurePositiveWhole(3);
	start = Args(3).Integer() - 1;
    }

    if (option != 'N' && option != 'M') {
	SetError(EIncorrectCallToRoutine);
	return;
    }
    
    int result = 0;
    ArgString& string = Args(0);
    ArgString& reference = Args(1);
    
    if (string.Length() == 0 || start > string.Length()) {
	SetReturn((int)0);
	return;
    }

    for (int i = start; i < string.Length(); i++) {
	int found = 0;
	for (int j = 0; j < reference.Length(); j++) {
	    if (string[i] == reference[j]) {
		found = 1;
		break;
	    }
	}
	if (option == 'M') {
	    if (found) {
		SetReturn((int)i + 1);
		return;
	    }
	} else {
	    if (!found) {
		SetReturn((int)i + 1);
		return;
	    }
	}
    }
    SetReturn((int)0);
}

void WORD() {
    // (string,n)
    EnsureExactArgs(2);
    EnsurePositiveWhole(1);
    
    int n = Args(1).Integer();
    int length = 1;
    ArgString string = Args(0);
    ArgString result("");
    int index = 0;
    int wordnumber = 0;
    int readingspaces = 1;
    for (int i = 0; i < string.Length(); i++) {
	if (readingspaces && string[i] == ' ') {
	    // we're reading spaces and we got a space
	    // do nothing
	} else if (readingspaces && string[i] != ' ') {
	    // we're reading spaces and we got a non-space
	    wordnumber++;
	    readingspaces = 0;
	} else if (!readingspaces && string[i] != ' ') {
	    // we're reading words and we got a non-space
	    // readingspaces = 1;
	} else if (!readingspaces && string[i] == ' ') {
	    // we're reading words and we got a space
	    readingspaces = 1;
	}

	if (wordnumber >= n && wordnumber < n + length) {
	    result += ArgString(string[i]);
	} else {
	}
    }
    SetReturn(stripfront(striprear(result)));
}

void WORDINDEX() {
    // (string,n)
    EnsureExactArgs(2);
    EnsurePositiveWhole(1);
    
    int n = Args(1).Integer();
    int length = 1;
    ArgString string = Args(0);
    ArgString result("");
    int index = 0;
    int wordnumber = 0;
    int readingspaces = 1;
    int wordindex = 0;
    for (int i = 0; i < string.Length(); i++) {
	if (readingspaces && string[i] == ' ') {
	    // we're reading spaces and we got a space
	    // do nothing
	} else if (readingspaces && string[i] != ' ') {
	    // we're reading spaces and we got a non-space
	    wordnumber++;
	    readingspaces = 0;
	} else if (!readingspaces && string[i] != ' ') {
	    // we're reading words and we got a non-space
	    // readingspaces = 1;
	} else if (!readingspaces && string[i] == ' ') {
	    // we're reading words and we got a space
	    readingspaces = 1;
	}

	if (wordnumber >= n && wordnumber < n + length) {
	    wordindex = i + 1;
	    break;
	}
    }
    SetReturn((int)wordindex);
}

void WORDLENGTH() {
    // (string,n)
    EnsureExactArgs(2);
    EnsurePositiveWhole(1);
    
    int n = Args(1).Integer();
    int length = 1;
    ArgString string = Args(0);
    ArgString result("");
    int index = 0;
    int wordnumber = 0;
    int readingspaces = 1;
    for (int i = 0; i < string.Length(); i++) {
	if (readingspaces && string[i] == ' ') {
	    // we're reading spaces and we got a space
	    // do nothing
	} else if (readingspaces && string[i] != ' ') {
	    // we're reading spaces and we got a non-space
	    wordnumber++;
	    readingspaces = 0;
	} else if (!readingspaces && string[i] != ' ') {
	    // we're reading words and we got a non-space
	    // readingspaces = 1;
	} else if (!readingspaces && string[i] == ' ') {
	    // we're reading words and we got a space
	    readingspaces = 1;
	}

	if (wordnumber >= n && wordnumber < n + length) {
	    result += ArgString(string[i]);
	} else {
	}
    }
    SetReturn((int)stripfront(striprear(result)).Length());
}

void WORDPOS() {
    // (phrase,string[,start])
    EnsureMinArgs(2);
    EnsureMaxArgs(3);
    int start = 1;

    if (!ArgsOmitted(2)) {
	EnsurePositiveWhole(2);
	start = Args(2).Integer();
    }

    ArgString phrase = space(Args(0));
    ArgString string = space(Args(1));
    int wordnumber = 0;
    int readingspaces = 1;
    for (int i = 0; i < string.Length(); i++) {
	if (readingspaces && string[i] == ' ') {
	    // we're reading spaces and we got a space
	    // do nothing
	} else if (readingspaces && string[i] != ' ') {
	    // we're reading spaces and we got a non-space
	    wordnumber++;
	    readingspaces = 0;
	    if (wordnumber >= start) {
		if (substring(string, i, i + phrase.Length()) ==
		    phrase) {
		    SetReturn((int)wordnumber);
		    return;
		}
	    }
	} else if (!readingspaces && string[i] != ' ') {
	    // we're reading words and we got a non-space
	} else if (!readingspaces && string[i] == ' ') {
	    // we're reading words and we got a space
	    readingspaces = 1;
	}
    }
    SetReturn((int)0); // not found
}

void WORDS() {
    // (string)
    EnsureExactArgs(1);
    
    ArgString string = Args(0);
    int wordnumber = 0;
    int readingspaces = 1;
    for (int i = 0; i < string.Length(); i++) {
	if (readingspaces && string[i] == ' ') {
	    // we're reading spaces and we got a space
	    // do nothing
	} else if (readingspaces && string[i] != ' ') {
	    // we're reading spaces and we got a non-space
	    wordnumber++;
	    readingspaces = 0;
	} else if (!readingspaces && string[i] != ' ') {
	    // we're reading words and we got a non-space
	    // readingspaces = 1;
	} else if (!readingspaces && string[i] == ' ') {
	    // we're reading words and we got a space
	    readingspaces = 1;
	}
    }
    SetReturn((int)wordnumber);
}

void XRANGE() {
    // ([start][,end])
    char start = 0x00;
    char end = 0xff;
    if (!ArgsOmitted(0)) start = GetPad(0);
    if (!ArgsOmitted(1)) end = GetPad(1);
    SetReturn(xrange(start, end));
}

void X2B() {
    EnsureExactArgs(1);
    if (!validhex(Args(0))) {
	SetError(EInvalidHexadecimalOrBinaryString);
	return;
    }

    ArgString result("");
    ArgString string = upper(Args(0));
    for (int i = 0; i < Args(0).Length(); i++) {
	switch (string[i]) {
	  case '0' :
	    result += "0000";
	    break;
	  case '1' :
	    result += "0001";
	    break;
	  case '2' :
	    result += "0010";
	    break;
	  case '3' :
	    result += "0011";
	    break;
	  case '4' :
	    result += "0100";
	    break;
	  case '5' :
	    result += "0101";
	    break;
	  case '6' :
	    result += "0110";
	    break;
	  case '7' :
	    result += "0111";
	    break;
	  case '8' :
	    result += "1000";
	    break;
	  case '9' :
	    result += "1001";
	    break;
	  case 'A' :
	    result += "1010";
	    break;
	  case 'B' :
	    result += "1011";
	    break;
	  case 'C' :
	    result += "1100";
	    break;
	  case 'D' :
	    result += "1101";
	    break;
	  case 'E' :
	    result += "1110";
	    break;
	  case 'F' :
	    result += "1111";
	    break;
	};
    }
    SetReturn(result);
}

void X2C() {
    EnsureExactArgs(1);
    if (!validhex(Args(0))) {
	SetError(EInvalidHexadecimalOrBinaryString);
	return;
    }

    SetReturn(packhex(Args(0)));
}

void X2D() {
    EnsureMinArgs(1);
    EnsureMaxArgs(2);
    if (!validhex(Args(0))) {
	SetError(EInvalidHexadecimalOrBinaryString);
	return;
    }

    ArgString result;
    if (!ArgsOmitted(1)) {
	EnsureNonNegativeWhole(1);
	int n = Args(1).Integer();
	// n was specified -- possibly negative
	ArgString choppedhex = padonleft(Args(0), n, '0');
	ArgString packedhex = packhex(choppedhex);
	result = c2d(packedhex, packedhex.Length());
    } else {
	// unsigned integer should be returned
	ArgString packedhex = packhex(Args(0));
	result = uc2d(packedhex);
    }

    SetReturn(result);
}

Builtin::Builtin() : functions(0) {
}

Builtin::~Builtin() {
}

void Builtin::Preload() {
    functions["ABBREV"] = ABBREV;
    functions["ABS"] = ABS;
    functions["ADDRESS"] = ADDRESS;
    functions["ARG"] = ARG;
    functions["BITAND"] = BITAND;
    functions["BITOR"] = BITOR;
    functions["BITXOR"] = BITXOR;
    functions["B2X"] = B2X;
    functions["CENTER"] = CENTER;
    functions["CENTRE"] = CENTER;
    functions["CHARIN"] = CHARIN;
    functions["CHAROUT"] = CHAROUT;
    functions["CHARS"] = CHARS;
    functions["COMPARE"] = COMPARE;
    functions["CONDITION"] = CONDITION;
    functions["COPIES"] = COPIES;
    functions["C2D"] = C2D;
    functions["C2X"] = C2X;
    functions["DATATYPE"] = DATATYPE;
    functions["DATE"] = DATE;
    functions["DELSTR"] = DELSTR;
    functions["DELWORD"] = DELWORD;
    functions["DIGITS"] = DIGITS;
    functions["D2C"] = D2C;
    functions["D2X"] = D2X;
    functions["ERRORTEXT"] = ERRORTEXT;
    functions["FORM"] = FORM;
    functions["FORMAT"] = FORMAT;
    functions["FUZZ"] = FUZZ;
    functions["INSERT"] = INSERT;
    functions["LASTPOS"] = LASTPOS;
    functions["LEFT"] = LEFT;
    functions["LENGTH"] = LENGTH;
    functions["LINEIN"] = LINEIN;
    functions["LINEOUT"] = LINEOUT;
    functions["LINES"] = LINES;
    functions["MAX"] = MAX;
    functions["MIN"] = MIN;
    functions["OVERLAY"] = OVERLAY;
    functions["POS"] = POS;
    functions["QUEUED"] = QUEUED;
    functions["RANDOM"] = RANDOM;
    functions["REVERSE"] = REVERSE;
    functions["RIGHT"] = RIGHT;
    functions["SIGN"] = SIGN;
    functions["SOURCELINE"] = SOURCELINE;
    functions["SPACE"] = SPACE;
    functions["STREAM"] = STREAM;
    functions["STRIP"] = STRIP;
    functions["SUBSTR"] = SUBSTR;
    functions["SUBWORD"] = SUBWORD;
    functions["SYMBOL"] = SYMBOL;
    functions["TIME"] = TIME;
    functions["TRACE"] = TRACE;
    functions["TRANSLATE"] = TRANSLATE;
    functions["TRUNC"] = TRUNC;
    functions["VALUE"] = VALUE;
    functions["VERIFY"] = VERIFY;
    functions["WORD"] = WORD;
    functions["WORDINDEX"] = WORDINDEX;
    functions["WORDLENGTH"] = WORDLENGTH;
    functions["WORDPOS"] = WORDPOS;
    functions["WORDS"] = WORDS;
    functions["XRANGE"] = XRANGE;
    functions["X2B"] = X2B;
    functions["X2C"] = X2C;
    functions["X2D"] = X2D;
}

int Builtin::Exists(String f) {
    return functions.contains(f);
}

Functionp Builtin::GetFunction(String& f) {
    return functions[f];
}

void Builtin::Dispatch(String f) {
    if (Exists(f)) {
	functions[f]();
    }
}

