#ifndef _CLAUSE_H
#pragma once
#define _CLAUSE_H
class Expression;
class FunctionExpression;
class Variable;
#include "ArgStr.h"
#include <String.h>
#include "InternalData.h"
#include "Implementation.h"

class Clause {
  private:
    int linenumber;
    int clausenumber; // within a line, what clause is it?
    int ID;
  protected:
    void SetID(int);
  public:
    Clause();
    virtual ~Clause();
    int GetLineNumber();
    void SetLineNumber(int);
    int GetClauseNumber();
    void SetClauseNumber(int);
    void TraceBefore(); // traces clause before execution
    virtual int Execute();
    virtual int NoExecute();
    int IsA(int);
    int GetClassId();
};

inline int Clause::IsA(int type) {
    return ID == type;
}

inline int Clause::GetClassId() {
    return ID;
}

inline int Clause::GetLineNumber() {
    return linenumber;
}

inline int Clause::GetClauseNumber() {
    return clausenumber;
}

class NopClause : public Clause{
  public:
    NopClause();
    virtual ~NopClause();
    virtual int Execute();
    virtual int NoExecute();
//    virtual int IsA(int);
//    virtual int GetClassId();
};

class OptionsClause : public Clause {
  private:
    Expression* exp;
  public:
    OptionsClause();
    virtual ~OptionsClause();
    void SetExpression(Expression*);
    virtual int Execute();
    virtual int NoExecute();
//    virtual int IsA(int);
//    virtual int GetClassId();
};

class NumericClause : public Clause {
  private:
    Expression* digitsexp;
    Expression* formexp;
    Expression* fuzzexp;
  public:
    NumericClause();
    virtual ~NumericClause();
    void SetDigits(Expression*);
    void SetForm(Expression*);
    void SetFuzz(Expression*);
    virtual int Execute();
    virtual int NoExecute();
//    virtual int IsA(int);
//    virtual int GetClassId();
};

class PushClause : public Clause {
  private:
    Expression* exp;
  public:
    PushClause();
    virtual ~PushClause();
    void SetExpression(Expression*);
    virtual int Execute();
    virtual int NoExecute();
//    virtual int IsA(int);
//    virtual int GetClassId();
};

class QueueClause : public Clause {
  private:
    Expression* exp;
  public:
    QueueClause();
    virtual ~QueueClause();
    void SetExpression(Expression*);
    virtual int Execute();
    virtual int NoExecute();
//    virtual int IsA(int);
//    virtual int GetClassId();
};

class TraceClause : public Clause {
  private:
    Expression* exp;
  public:
    TraceClause();
    virtual ~TraceClause();
    void SetExpression(Expression*);
    virtual int Execute();
    virtual int NoExecute();
//    virtual int IsA(int);
//    virtual int GetClassId();
};

class AddressClause : public Clause {
  private:
    int command; // set if this is a command to the specified environment
    ArgString environment;
    Expression* exp;
  public:
    AddressClause();
    virtual ~AddressClause();
    void SetCommand();
    void SetEnvironment(ArgString);
    void SetExpression(Expression*);
    virtual int Execute();
    virtual int NoExecute();
//    virtual int IsA(int);
//    virtual int GetClassId();
};

class BreakClause : public Clause {
  public:
    BreakClause();
    virtual ~BreakClause();
    virtual int Execute();
    virtual int NoExecute();
//    virtual int IsA(int);
//    virtual int GetClassId();
};

class DropClause : public Clause{
  private:
    // we need an drop list here
    ArgString droplist[MaxDrops];
    int dropisref[MaxDrops];
    int count;
    void CheckBound(int);
  public:
    DropClause();
    virtual ~DropClause();
    virtual int Execute();
    virtual int NoExecute();
//    virtual int IsA(int);
//    virtual int GetClassId();
    void AddVariable(ArgString);
    void AddVariableRef(ArgString);
};

class LeaveClause : public Clause {
  private:
    ArgString variable;
  public:
    LeaveClause();
    virtual ~LeaveClause();
    virtual int Execute();
    virtual int NoExecute();
//    virtual int IsA(int);
//    virtual int GetClassId();
    void SetVariable(ArgString);
    ArgString GetVariable();
};

class IterateClause : public Clause {
  private:
    ArgString variable;
  public:
    IterateClause();
    virtual ~IterateClause();
    virtual int Execute();
    virtual int NoExecute();
//    virtual int IsA(int);
//    virtual int GetClassId();
    void SetVariable(ArgString);
    ArgString GetVariable();
};

enum ParseType {PDone, PVariable, PPeriod, PString, PAbsoluteInteger, PPositiveInteger, PNegativeInteger, PAbsoluteReference, PPositiveReference, PNegativeReference, PStringReference};

class ParseClause : public Clause {
  private:
    ArgString ptemplate[MaxTemplates][MaxPatterns];
    ParseType ttype[MaxTemplates][MaxPatterns]; // this is the type of thing in template[][]
    int currenttemplate; // which template we are filling
    int currentposition; // which thing in template[currenttemplate] we are filling
    int upper;
    void CheckBound(int, int);
  public:
    ParseClause();
    virtual ~ParseClause();
    virtual int Execute();
    virtual int NoExecute();
//    virtual int IsA(int);
//    virtual int GetClassId();
    
    void Upper();
    void AddPattern(ArgString, ParseType);
    void NextTemplate();
    virtual ArgString GetFirst();
    virtual ArgString GetNext();
};

class ParseVarClause : public ParseClause {
  private:
    ArgString variable;
    int secondary;
    ArgString value;
  public:
    ParseVarClause();
    virtual ~ParseVarClause();
    void SetVariable(ArgString, int);
    virtual ArgString GetFirst();
    virtual ArgString GetNext();
};

class ParseVersionClause : public ParseClause {
  public:
    ParseVersionClause();
    virtual ~ParseVersionClause();
    virtual ArgString GetFirst();
    virtual ArgString GetNext();
};

class ParseSourceClause : public ParseClause {
  public:
    ParseSourceClause();
    virtual ~ParseSourceClause();
    virtual ArgString GetFirst();
    virtual ArgString GetNext();
};

class ParseArgClause : public ParseClause {
  private:
    int number;
  public:
    ParseArgClause();
    virtual ~ParseArgClause();
    virtual ArgString GetFirst();
    virtual ArgString GetNext();
};

class ParseLineinClause : public ParseClause {
  public:
    ParseLineinClause();
    virtual ~ParseLineinClause();
    virtual ArgString GetFirst();
    virtual ArgString GetNext();
};

class ParseExternalClause : public ParseClause {
  public:
    ParseExternalClause();
    virtual ~ParseExternalClause();
    virtual ArgString GetFirst();
    virtual ArgString GetNext();
};

class ParsePullClause : public ParseClause {
  public:
    ParsePullClause();
    virtual ~ParsePullClause();
    virtual ArgString GetFirst();
    virtual ArgString GetNext();
};

class ParseValueClause : public ParseClause {
  private:
    Expression* exp;
  public:
    ParseValueClause();
    virtual ~ParseValueClause();
    virtual ArgString GetFirst();
    virtual ArgString GetNext();
    void SetExpression(Expression*);
};

class CallClause : public Clause{
  private:
    FunctionExpression* exp;
    ArgString SignalLabel;
    Signals TurnOn_;
    Signals TurnOff_;
  public:
    CallClause();
    virtual ~CallClause();
    virtual int Execute();
    virtual int NoExecute();
//    virtual int IsA(int);
//    virtual int GetClassId();
    void SetExpression(FunctionExpression*);
    void TurnOn(Signals, ArgString);
    void TurnOn(Signals);
    void TurnOff(Signals);
};

class InterpretClause : public Clause {
    Expression* exp;
  public:
    InterpretClause();
    ~InterpretClause();
    void SetExpression(Expression*);
//    virtual int IsA(int);
//    virtual int GetClassId();
    virtual int Execute();
    virtual int NoExecute();
};

class ProcedureClause : public Clause{
  private:
    // we need an expose list here
    ArgString exposelist[MaxExposes];
    int exposeisref[MaxExposes];
    int count;
    void CheckBound(int);
  public:
    ProcedureClause();
    virtual ~ProcedureClause();
    virtual int Execute();
    virtual int NoExecute();
//    virtual int IsA(int);
//    virtual int GetClassId();
    void AddVariable(ArgString);
    void AddVariableRef(ArgString);
};

class SignalClause : public Clause{
  private:
    Expression* exp;
    ArgString SignalLabel;
    Signals TurnOn_;
    Signals TurnOff_;
  public:
    SignalClause();
    virtual ~SignalClause();
    virtual int Execute();
    virtual int NoExecute();
//    virtual int IsA(int);
//    virtual int GetClassId();
    void SetExpression(Expression*);
    void TurnOn(Signals, ArgString);
    void TurnOn(Signals);
    void TurnOff(Signals);
};

class ErrorClause : public Clause {
  private:
  public:
    ErrorClause();
    virtual int Execute();
    virtual int NoExecute();
//    virtual int IsA(int);
//    virtual int GetClassId();
};

class NullClause : public Clause {
  public:
    NullClause();
//    virtual int IsA(int);
//    virtual int GetClassId();
};

class LabelClause : public Clause {
  private:
    String label;
  public:
    LabelClause(ArgString);
    int Matches(String);
//    virtual int IsA(int);
//    virtual int GetClassId();
};

class DoClause : public Clause {
  private:
    int DoPC; // start of the Do statement in program counter terms
    // do forever
    int Forever;

    int EndLabelChecked;
    int EndLabelMatched;

    ArgString symbol;
    String ssymbol;
    Variable* vsymbol;
    int secondary; // type of symbol

    Expression* From;
    RexxNumber FromNumber; // calc version of from
    long FromLong;

    int LongBy;
    int LongTo;
    int LongFrom;

    // limit
    Expression* To;
    RexxNumber ToNumber; // calc version of to
    long ToLong;

    // increment
    Expression* By;
    RexxNumber ByNumber; // calc version of by
    long ByLong;

    // max iterations
    Expression* For;
    int ForInt; // calculated version of For

    // while expression
    Expression* While;
    int WhileInt; // boolean

    // until expression
    Expression* Until;
    int UntilInt; // boolean

    inline int IsRepetitive();
    inline int DoEndStuff();
  public:
    DoClause();
    ~DoClause();
//    virtual int IsA(int);
//    virtual int GetClassId();
    virtual int Execute();
    virtual int NoExecute();
    // returns 0 or 1, depending on success -- could conflict
    inline void SetTop(); // set the start of the do loop
    inline void JumpTop(); // jump to the beginning of the loop
    inline int StartEvaluate(); // evaluate things at start
    inline int StartDone(); // returns 1 when start conditions are false
    inline int EndDone();  // does everything we do at the end of a loop, then checks
    inline int WhileEvaluate();
    inline int UntilEvaluate();
    inline int EndEvaluate();
    inline int UpdateIndex();
    inline void WriteSymbol(RexxNumber&);
    inline ArgString& ReadSymbol();
    int SetForever();
    int SetSymbol(ArgString&, int);
    int SetFrom(Expression*);
    int SetTo(Expression*);
    int SetBy(Expression*);
    int SetFor(Expression*);
    int SetWhile(Expression*);
    int SetUntil(Expression*);
};

class EndClause : public Clause {
  private:
    ArgString variable;
  public:
    EndClause();

//    virtual int IsA(int);
//    virtual int GetClassId();
    virtual int Execute();
    virtual int NoExecute();
    void SetVariable(ArgString&);
    ArgString& GetVariable();
};

class WhenClause : public Clause {
  private:
    Expression* exp;
  public:
    WhenClause();
    void SetExpression(Expression*);
//    virtual int IsA(int);
//    virtual int GetClassId();
    int PrivateExecute();
    int PrivateNoExecute();
    virtual int Execute();
    virtual int NoExecute();
};

class OtherwiseClause : public Clause {
  public:
    OtherwiseClause();
//    virtual int IsA(int);
//    virtual int GetClassId();
    virtual int Execute();
    virtual int NoExecute();
};

class SelectClause : public Clause {
  public:
    SelectClause();
//    virtual int IsA(int);
//    virtual int GetClassId();
    virtual int Execute();
    virtual int NoExecute();
};

class IfClause : public Clause {
  private:
    Expression* exp;
  public:
    IfClause();
    void SetExpression(Expression*);
//    virtual int IsA(int);
//    virtual int GetClassId();
    virtual int Execute();
    virtual int NoExecute();
};

class ThenClause : public Clause {
  public:
    ThenClause();
//    virtual int IsA(int);
//    virtual int GetClassId();
    virtual int Execute();
    virtual int NoExecute();
};

class ElseClause : public Clause {
  public:
    ElseClause();
//    virtual int IsA(int);
//    virtual int GetClassId();
    virtual int Execute();
    virtual int NoExecute();
};

class SayClause : public Clause {
  private:
    Expression* exp;
  public:
    SayClause();
    void SetExpression(Expression*);
//    virtual int IsA(int);
//    virtual int GetClassId();
    virtual int Execute();
    virtual int NoExecute();
};

class CommandClause : public Clause {
  private:
    Expression* exp;
  public:
    CommandClause();
    void SetExpression(Expression*);
//    virtual int IsA(int);
//    virtual int GetClassId();
    virtual int Execute();
    virtual int NoExecute();
};

class AssignmentClause : public Clause {
  private:
//    String symbol;
//    int secondary; // type of symbol
    Variable* vsymbol;
    Expression* exp;
  public:
    AssignmentClause(ArgString&, int, Expression*);
    virtual ~AssignmentClause();
//    virtual int IsA(int);
//    virtual int GetClassId();
    virtual int Execute();
    virtual int NoExecute();
};

class ExitClause : public Clause {
    Expression* exp;
  public:
    ExitClause();
    void SetExpression(Expression*);
//    virtual int IsA(int);
//    virtual int GetClassId();
    virtual int Execute();
    virtual int NoExecute();
};

// this gets tacked on at the end of a parse
class EofClause : public Clause {
  public:
    EofClause();
//    virtual int IsA(int);
//    virtual int GetClassId();
    virtual int Execute();
    virtual int NoExecute();
};

class ReturnClause : public Clause {
    Expression* exp;
  public:
    ReturnClause();
    void SetExpression(Expression*);
//    virtual int IsA(int);
//    virtual int GetClassId();
    virtual int Execute();
    virtual int NoExecute();
};


#endif
