// This may look like C code, but it is really -*- C++ -*-
/* 
Copyright (C) 1988 Free Software Foundation
    written by Doug Lea (dl@rocky.oswego.edu)
    based on code by Marc Shapiro (shapiro@sor.inria.fr)

This file is part of GNU CC.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY.  No author or distributor
accepts responsibility to anyone for the consequences of using it
or for whether it serves any particular purpose or works at all,
unless he says so in writing.  Refer to the GNU CC General Public
License for full details.

Everyone is granted permission to copy, modify and redistribute
GNU CC, but only under the conditions described in the
GNU CC General Public License.   A copy of this license is
supposed to have been given to you along with GNU CC so you
can know your rights and responsibilities.  It should be in a
file named COPYING.  Among other things, the copyright notice
and this notice must be preserved on all copies.  
*/

#ifdef __GNUG__
#pragma implementation
#endif
#include "Clausep.XPlex.h"


ClausepXPlex:: ClausepXPlex()
{
  lo = fnc = 0;
  csize = DEFAULT_INITIAL_CAPACITY;
  Clausep* data = new Clausep[csize];
  set_cache(new ClausepIChunk(data,  lo, lo, fnc, lo+csize));
  hd = ch;
}

ClausepXPlex:: ClausepXPlex(int chunksize)
{
  if (chunksize == 0) error("invalid constructor specification");
  lo = fnc = 0;
  if (chunksize > 0)
  {
    csize = chunksize;
    Clausep* data = new Clausep[csize];
    set_cache(new ClausepIChunk(data,  lo, lo, fnc, csize));
    hd = ch;
  }
  else
  {
    csize = -chunksize;
    Clausep* data = new Clausep[csize];
    set_cache(new ClausepIChunk(data,  chunksize, lo, fnc, fnc));
    hd = ch;
  }
}


ClausepXPlex:: ClausepXPlex(int l, int chunksize)
{
  if (chunksize == 0) error("invalid constructor specification");
  lo = fnc = l;
  if (chunksize > 0)
  {
    csize = chunksize;
    Clausep* data = new Clausep[csize];
    set_cache(new ClausepIChunk(data,  lo, lo, fnc, csize+lo));
    hd = ch;
  }
  else
  {
    csize = -chunksize;
    Clausep* data = new Clausep[csize];
    set_cache(new ClausepIChunk(data,  chunksize+lo, lo, fnc, fnc));
    hd = ch;
  }
}

void ClausepXPlex::make_initial_chunks(int up)
{
  int need = fnc - lo;
  hd = 0;
  if (up)
  {
    int l = lo;
    do
    {
      int sz;
      if (need >= csize)
        sz = csize;
      else
        sz = need;
      Clausep* data = new Clausep [csize];
      ClausepIChunk* h = new ClausepIChunk(data,  l, l, l+sz, l+csize);
      if (hd != 0)
        h->link_to_next(hd);
      else
        hd = h;
      l += sz;
      need -= sz;
    } while (need > 0);
  }
  else
  {
    int hi = fnc;
    do
    {
      int sz;
      if (need >= csize)
        sz = csize;
      else
        sz = need;
      Clausep* data = new Clausep [csize];
      ClausepIChunk* h = new ClausepIChunk(data,  hi-csize, hi-sz, hi, hi);
      if (hd != 0)
        h->link_to_next(hd);
      hd = h;
      hi -= sz;
      need -= sz;
    } while (need > 0);
  }
  set_cache(hd);
}

ClausepXPlex:: ClausepXPlex(int l, int hi, const Clausep  initval, int chunksize)
{
  lo = l;
  fnc = hi + 1;
  if (chunksize == 0)
  {
    csize = fnc - l;
    make_initial_chunks(1);
  }
  else if (chunksize < 0)
  {
    csize = -chunksize;
    make_initial_chunks(0);
  }
  else
  {
    csize = chunksize;
    make_initial_chunks(1);
  }
  fill(initval);
}

ClausepXPlex::ClausepXPlex(const ClausepXPlex& a)
{
  lo = a.lo;
  fnc = a.fnc;
  csize = a.csize;
  make_initial_chunks();
  for (int i = a.low(); i < a.fence(); a.next(i)) (*this)[i] = a[i];
}

void ClausepXPlex::operator= (const ClausepXPlex& a)
{
  if (&a != this)
  {
    invalidate();
    lo = a.lo;
    fnc = a.fnc;
    csize = a.csize;
    make_initial_chunks();
    for (int i = a.low(); i < a.fence(); a.next(i)) (*this)[i] = a[i];
  }
}


void ClausepXPlex::cache(int idx) const
{
  const ClausepIChunk* tail = tl();
  const ClausepIChunk* t = ch;
  while (idx >= t->fence_index())
  {
    if (t == tail) index_error();
    t = (t->next());
  }
  while (idx < t->low_index())
  {
    if (t == hd) index_error();
    t = (t->prev());
  }
  set_cache(t);
}


void ClausepXPlex::cache(const Clausep* p) const
{
  const ClausepIChunk* old = ch;
  const ClausepIChunk* t = ch;
  while (!t->actual_pointer(p))
  {
    t = (t->next());
    if (t == old) index_error();
  }
  set_cache(t);
}

int ClausepXPlex::owns(Pix px) const
{
  Clausep* p = (Clausep*)px;
  const ClausepIChunk* old = ch;
  const ClausepIChunk* t = ch;
  while (!t->actual_pointer(p))
  {
    t = (t->next());
    if (t == old) { set_cache(t); return 0; }
  }
  set_cache(t);
  return 1;
}


Clausep* ClausepXPlex::dosucc(const Clausep* p) const
{
  if (p == 0) return 0;
  const ClausepIChunk* old = ch;
  const ClausepIChunk* t = ch;
 
  while (!t->actual_pointer(p))
  {
    t = (t->next());
    if (t == old)  return 0;
  }
  int i = t->index_of(p) + 1;
  if (i >= fnc) return 0;
  if (i >= t->fence_index()) t = (t->next());
  set_cache(t);
  return (t->pointer_to(i));
}

Clausep* ClausepXPlex::dopred(const Clausep* p) const
{
  if (p == 0) return 0;
  const ClausepIChunk* old = ch;
  const ClausepIChunk* t = ch;
  while (!t->actual_pointer(p))
  {
    t = (t->prev());
    if (t == old) return 0;
  }
  int i = t->index_of(p) - 1;
  if (i < lo) return 0;
  if (i < t->low_index()) t = (t->prev());
  set_cache(t);
  return (t->pointer_to(i));
}


int ClausepXPlex::add_high(const Clausep  elem)
{
  ClausepIChunk* t = tl();
  if (!t->can_grow_high())
  {
    if (t->ClausepIChunk::empty() && one_chunk())
      t->clear(fnc);
    else
    {
      Clausep* data = new Clausep [csize];
      t = (new ClausepIChunk(data,  fnc, fnc, fnc,fnc+csize));
      t->link_to_prev(tl());
    }
  }
  *((t->ClausepIChunk::grow_high())) = elem;
  set_cache(t);
  return fnc++;
}

int ClausepXPlex::del_high ()
{
  if (empty()) empty_error();
  ClausepIChunk* t = tl();
  t->ClausepIChunk::shrink_high();
  if (t->ClausepIChunk::empty() && !one_chunk())
  {
    ClausepIChunk* pred = t->prev();
    del_chunk(t);
    t = pred;
  }
  set_cache(t);
  return --fnc - 1;
}

int ClausepXPlex::add_low (const Clausep  elem)
{
  ClausepIChunk* t = hd;
  if (!t->can_grow_low())
  {
    if (t->ClausepIChunk::empty() && one_chunk())
      t->cleardown(lo);
    else
    {
      Clausep* data = new Clausep [csize];
      hd = new ClausepIChunk(data,  lo-csize, lo, lo, lo);
      hd->link_to_next(t);
      t = hd;
    }
  }
  *((t->ClausepIChunk::grow_low())) = elem;
  set_cache(t);
  return --lo;
}


int ClausepXPlex::del_low ()
{
  if (empty()) empty_error();
  ClausepIChunk* t = hd;
  t->ClausepIChunk::shrink_low();
  if (t->ClausepIChunk::empty() && !one_chunk())
  {
    hd = t->next();
    del_chunk(t);
    t = hd;
  }
  set_cache(t);
  return ++lo;
}

void ClausepXPlex::append (const ClausepXPlex& a)
{
  for (int i = a.low(); i < a.fence(); a.next(i)) add_high(a[i]);
}

void ClausepXPlex::prepend (const ClausepXPlex& a)
{
  for (int i = a.high(); i > a.ecnef(); a.prev(i)) add_low(a[i]);
}

void ClausepXPlex::reverse()
{
  Clausep tmp;
  int l = lo;
  int h = fnc - 1;
  ClausepIChunk* loch = hd;
  ClausepIChunk* hich = tl();
  while (l < h)
  {
    Clausep* lptr = loch->pointer_to(l);
    Clausep* hptr = hich->pointer_to(h);
    tmp = *lptr;
    *lptr = *hptr;
    *hptr = tmp;
    if (++l >= loch->fence_index()) loch = loch->next();
    if (--h < hich->low_index()) hich = hich->prev();
  }
}

void ClausepXPlex::fill(const Clausep  x)
{
  for (int i = lo; i < fnc; ++i) (*this)[i] = x;
}

void ClausepXPlex::fill(const Clausep  x, int l, int hi)
{
  for (int i = l; i <= hi; ++i) (*this)[i] = x;
}


void ClausepXPlex::clear()
{
  if (fnc != lo)
  {
    ClausepIChunk* t = tl();
    while (t != hd)
    {
      ClausepIChunk* prv = t->prev();
      del_chunk(t);
      t = prv;
    }
    t->ClausepIChunk::clear(lo);
    set_cache(t);
    fnc = lo;
  }
}


int ClausepXPlex::OK () const
{
  int v = hd != 0 && ch != 0;     // at least one chunk

  v &= fnc == tl()->fence_index();// last chunk fence == plex fence
  v &= lo == ((hd))->ClausepIChunk::low_index();    // first lo == plex lo

// loop for others:
  int found_ch = 0;                   // to make sure ch is in list;
  const ClausepIChunk* t = (hd);
  for (;;)
  {
    if (t == ch) ++found_ch;
    v &= t->ClausepIChunk::OK();              // each chunk is OK
    if (t == tl())
      break;
    else                              // and has indices contiguous to succ
    {
      v &= t->top_index() == t->next()->base_index();
      if (t != hd)                  // internal chunks full
      {
        v &= !t->empty();
        v &= !t->can_grow_low();
        v &= !t->can_grow_high();
      }
      t = t->next();
    }
  }
  v &= found_ch == 1;
  if (!v) error("invariant failure");
  return v;
}
