// This may look like C code, but it is really -*- C++ -*-
/* 
Copyright (C) 1988 Free Software Foundation
    written by Doug Lea (dl@rocky.oswego.edu)
    based on code by Marc Shapiro (shapiro@sor.inria.fr)

This file is part of GNU CC.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY.  No author or distributor
accepts responsibility to anyone for the consequences of using it
or for whether it serves any particular purpose or works at all,
unless he says so in writing.  Refer to the GNU CC General Public
License for full details.

Everyone is granted permission to copy, modify and redistribute
GNU CC, but only under the conditions described in the
GNU CC General Public License.   A copy of this license is
supposed to have been given to you along with GNU CC so you
can know your rights and responsibilities.  It should be in a
file named COPYING.  Among other things, the copyright notice
and this notice must be preserved on all copies.  
*/

#ifndef _ClausepXPlex_h
#ifdef __GNUG__
#pragma once
#pragma interface
#endif
#define _ClausepXPlex_h 1

#include "Clausep.Plex.h"

class ClausepXPlex: public ClausepPlex
{
  ClausepIChunk*       ch;           // cached chunk

  void             make_initial_chunks(int up = 1);

  void             cache(int idx) const;
  void             cache(const Clausep* p) const;

  Clausep*             dopred(const Clausep* p) const;
  Clausep*             dosucc(const Clausep* p) const;

  void             set_cache(const ClausepIChunk* t) const; // logically, 
                                               // not physically const
public:
                   ClausepXPlex();                 // set low = 0;
                                               // fence = 0;
                                               // csize = default

                   ClausepXPlex(int ch_size);      // low = 0; 
                                               // fence = 0;
                                               // csize = ch_size

                   ClausepXPlex(int lo,            // low = lo; 
                            int ch_size);      // fence=lo
                                               // csize = ch_size

                   ClausepXPlex(int lo,            // low = lo
                            int hi,            // fence = hi+1
                            const Clausep  initval,// fill with initval,
                            int ch_size = 0);  // csize= ch_size
                                               // or fence-lo if 0

                   ClausepXPlex(const ClausepXPlex&);
  
  void             operator= (const ClausepXPlex&);

// virtuals


  Clausep&             high_element ();
  Clausep&             low_element ();

  const Clausep&       high_element () const;
  const Clausep&       low_element () const;

  Pix              first() const;
  Pix              last() const;
  void             prev(Pix& ptr) const;
  void             next(Pix& ptr) const;
  int              owns(Pix p) const;
  Clausep&             operator () (Pix p);
  const Clausep&       operator () (Pix p) const;

  int              low() const; 
  int              high() const;
  int              valid(int idx) const;
  void             prev(int& idx) const;
  void             next(int& x) const;
  Clausep&             operator [] (int index);
  const Clausep&       operator [] (int index) const;
    
  int              Pix_to_index(Pix p) const;
  Pix              index_to_Pix(int idx) const;    

  int              can_add_high() const;
  int              can_add_low() const;
  int              full() const;

  int              add_high(const Clausep  elem);
  int              del_high ();
  int              add_low (const Clausep  elem);
  int              del_low ();

  void             fill(const Clausep  x);
  void             fill(const Clausep  x, int from, int to);
  void             clear();
  void             reverse();
  void             append(const ClausepXPlex& a); 
  void             prepend(const ClausepXPlex& a);
    
  int              OK () const; 

};

inline void ClausepXPlex::prev(int& idx) const
{
  --idx;
}

inline void ClausepXPlex::next(int& idx) const
{
  ++idx;
}

inline  int ClausepXPlex::full () const
{
  return 0;
}

inline int ClausepXPlex::can_add_high() const
{
  return 1;
}

inline int ClausepXPlex::can_add_low() const
{
  return 1;
}

inline  int ClausepXPlex::valid (int idx) const
{
  return idx >= lo && idx < fnc;
}

inline int ClausepXPlex::low() const
{
  return lo;
}

inline int ClausepXPlex::high() const
{
  return fnc - 1;
}

inline Clausep& ClausepXPlex:: operator [] (int idx)
{
  if (!ch->actual_index(idx)) cache(idx);
  return *(ch->pointer_to(idx));
}

inline const Clausep& ClausepXPlex:: operator [] (int idx) const
{
  if (!ch->actual_index(idx)) cache(idx);
  return *((const Clausep*)(ch->pointer_to(idx)));
}

inline  Clausep& ClausepXPlex::low_element ()
{
  if (empty()) index_error();
  return *(hd->pointer_to(lo));
}

inline  const Clausep& ClausepXPlex::low_element () const
{
  if (empty()) index_error();
  return *((const Clausep*)(hd->pointer_to(lo)));
}

inline  Clausep& ClausepXPlex::high_element ()
{
  if (empty()) index_error();
  return *(tl()->pointer_to(fnc - 1));
}

inline const Clausep& ClausepXPlex::high_element () const
{
  if (empty()) index_error();
  return *((const Clausep*)(tl()->pointer_to(fnc - 1)));
}

inline  int ClausepXPlex::Pix_to_index(Pix px) const
{
  Clausep* p = (Clausep*)px;
  if (!ch->actual_pointer(p)) cache(p);
  return ch->index_of(p);
}

inline  Pix ClausepXPlex::index_to_Pix(int idx) const
{
  if (!ch->actual_index(idx)) cache(idx);
  return (Pix)(ch->pointer_to(idx));
}

inline Pix ClausepXPlex::first() const
{
  return Pix(hd->ClausepIChunk::first_pointer());
}

inline Pix ClausepXPlex::last() const
{
  return Pix(tl()->ClausepIChunk::last_pointer());
}

inline void ClausepXPlex::prev(Pix& p) const
{
  Pix q = Pix(ch->ClausepIChunk::pred((Clausep*) p));
  p = (q == 0)? Pix(dopred((const Clausep*) p)) : q;
}

inline void ClausepXPlex::next(Pix& p) const
{
  Pix q = Pix(ch->ClausepIChunk::succ((Clausep*) p));
  p = (q == 0)? Pix(dosucc((const Clausep*)p)) : q;
}

inline Clausep& ClausepXPlex:: operator () (Pix p)
{
  return *((Clausep*)p);
}

inline const Clausep& ClausepXPlex:: operator () (Pix p) const
{
  return *((const Clausep*)p);
}

inline void ClausepXPlex::set_cache(const ClausepIChunk* t) const
{
  ((ClausepXPlex*)(this))->ch = (ClausepIChunk*)t;
}

#endif
