#include "Express.h"
#include "ArgStr.h"
#include "SymTabs.h"
#include "InternalEnv.h"
#include "Expressionp.SLList.h"
#include "support.h"
#include <String.h>
#include <math.h>
#include "Glob.h"
#include "SymTab.h"
#include "stringsupport.h"
#include "Trace.h"

#define BADRETURN {returnvalue.Invalidate(); return returnvalue;}

static ArgString ASTrue((int)1);
static ArgString ASFalse((int)0);

extern Glob glob;
// extern String rexx_compound;

// still need to do proper REXX arithmetic
// still need to do proper comparisons (inexact stuff) -- strip off begin/end whitespace and pad shorter string

Expression::Expression() {}

Expression::~Expression() {}

const ArgString& Expression::Evaluate() {
    static ArgString returnvalue;
    SetError(EInterpretationError);
    BADRETURN;
}

//

ErrorExpression::ErrorExpression(Errors e) : error(e) {}

ErrorExpression::~ErrorExpression() {}

const ArgString& ErrorExpression::Evaluate() {
    static ArgString returnvalue;
    SetError(error);
    BADRETURN;
}

//

FunctionExpression::FunctionExpression(const ArgString& a, ExpressionpSLList* args, int skip) : IsCall(0), ShouldSkipInternalSearch(skip), functiontype(UnresolvedFunction), Internal(0), Builtin(0) {
    label = a.String();
    // there was a side-effect in rexxparse.y
    if (args->length() == 1) {
	// check if it's a 0 -- if so, clear the list
	Pix i = args->first();
	if (!(*args)(i)) args->clear();
    }
    arguments = args;
}

FunctionExpression::~FunctionExpression() {
    for (Pix i = arguments->first(); i != 0; arguments->next(i)) {
	if ((*arguments)(i)) delete (*arguments)(i);
    }
    delete arguments;
}

void FunctionExpression::SetCall() {
    IsCall = 1;
}

extern ArgString function(String, int, int = 0); // prototype of function in main.cc
extern ArgString internalfunction(int);

const ArgString& FunctionExpression::Evaluate() {
    if (functiontype == UnresolvedFunction) {
	// do the majority of the resolving here
	if (!ShouldSkipInternalSearch && glob.Clauses->Exists(label, GetStart(), GetEnd())) {
	    // an internal function
	    Internal = glob.Clauses->Find(label, GetStart(), GetEnd());
	    assert(Internal != -1);
	    functiontype = InternalFunction;
	} else if (glob.BuiltinFunctions.Exists(label)) {
	    // a builtin function
	    Builtin = glob.BuiltinFunctions.GetFunction(label);
	    functiontype = BuiltinFunction;
	} else {
	    // possibly an external function
	    functiontype = ExternalFunction;
	}
    }
    ArgString& result =  ActualEvaluate();
    return result;
}

const ArgString& FunctionExpression::ActualEvaluate() {
    static ArgString returnvalue;
    
    glob.Environments.PushEnv();
    // evaluate arguments here
    int count = 0;
    for (Pix i = arguments->first(); i != 0; arguments->next(i)) {
	if ((*arguments)(i)) {
	    Args(count) = (*arguments)(i)->Evaluate();
	    ArgsOmitted(count) = 0;
	    // check if we should trap here
	    if (!Args(count).IsValid()) BADRETURN;
	} else {
	    // Args(count) = ArgString(""); // put an invalid argument string here
	    ArgsOmitted(count) = 1;
	}
	count++;
    }
    // SetNumArguments(count);
    int previousline = glob.LineNumber;
    // should set SIGL here, right before we actually call a routine!
    glob.SymbolTables.Write("SIGL", ArgString(glob.LineNumber));

//    returnvalue =  function(label, IsCall, ShouldSkipInternalSearch);
    switch (functiontype) {
      case InternalFunction :
	returnvalue = internalfunction(Internal);
	break;
      case BuiltinFunction :
	Builtin();
	returnvalue = ReturnValue();
	break;
      case ExternalFunction :
	returnvalue = function(label, IsCall, ShouldSkipInternalSearch);
	break;
      case UnresolvedFunction :
	abort();
    };

    if (HasSymbolTable()) delete glob.SymbolTables.PopSymbolTable(); // pop the one we created if we have one
    delete glob.Environments.PopEnv();

    if (!returnvalue.IsValid()) {
	// ensure that we return a value
	if (IsReturn()) {
	    // we must have 'return'ed without specifying a value
	    if (!IsCall) {
		SetError(ENoDataSpecifiedOnFunctionReturn);
	    }
	}
    }
    glob.LineNumber = previousline;
    return returnvalue;
}

//

LiteralExpression::LiteralExpression(const ArgString& a) {
    string = a;
}

LiteralExpression::~LiteralExpression() {
}

const ArgString& LiteralExpression::Evaluate() {
    return string;
}

//

SimpleSymExpression::SimpleSymExpression(const ArgString& a) : vsymbol(a.String()) {
//    symbol = a.String();
}

SimpleSymExpression::~SimpleSymExpression() {}

const ArgString& SimpleSymExpression::Evaluate() {
    static ArgString returnvalue;

    ArgString& temp = vsymbol.Read();

    //ArgString& temp = glob.SymbolTables.Read(symbol);

    if (glob.UninitializedAccess) {
	if (SignalOn(SigNoValue)) {
	    SetError(ENoValue);
	    SetSignalDescription(SigNoValue, temp);
	    BADRETURN;
	}
    }

    if (temp.IsValid()) {
	return temp;
    } else {
	abort(); // should never get here
    }
}

//

StemSymExpression::StemSymExpression(const ArgString& a) : vsymbol(a.String()) {
//    symbol = a.String();
}

StemSymExpression::~StemSymExpression() {}

const ArgString& StemSymExpression::Evaluate() {
    static ArgString returnvalue;

    ArgString& temp = vsymbol.Read();
    //    ArgString& temp = glob.SymbolTables.ReadStem(symbol);

    if (glob.UninitializedAccess) {
	// do something here if we signal on novalue
	if (SignalOn(SigNoValue)) {
	    SetError(ENoValue);
	    SetSignalDescription(SigNoValue, temp);
	    BADRETURN;
	}
    }

    if (temp.IsValid()) {
	return temp;
    } else {
	abort();
    }
}

//

CompoundSymExpression::CompoundSymExpression(const ArgString& a) : vsymbol(a.String()) {
//    symbol = a.String();
}

CompoundSymExpression::~CompoundSymExpression() {}

const ArgString& CompoundSymExpression::Evaluate() {
    static ArgString returnvalue;
    ArgString& temp = vsymbol.Read();
//    ArgString& temp = glob.SymbolTables.ReadCompound(symbol);

    if (glob.UninitializedAccess) {
	// do something here if we signal on novalue
	if (SignalOn(SigNoValue)) {
	    SetError(ENoValue);
	    SetSignalDescription(SigNoValue, temp);
	    BADRETURN;
	}
    }

    if (temp.IsValid()) {
	return temp;
    } else {
	abort();
    }
}

//

UnaryExpression::UnaryExpression(Expression* e) : exp(e) {}

UnaryExpression::~UnaryExpression() {
    delete exp;
}

//

PositiveExpression::PositiveExpression(Expression* e) : UnaryExpression(e) {}

PositiveExpression::~PositiveExpression() {}

const ArgString& PositiveExpression::Evaluate() {
    static ArgString returnvalue;

    ArgString& a = exp->Evaluate();
    if (!a.IsValid()) BADRETURN;
    if (!a.IsANumber()) {
	SetError(EBadArithmeticConversion);
	BADRETURN;
    }

    static RexxNumber temp;
    add(0, a.Number(), GetDigits(), temp);
    returnvalue = Form(temp, GetDigits());
    return returnvalue;
}

//

NegativeExpression::NegativeExpression(Expression* e) : UnaryExpression(e) {}

NegativeExpression::~NegativeExpression() {}

const ArgString& NegativeExpression::Evaluate() {
    static ArgString returnvalue;

    ArgString& a = exp->Evaluate();
    if (!a.IsValid()) BADRETURN;
    if (!a.IsANumber()) {
	SetError(EBadArithmeticConversion);
	BADRETURN;
    }
//    static RexxNumber n;
//    n = a.Number();
//    n.negate();
    static RexxNumber temp;
//    add(0, n, GetDigits(), temp);
    subtract(0, a.Number(), GetDigits(), temp);
    returnvalue = Form(temp, GetDigits());
    return returnvalue;
}

//

BinaryExpression::BinaryExpression(Expression* e1, Expression* e2) : exp1(e1), exp2(e2) {}

BinaryExpression::~BinaryExpression() {
    delete exp1;
    delete exp2;
}

//

ExactEqualExpression::ExactEqualExpression(Expression* e1, Expression* e2) : BinaryExpression(e1, e2) {}

ExactEqualExpression::~ExactEqualExpression() {}

const ArgString& ExactEqualExpression::Evaluate() {
    static ArgString s1;
    static ArgString returnvalue;

    s1 = exp1->Evaluate();
    if (!s1.IsValid()) BADRETURN;
    ArgString& s2 = exp2->Evaluate();
    if (!s2.IsValid()) BADRETURN;
    // should do some trapping here!
//    returnvalue = (int)(s1 == s2);
//    return returnvalue;
    return (s1 == s2) ? ASTrue : ASFalse;
}

//

ExactNotEqualExpression::ExactNotEqualExpression(Expression* e1, Expression* e2) : BinaryExpression(e1, e2) {}

ExactNotEqualExpression::~ExactNotEqualExpression() {}

const ArgString& ExactNotEqualExpression::Evaluate() {
    static ArgString s1;
    static ArgString returnvalue;

    s1 = exp1->Evaluate();
    if (!s1.IsValid()) BADRETURN;
    ArgString& s2 = exp2->Evaluate();
    if (!s2.IsValid()) BADRETURN;

//    returnvalue = (int)(s1 != s2);
//    return returnvalue;
    return (s1 != s2) ? ASTrue : ASFalse;
}

//

ExactLessThanExpression::ExactLessThanExpression(Expression* e1, Expression* e2) : BinaryExpression(e1, e2) {}

ExactLessThanExpression::~ExactLessThanExpression() {}

const ArgString& ExactLessThanExpression::Evaluate() {
    static ArgString s1;
    static ArgString returnvalue;

    s1 = exp1->Evaluate();
    if (!s1.IsValid()) BADRETURN;
    ArgString& s2 = exp2->Evaluate();
    if (!s2.IsValid()) BADRETURN;
    // should do some trapping here!
//    returnvalue = (int)(s1 < s2);
//    return returnvalue;
    return (s1 < s2) ? ASTrue : ASFalse;
}

//

ExactGreaterThanExpression::ExactGreaterThanExpression(Expression* e1, Expression* e2) : BinaryExpression(e1, e2) {}

ExactGreaterThanExpression::~ExactGreaterThanExpression() {}

const ArgString& ExactGreaterThanExpression::Evaluate() {
    static ArgString s1;
    static ArgString returnvalue;

    s1 = exp1->Evaluate();
    if (!s1.IsValid()) BADRETURN;
    ArgString& s2 = exp2->Evaluate();
    if (!s2.IsValid()) BADRETURN;

//    returnvalue = (int)(s1 > s2);
//    return returnvalue;
    return (s1 > s2) ? ASTrue : ASFalse;
}

//

ExactGreaterEqualExpression::ExactGreaterEqualExpression(Expression* e1, Expression* e2) : BinaryExpression(e1, e2) {}

ExactGreaterEqualExpression::~ExactGreaterEqualExpression() {}

const ArgString& ExactGreaterEqualExpression::Evaluate() {
    static ArgString s1;
    static ArgString returnvalue;

    s1 = exp1->Evaluate();
    if (!s1.IsValid()) BADRETURN;
    ArgString& s2 = exp2->Evaluate();
    if (!s2.IsValid()) BADRETURN;

//    returnvalue = (int)(s1 >= s2);
//    return returnvalue;
    return (s1 >= s2) ? ASTrue : ASFalse;
}

//

ExactLessEqualExpression::ExactLessEqualExpression(Expression* e1, Expression* e2) : BinaryExpression(e1, e2) {}

ExactLessEqualExpression::~ExactLessEqualExpression() {}

const ArgString& ExactLessEqualExpression::Evaluate() {
    static ArgString s1;
    static ArgString returnvalue;
    char* c;
    s1 = exp1->Evaluate();
    if (!s1.IsValid()) BADRETURN;
    ArgString& s2 = exp2->Evaluate();
    if (!s2.IsValid()) BADRETURN;

//    returnvalue = (int)(s1 <= s2);
//    return returnvalue;
    return (s1 <= s2) ? ASTrue : ASFalse;
}



//

EqualExpression::EqualExpression(Expression* e1, Expression* e2) : BinaryExpression(e1, e2) {}

EqualExpression::~EqualExpression() {}

const ArgString& EqualExpression::Evaluate() {
    static ArgString s1;
    static ArgString s2;
    static ArgString returnvalue;

    s1 = exp1->Evaluate();
    if (!s1.IsValid()) BADRETURN;
    s2 = exp2->Evaluate();
    if (!s2.IsValid()) BADRETURN;

    // should do some trapping here!
    if (s1.IsANumber() && s2.IsANumber()) {
	// need to do rexx comparison here
//	RexxNumber& n1 = s1.Number();
//	static RexxNumber n2;
//	n2 = s2.Number();
//	n2.negate();
	static RexxNumber comp;
	add(s1.Number(), s2.Number(), GetDigits() - GetFuzz(), comp);
	subtract(s1.Number(), s2.Number(), GetDigits() - GetFuzz(), comp);
//	returnvalue = (int)comp.sign() == 0;
//	return returnvalue;
	return (comp.sign() == 0) ? ASTrue : ASFalse;
    }

    // otherwise, doing a string comparison
    s1 = stripfront(striprear(s1));
    s2 = stripfront(striprear(s2));
    if (s1.Length() > s2.Length()) {
	s2 = padonright(s2, s1.Length(), ' ');
    } else {
	s1 = padonright(s1, s2.Length(), ' ');
    }
//    returnvalue = (int)(s1 == s2);
//    return returnvalue;
    return (s1 == s2) ? ASTrue : ASFalse;
}

//

NotEqualExpression::NotEqualExpression(Expression* e1, Expression* e2) : BinaryExpression(e1, e2) {}

NotEqualExpression::~NotEqualExpression() {}

const ArgString& NotEqualExpression::Evaluate() {
    static ArgString s1;
    static ArgString s2;
    static ArgString returnvalue;

    s1 = exp1->Evaluate();
    if (!s1.IsValid()) BADRETURN;
    s2 = exp2->Evaluate();
    if (!s2.IsValid()) BADRETURN;

    if (s1.IsANumber() && s2.IsANumber()) {
	// need to do rexx comparison here
//	RexxNumber& n1 = s1.Number();
//	static RexxNumber n2;
//	n2 = s2.Number();
//	n2.negate();
	static RexxNumber comp;
//	add(n1, n2, GetDigits() - GetFuzz(), comp);
	subtract(s1.Number(), s2.Number(), GetDigits() - GetFuzz(), comp);
//	returnvalue = (int)comp.sign() != 0;
//	return returnvalue;
	return (comp.sign() != 0) ? ASTrue : ASFalse;
    }
    // otherwise, doing a string comparison
    s1 = stripfront(striprear(s1));
    s2 = stripfront(striprear(s2));
    if (s1.Length() > s2.Length()) {
	s2 = padonright(s2, s1.Length(), ' ');
    } else {
	s1 = padonright(s1, s2.Length(), ' ');
    }
//    returnvalue = (int)(s1 != s2);
//    return returnvalue;
    return (s1 != s2) ? ASTrue : ASFalse;
}

//

LessThanExpression::LessThanExpression(Expression* e1, Expression* e2) : BinaryExpression(e1, e2) {}

LessThanExpression::~LessThanExpression() {}

const ArgString& LessThanExpression::Evaluate() {
    static ArgString s1;
    static ArgString s2;
    static ArgString returnvalue;

    s1 = exp1->Evaluate();
    if (!s1.IsValid()) BADRETURN;
    s2 = exp2->Evaluate();
    if (!s2.IsValid()) BADRETURN;
    // should do some trapping here!
    if (s1.IsANumber() && s2.IsANumber()) {
	// need to do rexx comparison here
//	RexxNumber& n1 = s1.Number();
//	static RexxNumber n2;
//	n2 = s2.Number();
//	n2.negate();
	static RexxNumber comp;
//	add(n1, n2, GetDigits() - GetFuzz(), comp);
	subtract(s1.Number(), s2.Number(), GetDigits() - GetFuzz(), comp);
	
//	returnvalue = (int)comp.sign() < 0;
//	return returnvalue;
	return (comp.sign() < 0) ? ASTrue : ASFalse;
    }
    // otherwise, doing a string comparison
    s1 = stripfront(striprear(s1));
    s2 = stripfront(striprear(s2));
    if (s1.Length() > s2.Length()) {
	s2 = padonright(s2, s1.Length(), ' ');
    } else {
	s1 = padonright(s1, s2.Length(), ' ');
    }
//    returnvalue = (int)(s1 < s2);
//    return returnvalue;
    return (s1 < s2) ? ASTrue : ASFalse;
}

//

GreaterThanExpression::GreaterThanExpression(Expression* e1, Expression* e2) : BinaryExpression(e1, e2) {}

GreaterThanExpression::~GreaterThanExpression() {}

const ArgString& GreaterThanExpression::Evaluate() {
    static ArgString s1;
    static ArgString s2;
    static ArgString returnvalue;

    s1 = exp1->Evaluate();
    if (!s1.IsValid()) BADRETURN;
    s2 = exp2->Evaluate();
    if (!s2.IsValid()) BADRETURN;

    if (s1.IsANumber() && s2.IsANumber()) {
	// need to do rexx comparison here
//	RexxNumber& n1 = s1.Number();
//	static RexxNumber n2;
//	n2 = s2.Number();
//	n2.negate();
	static RexxNumber comp;
//	add(n1, n2, GetDigits() - GetFuzz(), comp);
	subtract(s1.Number(), s2.Number(), GetDigits() - GetFuzz(), comp);
//	returnvalue = (int)comp.sign() > 0;
//	return returnvalue;
	return (comp.sign() > 0) ? ASTrue : ASFalse;
    }
    // otherwise, doing a string comparison
    s1 = stripfront(striprear(s1));
    s2 = stripfront(striprear(s2));
    if (s1.Length() > s2.Length()) {
	s2 = padonright(s2, s1.Length(), ' ');
    } else {
	s1 = padonright(s1, s2.Length(), ' ');
    }
//    returnvalue = (int)(s1 > s2);
//    return returnvalue;
    return (s1 > s2) ? ASTrue : ASFalse;
}

//

GreaterEqualExpression::GreaterEqualExpression(Expression* e1, Expression* e2) : BinaryExpression(e1, e2) {}

GreaterEqualExpression::~GreaterEqualExpression() {}

const ArgString& GreaterEqualExpression::Evaluate() {
    static ArgString s1;
    static ArgString s2;
    static ArgString returnvalue;

    s1 = exp1->Evaluate();
    if (!s1.IsValid()) BADRETURN;
    s2 = exp2->Evaluate();
    if (!s2.IsValid()) BADRETURN;

    // should do some trapping here!
    if (s1.IsANumber() && s2.IsANumber()) {
	// need to do rexx comparison here
//	RexxNumber& n1 = s1.Number();
//	static RexxNumber n2;
//	n2 = s2.Number();
//	n2.negate();
	static RexxNumber comp;
//	add(n1, n2, GetDigits() - GetFuzz(), comp);
	subtract(s1.Number(), s2.Number(), GetDigits() - GetFuzz(), comp);	
//	returnvalue = (int)comp.sign() >= 0;
//	return returnvalue;
	return (comp.sign() >= 0) ? ASTrue : ASFalse;
    }

    // otherwise, doing a string comparison
    s1 = stripfront(striprear(s1));
    s2 = stripfront(striprear(s2));
    if (s1.Length() > s2.Length()) {
	s2 = padonright(s2, s1.Length(), ' ');
    } else {
	s1 = padonright(s1, s2.Length(), ' ');
    }
//    returnvalue = (int)(s1 >= s2);
//    return returnvalue;
    return (s1 >= s2) ? ASTrue : ASFalse;
}

//

LessEqualExpression::LessEqualExpression(Expression* e1, Expression* e2) : BinaryExpression(e1, e2) {}

LessEqualExpression::~LessEqualExpression() {}

const ArgString& LessEqualExpression::Evaluate() {
    static ArgString s1;
    static ArgString s2;
    static ArgString returnvalue;

    s1 = exp1->Evaluate();
    if (!s1.IsValid()) BADRETURN;
    s2 = exp2->Evaluate();
    if (!s2.IsValid()) BADRETURN;

    if (s1.IsANumber() && s2.IsANumber()) {
	// need to do rexx comparison here
//	RexxNumber& n1 = s1.Number();
//	static RexxNumber n2;
//	n2 = s2.Number();
//	n2.negate();
	static RexxNumber comp;
//	add(n1, n2, GetDigits() - GetFuzz(), comp);
	subtract(s1.Number(), s2.Number(), GetDigits() - GetFuzz(), comp);
//	returnvalue = (int)comp.sign() <= 0;
//	return returnvalue;
	return (comp.sign() <= 0) ? ASTrue : ASFalse;
    }

    // otherwise, doing a string comparison
    s1 = stripfront(striprear(s1));
    s2 = stripfront(striprear(s2));
    if (s1.Length() > s2.Length()) {
	s2 = padonright(s2, s1.Length(), ' ');
    } else {
	s1 = padonright(s1, s2.Length(), ' ');
    }

//    returnvalue = (int)(s1 <= s2);
//    return returnvalue;
    return (s1 <= s2) ? ASTrue : ASFalse;
}

//

ConcatExpression::ConcatExpression(Expression* e1, Expression* e2) : BinaryExpression(e1, e2) {}

ConcatExpression::~ConcatExpression() {}

const ArgString& ConcatExpression::Evaluate() {
    static ArgString s1;
    static ArgString returnvalue;

    s1 = exp1->Evaluate();
    if (!s1.IsValid()) BADRETURN;
    ArgString& s2 = exp2->Evaluate();
    if (!s2.IsValid()) BADRETURN;

    s1 += s2;

    returnvalue = s1;
    return returnvalue;
}

//

BlankExpression::BlankExpression(Expression* e1, Expression* e2) : BinaryExpression(e1, e2) {}

BlankExpression::~BlankExpression() {}

const ArgString& BlankExpression::Evaluate() {
    static ArgString s1;
    static ArgString returnvalue;

    s1 = exp1->Evaluate();
    if (!s1.IsValid()) BADRETURN;
    ArgString& s2 = exp2->Evaluate();
    if (!s2.IsValid()) BADRETURN;

    s1 += ' ';
    s1 += s2;
    returnvalue = s1;
    return returnvalue;
}

//

AddExpression::AddExpression(Expression* e1, Expression* e2) : BinaryExpression(e1, e2) {}

AddExpression::~AddExpression() {}

const ArgString& AddExpression::Evaluate() {
    static ArgString s1;
    static ArgString returnvalue;

    s1 = exp1->Evaluate();
    if (!s1.IsValid()) BADRETURN;
    ArgString& s2 = exp2->Evaluate();
    if (!s2.IsValid()) BADRETURN;

    if (!s1.IsANumber() || !s2.IsANumber()) {
	SetError(EBadArithmeticConversion);
	BADRETURN;
    }

    static RexxNumber temp;
    add(s1.Number(), s2.Number(), GetDigits(), temp);
    returnvalue = Form(temp, GetDigits());
    return returnvalue;
}

//

SubtractExpression::SubtractExpression(Expression* e1, Expression* e2) : BinaryExpression(e1, e2) {}

SubtractExpression::~SubtractExpression() {}

const ArgString& SubtractExpression::Evaluate() {
    static ArgString s1;
    static ArgString returnvalue;

    s1 = exp1->Evaluate();
    if (!s1.IsValid()) BADRETURN;
    ArgString& s2 = exp2->Evaluate();
    if (!s2.IsValid()) BADRETURN;

    if (!s1.IsANumber() || !s2.IsANumber()) {
	SetError(EBadArithmeticConversion);
	BADRETURN;
    }

//    RexxNumber& n1 = s1.Number();
//    static RexxNumber n2;
//    n2 = s2.Number();
//    n2.negate();
    static RexxNumber temp;
//    add(n1, n2, GetDigits(), temp);
    subtract(s1.Number(), s2.Number(), GetDigits(), temp);
    
    returnvalue =  Form(temp, GetDigits());
    return returnvalue;
}

//

RemainderExpression::RemainderExpression(Expression* e1, Expression* e2) : BinaryExpression(e1, e2) {}

RemainderExpression::~RemainderExpression() {}

const ArgString& RemainderExpression::Evaluate() {
    static ArgString s1;
    static ArgString returnvalue;

    s1 = exp1->Evaluate();
    if (!s1.IsValid()) BADRETURN;
    ArgString& s2 = exp2->Evaluate();
    if (!s2.IsValid()) BADRETURN;

    if (!s1.IsANumber() || !s2.IsANumber()) {
	SetError(EBadArithmeticConversion);
	BADRETURN;
    }

    static RexxNumber q, r;
    int success = integerdivide(s1.Number(), s2.Number(), GetDigits(), q, r);
    if (!success) {
	SetError(EArithmeticOverflowUnderflow);
	BADRETURN;
    }

    returnvalue = Form(r, GetDigits());
    return returnvalue;
}

//

IntDivideExpression::IntDivideExpression(Expression* e1, Expression* e2) : BinaryExpression(e1, e2) {}

IntDivideExpression::~IntDivideExpression() {}

const ArgString& IntDivideExpression::Evaluate() {
    static ArgString s1;
    static ArgString returnvalue;

    s1 = exp1->Evaluate();
    if (!s1.IsValid()) BADRETURN;
    ArgString& s2 = exp2->Evaluate();
    if (!s2.IsValid()) BADRETURN;

    if (!s1.IsANumber() || !s2.IsANumber()) {
	SetError(EBadArithmeticConversion);
	BADRETURN;
    }

    static RexxNumber q, r;
    int success = integerdivide(s1.Number(), s2.Number(), GetDigits(), q, r);
    if (!success) {
	SetError(EArithmeticOverflowUnderflow);
	BADRETURN;
    }

    returnvalue = Form(q, GetDigits());
    return returnvalue;
}

//

DivideExpression::DivideExpression(Expression* e1, Expression* e2) : BinaryExpression(e1, e2) {}

DivideExpression::~DivideExpression() {}

const ArgString& DivideExpression::Evaluate() {
    static ArgString s1;
    static ArgString returnvalue;

    s1 = exp1->Evaluate();
    if (!s1.IsValid()) BADRETURN;
    ArgString& s2 = exp2->Evaluate();
    if (!s2.IsValid()) BADRETURN;

    if (!s1.IsANumber() || !s2.IsANumber()) {
	SetError(EBadArithmeticConversion);
	BADRETURN;
    }

    static RexxNumber temp;
    divide(s1.Number(), s2.Number(), GetDigits(), temp);
    returnvalue = Form(temp, GetDigits());
    return returnvalue;
}

//

MultiplyExpression::MultiplyExpression(Expression* e1, Expression* e2) : BinaryExpression(e1, e2) {}

MultiplyExpression::~MultiplyExpression() {}

const ArgString& MultiplyExpression::Evaluate() {
    static ArgString s1;
    static ArgString returnvalue;

    s1 = exp1->Evaluate();
    if (!s1.IsValid()) BADRETURN;
    ArgString& s2 = exp2->Evaluate();
    if (!s2.IsValid()) BADRETURN;

    if (!s1.IsANumber() || !s2.IsANumber()) {
	SetError(EBadArithmeticConversion);
	BADRETURN;
    }

    static RexxNumber temp;
    multiply(s1.Number(), s2.Number(), GetDigits(), temp);
    returnvalue = Form(temp, GetDigits());
    return returnvalue;
}

//

ExponentExpression::ExponentExpression(Expression* e1, Expression* e2) : BinaryExpression(e1, e2) {}

ExponentExpression::~ExponentExpression() {}

const ArgString& ExponentExpression::Evaluate() {
    static ArgString s1;
    static ArgString returnvalue;

    s1 = exp1->Evaluate();
    if (!s1.IsValid()) BADRETURN;
    ArgString& s2 = exp2->Evaluate();
    if (!s2.IsValid()) BADRETURN;

    if (!s1.IsANumber() || !s2.IsANumber() || !s2.IsAWhole()) {
	SetError(EBadArithmeticConversion);
	BADRETURN;
    }

    static RexxNumber temp;
    power(s1.Number(), s2.Number(), GetDigits(), temp);
    returnvalue = Form(temp, GetDigits());
    return returnvalue;
}

//

AndExpression::AndExpression(Expression* e1, Expression* e2) : BinaryExpression(e1, e2) {}

AndExpression::~AndExpression() {
}

const ArgString& AndExpression::Evaluate() {
    static ArgString s1;
    static ArgString returnvalue;

    s1 = exp1->Evaluate();
    if (!s1.IsValid()) BADRETURN;
    ArgString& s2 = exp2->Evaluate();
    if (!s2.IsValid()) BADRETURN;

    if (!s1.IsABoolean() || !s2.IsABoolean()) {
	SetError(ELogicalValueNotZeroOrOne);
	BADRETURN;
    }
    int i1 = s1.Boolean();
    int i2 = s2.Boolean();

//    returnvalue = (int)(i1 && i2);
//    return returnvalue;
    return (i1 && i2) ? ASTrue : ASFalse;
}

//

OrExpression::OrExpression(Expression* e1, Expression* e2) : BinaryExpression(e1, e2) {}

OrExpression::~OrExpression() {
}

const ArgString& OrExpression::Evaluate() {
    static ArgString s1;
    static ArgString returnvalue;

    s1 = exp1->Evaluate();
    if (!s1.IsValid()) BADRETURN;
    ArgString& s2 = exp2->Evaluate();

    if (!s2.IsValid()) BADRETURN;

    if (!s1.IsABoolean() || !s2.IsABoolean()) {
	SetError(ELogicalValueNotZeroOrOne);
	BADRETURN;
    }

    int i1 = s1.Boolean();
    int i2 = s2.Boolean();

//    returnvalue = (int)(i1 || i2);
//    return returnvalue;
    return (i1 || i2) ? ASTrue : ASFalse;
}

//

XorExpression::XorExpression(Expression* e1, Expression* e2) : BinaryExpression(e1, e2) {}

XorExpression::~XorExpression() {
}

const ArgString& XorExpression::Evaluate() {
    static ArgString s1;
    static ArgString returnvalue;

    s1 = exp1->Evaluate();
    if (!s1.IsValid()) BADRETURN;
    ArgString& s2 = exp2->Evaluate();

    if (!s2.IsValid()) BADRETURN;

    if (!s1.IsABoolean() || !s2.IsABoolean()) {
	SetError(ELogicalValueNotZeroOrOne);
	BADRETURN;
    }
    int i1 = s1.Boolean();
    int i2 = s2.Boolean();
//    returnvalue = (int)((i1 && !i2) || (!i1 && i2));
//    return returnvalue;
    return ((i1 && !i2) || (!i1 && i2)) ? ASTrue : ASFalse;
}

//

NotExpression::NotExpression(Expression* e) : UnaryExpression(e) {}

NotExpression::~NotExpression() {
}

const ArgString& NotExpression::Evaluate() {
    static ArgString returnvalue;

    ArgString& a = exp->Evaluate();
    if (!a.IsValid()) BADRETURN;

    if (!a.IsABoolean()) {
	SetError(ELogicalValueNotZeroOrOne);
	BADRETURN;
    }
//    returnvalue = (int)a.Boolean();
//    return returnvalue;
    return (a.Boolean()) ? ASTrue : ASFalse;
}

