#ifndef _EXPRESSION_H
#pragma once
#define _EXPRESSION_H

#include <String.h>

#include "ArgStr.h"
#include "InternalData.h"
#include "Variable.h"
class ExpressionpSLList;

class Expression {
  public:
    Expression();
    virtual ~Expression();
    virtual const ArgString& Evaluate();
};

class ErrorExpression : public Expression {
  private:
    Errors error;
  public:
    ErrorExpression(Errors e);
    virtual ~ErrorExpression();
    virtual const ArgString& Evaluate();
};

enum FunctionType {UnresolvedFunction, InternalFunction, BuiltinFunction, ExternalFunction};
typedef void (*Functionp)();

class FunctionExpression : public Expression {
  private:
    String label;
    FunctionType functiontype; // indicates whether we have an internal, builtin, or possibly external function
    Functionp Builtin; // if functiontype == BuiltinFunction, this is valid
    int Internal; // if functiontype == InternalFunction, this is valid
    ExpressionpSLList* arguments;
    int IsCall;
    int ShouldSkipInternalSearch;
    const ArgString& ActualEvaluate();
  public:
    FunctionExpression(const ArgString&, ExpressionpSLList*, int = 0);
    virtual ~FunctionExpression();
    void SetCall(); // set if this is being evaluated as a call
    virtual const ArgString& Evaluate();
};

class LiteralExpression : public Expression {
  private:
    ArgString string;
  public:
    LiteralExpression(const ArgString&);
    virtual ~LiteralExpression();
    virtual const ArgString& Evaluate();
};

class SimpleSymExpression : public Expression {
  private:
//    String symbol;
    SimpleVariable vsymbol;
  public:
    SimpleSymExpression(const ArgString&);
    virtual ~SimpleSymExpression();
    virtual const ArgString& Evaluate();
};

class StemSymExpression : public Expression {
  private:
//    String symbol;
    StemVariable vsymbol;
  public:
    StemSymExpression(const ArgString&);
    virtual ~StemSymExpression();
    virtual const ArgString& Evaluate();
};

class CompoundSymExpression : public Expression {
  private:
//    String symbol;
    CompoundVariable vsymbol;
  public:
    CompoundSymExpression(const ArgString&);
    virtual ~CompoundSymExpression();
    virtual const ArgString& Evaluate();
};

class UnaryExpression : public Expression {
  protected:
    Expression* exp;
  public:
    UnaryExpression(Expression*);
    virtual ~UnaryExpression();
};

class PositiveExpression : public UnaryExpression {
  private:
  public:
    PositiveExpression(Expression*);
    virtual ~PositiveExpression();
    virtual const ArgString& Evaluate();
};

class NegativeExpression : public UnaryExpression {
  private:
  public:
    NegativeExpression(Expression*);
    virtual ~NegativeExpression();
    virtual const ArgString& Evaluate();
};

class BinaryExpression : public Expression {
  protected:
    Expression* exp1;
    Expression* exp2;
  public:
    BinaryExpression(Expression*, Expression*);
    virtual ~BinaryExpression();
};

class EqualExpression : public BinaryExpression {
  public:
    EqualExpression(Expression*, Expression*);
    virtual ~EqualExpression();
    virtual const ArgString& Evaluate();
};

class NotEqualExpression : public BinaryExpression {
  public:
    NotEqualExpression(Expression*, Expression*);
    virtual ~NotEqualExpression();
    virtual const ArgString& Evaluate();
};

class LessEqualExpression : public BinaryExpression {
  public:
    LessEqualExpression(Expression*, Expression*);
    virtual ~LessEqualExpression();
    virtual const ArgString& Evaluate();
};

class GreaterEqualExpression : public BinaryExpression {
  public:
    GreaterEqualExpression(Expression*, Expression*);
    virtual ~GreaterEqualExpression();
    virtual const ArgString& Evaluate();
};

class LessThanExpression : public BinaryExpression {
  public:
    LessThanExpression(Expression*, Expression*);
    virtual ~LessThanExpression();
    virtual const ArgString& Evaluate();
};

class GreaterThanExpression : public BinaryExpression {
  public:
    GreaterThanExpression(Expression*, Expression*);
    virtual ~GreaterThanExpression();
    virtual const ArgString& Evaluate();
};

class ExactEqualExpression : public BinaryExpression {
  public:
    ExactEqualExpression(Expression*, Expression*);
    virtual ~ExactEqualExpression();
    virtual const ArgString& Evaluate();
};

class ExactNotEqualExpression : public BinaryExpression {
  public:
    ExactNotEqualExpression(Expression*, Expression*);
    virtual ~ExactNotEqualExpression();
    virtual const ArgString& Evaluate();
};

class ExactLessEqualExpression : public BinaryExpression {
  public:
    ExactLessEqualExpression(Expression*, Expression*);
    virtual ~ExactLessEqualExpression();
    virtual const ArgString& Evaluate();
};

class ExactGreaterEqualExpression : public BinaryExpression {
  public:
    ExactGreaterEqualExpression(Expression*, Expression*);
    virtual ~ExactGreaterEqualExpression();
    virtual const ArgString& Evaluate();
};

class ExactLessThanExpression : public BinaryExpression {
  public:
    ExactLessThanExpression(Expression*, Expression*);
    virtual ~ExactLessThanExpression();
    virtual const ArgString& Evaluate();
};

class ExactGreaterThanExpression : public BinaryExpression {
  public:
    ExactGreaterThanExpression(Expression*, Expression*);
    virtual ~ExactGreaterThanExpression();
    virtual const ArgString& Evaluate();
};

class ConcatExpression : public BinaryExpression {
  public:
    ConcatExpression(Expression*, Expression*);
    virtual ~ConcatExpression();
    virtual const ArgString& Evaluate();
};

class BlankExpression : public BinaryExpression {
  public:
    BlankExpression(Expression*, Expression*);
    virtual ~BlankExpression();
    virtual const ArgString& Evaluate();
};

class AddExpression : public BinaryExpression {
  public:
    AddExpression(Expression*, Expression*);
    virtual ~AddExpression();
    virtual const ArgString& Evaluate();
};

class SubtractExpression : public BinaryExpression {
  public:
    SubtractExpression(Expression*, Expression*);
    virtual ~SubtractExpression();
    virtual const ArgString& Evaluate();
};

class RemainderExpression : public BinaryExpression {
  public:
    RemainderExpression(Expression*, Expression*);
    virtual ~RemainderExpression();
    virtual const ArgString& Evaluate();
};

class IntDivideExpression : public BinaryExpression {
  public:
    IntDivideExpression(Expression*, Expression*);
    virtual ~IntDivideExpression();
    virtual const ArgString& Evaluate();
};

class DivideExpression : public BinaryExpression {
  public:
    DivideExpression(Expression*, Expression*);
    virtual ~DivideExpression();
    virtual const ArgString& Evaluate();
};

class MultiplyExpression : public BinaryExpression {
  public:
    MultiplyExpression(Expression*, Expression*);
    virtual ~MultiplyExpression();
    virtual const ArgString& Evaluate();
};

class ExponentExpression : public BinaryExpression {
  public:
    ExponentExpression(Expression*, Expression*);
    virtual ~ExponentExpression();
    virtual const ArgString& Evaluate();
};

class AndExpression : public BinaryExpression {
  public:
    AndExpression(Expression*, Expression*);
    virtual ~AndExpression();
    virtual const ArgString& Evaluate();
};

class OrExpression : public BinaryExpression {
  public:
    OrExpression(Expression*, Expression*);
    virtual ~OrExpression();
    virtual const ArgString& Evaluate();
};

class XorExpression : public BinaryExpression {
  public:
    XorExpression(Expression*, Expression*);
    virtual ~XorExpression();
    virtual const ArgString& Evaluate();
};

class NotExpression : public UnaryExpression {
  public:
    NotExpression(Expression*);
    virtual ~NotExpression();
    virtual const ArgString& Evaluate();
};

/*
class ExactInequalExpression : public BinaryExpression {
  public:
    ExactInequalExpression(Expression*, Expression*);
    virtual ~ExactInequalExpression();
    virtual const ArgString& Evaluate();
};

class InequalExpression : public BinaryExpression {
  public:
    InequalExpression(Expression*, Expression*);
    virtual ~InequalExpression();
    virtual const ArgString& Evaluate();
};
*/

/*
class GreaterExpression : public BinaryExpression {
  public:
    GreaterExpression(Expression*, Expression*);
    virtual ~GreaterExpression();
    virtual const ArgString& Evaluate();
};
*/

/*
class LessExpression : public BinaryExpression {
  public:
    LessExpression(Expression*, Expression*);
    virtual ~LessExpression();
    virtual const ArgString& Evaluate();
};
*/

#endif






