// This may look like C code, but it is really -*- C++ -*-
/* 
Copyright (C) 1988 Free Software Foundation
    written by Doug Lea (dl@rocky.oswego.edu)

This file is part of GNU CC.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY.  No author or distributor
accepts responsibility to anyone for the consequences of using it
or for whether it serves any particular purpose or works at all,
unless he says so in writing.  Refer to the GNU CC General Public
License for full details.

Everyone is granted permission to copy, modify and redistribute
GNU CC, but only under the conditions described in the
GNU CC General Public License.   A copy of this license is
supposed to have been given to you along with GNU CC so you
can know your rights and responsibilities.  It should be in a
file named COPYING.  Among other things, the copyright notice
and this notice must be preserved on all copies.  
*/

#ifdef __GNUG__
#pragma implementation
#endif
#include <values.h>
#include <stream.h>
#include "Expressionp.SLList.h"


void ExpressionpSLList::error(const char* msg)
{
  (*lib_error_handler)("SLList", msg);
}

int ExpressionpSLList::length()
{
  int l = 0;
  ExpressionpSLListNode* t = last;
  if (t != 0) do { ++l; t = t->tl; } while (t != last);
  return l;
}

ExpressionpSLList::ExpressionpSLList(ExpressionpSLList& a)
{
  if (a.last == 0)
    last = 0;
  else
  {
    ExpressionpSLListNode* p = a.last->tl;
    ExpressionpSLListNode* h = new ExpressionpSLListNode(p->hd);
    last = h;
    for (;;)
    {
      if (p == a.last)
      {
        last->tl = h;
        return;
      }
      p = p->tl;
      ExpressionpSLListNode* n = new ExpressionpSLListNode(p->hd);
      last->tl = n;
      last = n;
    }
  }
}

ExpressionpSLList& ExpressionpSLList::operator = (ExpressionpSLList& a)
{
  if (last != a.last)
  {
    clear();
    if (a.last != 0)
    {
      ExpressionpSLListNode* p = a.last->tl;
      ExpressionpSLListNode* h = new ExpressionpSLListNode(p->hd);
      last = h;
      for (;;)
      {
        if (p == a.last)
        {
          last->tl = h;
          break;
        }
        p = p->tl;
        ExpressionpSLListNode* n = new ExpressionpSLListNode(p->hd);
        last->tl = n;
        last = n;
      }
    }
  }
  return *this;
}

void ExpressionpSLList::clear()
{
  if (last == 0)
    return;

  ExpressionpSLListNode* p = last->tl;
  last->tl = 0;
  last = 0;

  while (p != 0)
  {
    ExpressionpSLListNode* nxt = p->tl;
    delete(p);
    p = nxt;
  }
}


Pix ExpressionpSLList::prepend(Expressionp  item)
{
  ExpressionpSLListNode* t = new ExpressionpSLListNode(item);
  if (last == 0)
    t->tl = last = t;
  else
  {
    t->tl = last->tl;
    last->tl = t;
  }
  return Pix(t);
}


Pix ExpressionpSLList::prepend(ExpressionpSLListNode* t)
{
  if (t == 0) return 0;
  if (last == 0)
    t->tl = last = t;
  else
  {
    t->tl = last->tl;
    last->tl = t;
  }
  return Pix(t);
}


Pix ExpressionpSLList::append(Expressionp  item)
{
  ExpressionpSLListNode* t = new ExpressionpSLListNode(item);
  if (last == 0)
    t->tl = last = t;
  else
  {
    t->tl = last->tl;
    last->tl = t;
    last = t;
  }
  return Pix(t);
}

Pix ExpressionpSLList::append(ExpressionpSLListNode* t)
{
  if (t == 0) return 0;
  if (last == 0)
    t->tl = last = t;
  else
  {
    t->tl = last->tl;
    last->tl = t;
    last = t;
  }
  return Pix(t);
}

void ExpressionpSLList::join(ExpressionpSLList& b)
{
  ExpressionpSLListNode* t = b.last;
  b.last = 0;
  if (last == 0)
    last = t;
  else if (t != 0)
  {
    ExpressionpSLListNode* f = last->tl;
    last->tl = t->tl;
    t->tl = f;
    last = t;
  }
}

Pix ExpressionpSLList::ins_after(Pix p, Expressionp  item)
{
  ExpressionpSLListNode* u = (ExpressionpSLListNode*)p;
  ExpressionpSLListNode* t = new ExpressionpSLListNode(item);
  if (last == 0)
    t->tl = last = t;
  else if (u == 0) // ins_after 0 means prepend
  {
    t->tl = last->tl;
    last->tl = t;
  }
  else
  {
    t->tl = u->tl;
    u->tl = t;
    if (u == last) 
      last = t;
  }
  return Pix(t);
}


void ExpressionpSLList::del_after(Pix p)
{
  ExpressionpSLListNode* u = (ExpressionpSLListNode*)p;
  if (last == 0 || u == last) error("cannot del_after last");
  if (u == 0) u = last; // del_after 0 means delete first
  ExpressionpSLListNode* t = u->tl;
  if (u == t)
    last = 0;
  else
  {
    u->tl = t->tl;
    if (last == t)
      last = u;
  }
  delete t;
}

int ExpressionpSLList::owns(Pix p)
{
  ExpressionpSLListNode* t = last;
  if (t != 0 && p != 0)
  {
    do
    {
      if (Pix(t) == p) return 1;
      t = t->tl;
    } while (t != last);
  }
  return 0;
}

Expressionp ExpressionpSLList::remove_front()
{
  if (last == 0) error("remove_front of empty list");
  ExpressionpSLListNode* t = last->tl;
  Expressionp res = t->hd;
  if (t == last)
    last = 0;
  else
    last->tl = t->tl;
  delete t;
  return res;
}

int ExpressionpSLList::remove_front(Expressionp& x)
{
  if (last == 0)
    return 0;
  else
  {
    ExpressionpSLListNode* t = last->tl;
    x = t->hd;
    if (t == last)
      last = 0;
    else
      last->tl = t->tl;
    delete t;
    return 1;
  }
}


void ExpressionpSLList::del_front()
{
  if (last == 0) error("del_front of empty list");
  ExpressionpSLListNode* t = last->tl;
  if (t == last)
    last = 0;
  else
    last->tl = t->tl;
  delete t;
}

int ExpressionpSLList::OK()
{
  int v = 1;
  if (last != 0)
  {
    ExpressionpSLListNode* t = last;
    long count = MAXLONG;      // Lots of chances to find last!
    do
    {
      count--;
      t = t->tl;
    } while (count > 0 && t != last);
    v &= count > 0;
  }
  if (!v) error("invariant failure");
  return v;
}
