// This may look like C code, but it is really -*- C++ -*-
/* 
Copyright (C) 1988 Free Software Foundation
    written by Doug Lea (dl@rocky.oswego.edu)

This file is part of GNU CC.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY.  No author or distributor
accepts responsibility to anyone for the consequences of using it
or for whether it serves any particular purpose or works at all,
unless he says so in writing.  Refer to the GNU CC General Public
License for full details.

Everyone is granted permission to copy, modify and redistribute
GNU CC, but only under the conditions described in the
GNU CC General Public License.   A copy of this license is
supposed to have been given to you along with GNU CC so you
can know your rights and responsibilities.  It should be in a
file named COPYING.  Among other things, the copyright notice
and this notice must be preserved on all copies.  
*/


#ifndef _ExpressionpSLList_h
#ifdef __GNUG__
#pragma once
#pragma interface
#endif
#define _ExpressionpSLList_h 1

#include <Pix.h>
#include "Expressionp.defs.h"

#ifndef _ExpressionpSLListNode_h
#define _ExpressionpSLListNode_h 1

struct ExpressionpSLListNode
{
  ExpressionpSLListNode*         tl;
  Expressionp                    hd;
                         ExpressionpSLListNode();
                         ExpressionpSLListNode(Expressionp  h, ExpressionpSLListNode* t = 0);
                         ~ExpressionpSLListNode();
};

inline ExpressionpSLListNode::ExpressionpSLListNode() {}

inline ExpressionpSLListNode::ExpressionpSLListNode(Expressionp  h, ExpressionpSLListNode* t) :hd(h), tl(t) {}

inline ExpressionpSLListNode::~ExpressionpSLListNode() {}

typedef ExpressionpSLListNode* ExpressionpSLListNodePtr;

#endif


class ExpressionpSLList
{
protected:
  ExpressionpSLListNode*        last;

public:
                        ExpressionpSLList();
                        ExpressionpSLList(ExpressionpSLList& a);
                        ~ExpressionpSLList();

  ExpressionpSLList&            operator = (ExpressionpSLList& a);

  int                   empty();
  int                   length();

  void                  clear();

  Pix                   prepend(Expressionp  item);
  Pix                   append(Expressionp  item);

  void                  join(ExpressionpSLList&);

  Pix                   prepend(ExpressionpSLListNode*);
  Pix                   append(ExpressionpSLListNode*);

  Expressionp&                  operator () (Pix p);
  Pix                   first();
  void                  next(Pix& p);
  int                   owns(Pix p);
  Pix                   ins_after(Pix p, Expressionp  item);
  void                  del_after(Pix p);

  Expressionp&                  front();
  Expressionp&                  rear();
  Expressionp                   remove_front();
  int                   remove_front(Expressionp& x);
  void                  del_front();

  void                  error(const char* msg);
  int                   OK();
};


inline ExpressionpSLList::~ExpressionpSLList()
{
  clear();
}

inline ExpressionpSLList::ExpressionpSLList()
{
  last = 0;
}

inline int ExpressionpSLList::empty()
{
  return last == 0;
}


inline Pix ExpressionpSLList::first()
{
  return (last == 0)? 0 : Pix(last->tl);
}

inline void ExpressionpSLList::next(Pix& p)
{
  p = (p == 0 || p == last)? 0 : Pix(((ExpressionpSLListNode*)(p))->tl);
}

inline Expressionp& ExpressionpSLList::operator () (Pix p)
{
  if (p == 0) error("null Pix");
  return ((ExpressionpSLListNode*)(p))->hd;
}

inline Expressionp& ExpressionpSLList::front()
{
  if (last == 0) error("front: empty list");
  return last->tl->hd;
}

inline Expressionp& ExpressionpSLList::rear()
{
  if (last == 0) error("rear: empty list");
  return last->hd;
}


#endif
