// This may look like C code, but it is really -*- C++ -*-
/* 
Copyright (C) 1988 Free Software Foundation
    written by Doug Lea (dl@rocky.oswego.edu)

This file is part of GNU CC.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY.  No author or distributor
accepts responsibility to anyone for the consequences of using it
or for whether it serves any particular purpose or works at all,
unless he says so in writing.  Refer to the GNU CC General Public
License for full details.

Everyone is granted permission to copy, modify and redistribute
GNU CC, but only under the conditions described in the
GNU CC General Public License.   A copy of this license is
supposed to have been given to you along with GNU CC so you
can know your rights and responsibilities.  It should be in a
file named COPYING.  Among other things, the copyright notice
and this notice must be preserved on all copies.  
*/

#ifdef __GNUG__
#pragma implementation
#endif
#include <values.h>
#include <stream.h>
#include "InternalDatap.SLList.h"


void InternalDatapSLList::error(const char* msg)
{
  (*lib_error_handler)("SLList", msg);
}

int InternalDatapSLList::length()
{
  int l = 0;
  InternalDatapSLListNode* t = last;
  if (t != 0) do { ++l; t = t->tl; } while (t != last);
  return l;
}

InternalDatapSLList::InternalDatapSLList(InternalDatapSLList& a)
{
  if (a.last == 0)
    last = 0;
  else
  {
    InternalDatapSLListNode* p = a.last->tl;
    InternalDatapSLListNode* h = new InternalDatapSLListNode(p->hd);
    last = h;
    for (;;)
    {
      if (p == a.last)
      {
        last->tl = h;
        return;
      }
      p = p->tl;
      InternalDatapSLListNode* n = new InternalDatapSLListNode(p->hd);
      last->tl = n;
      last = n;
    }
  }
}

InternalDatapSLList& InternalDatapSLList::operator = (InternalDatapSLList& a)
{
  if (last != a.last)
  {
    clear();
    if (a.last != 0)
    {
      InternalDatapSLListNode* p = a.last->tl;
      InternalDatapSLListNode* h = new InternalDatapSLListNode(p->hd);
      last = h;
      for (;;)
      {
        if (p == a.last)
        {
          last->tl = h;
          break;
        }
        p = p->tl;
        InternalDatapSLListNode* n = new InternalDatapSLListNode(p->hd);
        last->tl = n;
        last = n;
      }
    }
  }
  return *this;
}

void InternalDatapSLList::clear()
{
  if (last == 0)
    return;

  InternalDatapSLListNode* p = last->tl;
  last->tl = 0;
  last = 0;

  while (p != 0)
  {
    InternalDatapSLListNode* nxt = p->tl;
    delete(p);
    p = nxt;
  }
}


Pix InternalDatapSLList::prepend(InternalDatap  item)
{
  InternalDatapSLListNode* t = new InternalDatapSLListNode(item);
  if (last == 0)
    t->tl = last = t;
  else
  {
    t->tl = last->tl;
    last->tl = t;
  }
  return Pix(t);
}


Pix InternalDatapSLList::prepend(InternalDatapSLListNode* t)
{
  if (t == 0) return 0;
  if (last == 0)
    t->tl = last = t;
  else
  {
    t->tl = last->tl;
    last->tl = t;
  }
  return Pix(t);
}


Pix InternalDatapSLList::append(InternalDatap  item)
{
  InternalDatapSLListNode* t = new InternalDatapSLListNode(item);
  if (last == 0)
    t->tl = last = t;
  else
  {
    t->tl = last->tl;
    last->tl = t;
    last = t;
  }
  return Pix(t);
}

Pix InternalDatapSLList::append(InternalDatapSLListNode* t)
{
  if (t == 0) return 0;
  if (last == 0)
    t->tl = last = t;
  else
  {
    t->tl = last->tl;
    last->tl = t;
    last = t;
  }
  return Pix(t);
}

void InternalDatapSLList::join(InternalDatapSLList& b)
{
  InternalDatapSLListNode* t = b.last;
  b.last = 0;
  if (last == 0)
    last = t;
  else if (t != 0)
  {
    InternalDatapSLListNode* f = last->tl;
    last->tl = t->tl;
    t->tl = f;
    last = t;
  }
}

Pix InternalDatapSLList::ins_after(Pix p, InternalDatap  item)
{
  InternalDatapSLListNode* u = (InternalDatapSLListNode*)p;
  InternalDatapSLListNode* t = new InternalDatapSLListNode(item);
  if (last == 0)
    t->tl = last = t;
  else if (u == 0) // ins_after 0 means prepend
  {
    t->tl = last->tl;
    last->tl = t;
  }
  else
  {
    t->tl = u->tl;
    u->tl = t;
    if (u == last) 
      last = t;
  }
  return Pix(t);
}


void InternalDatapSLList::del_after(Pix p)
{
  InternalDatapSLListNode* u = (InternalDatapSLListNode*)p;
  if (last == 0 || u == last) error("cannot del_after last");
  if (u == 0) u = last; // del_after 0 means delete first
  InternalDatapSLListNode* t = u->tl;
  if (u == t)
    last = 0;
  else
  {
    u->tl = t->tl;
    if (last == t)
      last = u;
  }
  delete t;
}

int InternalDatapSLList::owns(Pix p)
{
  InternalDatapSLListNode* t = last;
  if (t != 0 && p != 0)
  {
    do
    {
      if (Pix(t) == p) return 1;
      t = t->tl;
    } while (t != last);
  }
  return 0;
}

InternalDatap InternalDatapSLList::remove_front()
{
  if (last == 0) error("remove_front of empty list");
  InternalDatapSLListNode* t = last->tl;
  InternalDatap res = t->hd;
  if (t == last)
    last = 0;
  else
    last->tl = t->tl;
  delete t;
  return res;
}

int InternalDatapSLList::remove_front(InternalDatap& x)
{
  if (last == 0)
    return 0;
  else
  {
    InternalDatapSLListNode* t = last->tl;
    x = t->hd;
    if (t == last)
      last = 0;
    else
      last->tl = t->tl;
    delete t;
    return 1;
  }
}


void InternalDatapSLList::del_front()
{
  if (last == 0) error("del_front of empty list");
  InternalDatapSLListNode* t = last->tl;
  if (t == last)
    last = 0;
  else
    last->tl = t->tl;
  delete t;
}

int InternalDatapSLList::OK()
{
  int v = 1;
  if (last != 0)
  {
    InternalDatapSLListNode* t = last;
    long count = MAXLONG;      // Lots of chances to find last!
    do
    {
      count--;
      t = t->tl;
    } while (count > 0 && t != last);
    v &= count > 0;
  }
  if (!v) error("invariant failure");
  return v;
}
