// This may look like C code, but it is really -*- C++ -*-
/* 
Copyright (C) 1988 Free Software Foundation
    written by Doug Lea (dl@rocky.oswego.edu)

This file is part of GNU CC.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY.  No author or distributor
accepts responsibility to anyone for the consequences of using it
or for whether it serves any particular purpose or works at all,
unless he says so in writing.  Refer to the GNU CC General Public
License for full details.

Everyone is granted permission to copy, modify and redistribute
GNU CC, but only under the conditions described in the
GNU CC General Public License.   A copy of this license is
supposed to have been given to you along with GNU CC so you
can know your rights and responsibilities.  It should be in a
file named COPYING.  Among other things, the copyright notice
and this notice must be preserved on all copies.  
*/


#ifndef _InternalDatapSLList_h
#ifdef __GNUG__
#pragma once
#pragma interface
#endif
#define _InternalDatapSLList_h 1

#include <Pix.h>
#include "InternalDatap.defs.h"

#ifndef _InternalDatapSLListNode_h
#define _InternalDatapSLListNode_h 1

struct InternalDatapSLListNode
{
  InternalDatapSLListNode*         tl;
  InternalDatap                    hd;
                         InternalDatapSLListNode();
                         InternalDatapSLListNode(InternalDatap  h, InternalDatapSLListNode* t = 0);
                         ~InternalDatapSLListNode();
};

inline InternalDatapSLListNode::InternalDatapSLListNode() {}

inline InternalDatapSLListNode::InternalDatapSLListNode(InternalDatap  h, InternalDatapSLListNode* t) :hd(h), tl(t) {}

inline InternalDatapSLListNode::~InternalDatapSLListNode() {}

typedef InternalDatapSLListNode* InternalDatapSLListNodePtr;

#endif


class InternalDatapSLList
{
protected:
  InternalDatapSLListNode*        last;

public:
                        InternalDatapSLList();
                        InternalDatapSLList(InternalDatapSLList& a);
                        ~InternalDatapSLList();

  InternalDatapSLList&            operator = (InternalDatapSLList& a);

  int                   empty();
  int                   length();

  void                  clear();

  Pix                   prepend(InternalDatap  item);
  Pix                   append(InternalDatap  item);

  void                  join(InternalDatapSLList&);

  Pix                   prepend(InternalDatapSLListNode*);
  Pix                   append(InternalDatapSLListNode*);

  InternalDatap&                  operator () (Pix p);
  Pix                   first();
  void                  next(Pix& p);
  int                   owns(Pix p);
  Pix                   ins_after(Pix p, InternalDatap  item);
  void                  del_after(Pix p);

  InternalDatap&                  front();
  InternalDatap&                  rear();
  InternalDatap                   remove_front();
  int                   remove_front(InternalDatap& x);
  void                  del_front();

  void                  error(const char* msg);
  int                   OK();
};


inline InternalDatapSLList::~InternalDatapSLList()
{
  clear();
}

inline InternalDatapSLList::InternalDatapSLList()
{
  last = 0;
}

inline int InternalDatapSLList::empty()
{
  return last == 0;
}


inline Pix InternalDatapSLList::first()
{
  return (last == 0)? 0 : Pix(last->tl);
}

inline void InternalDatapSLList::next(Pix& p)
{
  p = (p == 0 || p == last)? 0 : Pix(((InternalDatapSLListNode*)(p))->tl);
}

inline InternalDatap& InternalDatapSLList::operator () (Pix p)
{
  if (p == 0) error("null Pix");
  return ((InternalDatapSLListNode*)(p))->hd;
}

inline InternalDatap& InternalDatapSLList::front()
{
  if (last == 0) error("front: empty list");
  return last->tl->hd;
}

inline InternalDatap& InternalDatapSLList::rear()
{
  if (last == 0) error("rear: empty list");
  return last->hd;
}


#endif
