#include "RexxClauses.h"
#include "Clause.h"
#include "Expressionp.SLList.h"
#include "rexxparse.tok.h"


RexxClauses::RexxClauses() : pc(0), freespace(0), clauses(MinClauses) {
    clauses.fill(0); // fill them all with null
//    for (int i = 0; i < MaxClauses; i++) clauses[i] = 0;
}

RexxClauses::~RexxClauses() {
    Free(0);
}

void RexxClauses::CheckBound(int i) {
    if (i < 0) {
	cout << "RexxClauses::CheckBound() negative!\n";
	abort();
    }
    // no upper bound with xplex
/*
  else if (i > MaxClauses) {
	cout << "RexxClauses::CheckBound() >  MaxClauses\n";
	abort();
    }
*/
}

int RexxClauses::GetPC() {
    return pc;
}

int RexxClauses::Free() {
    return freespace;
}

void RexxClauses::Free(int start) {
    for (int i = start; i < freespace; i++) {
//	cout << "killing " << i << "\n";
	CheckBound(i);
	delete clauses[i];
	clauses[i] = 0;
    }
    freespace = start;
}

int RexxClauses::Goto(int pos) {
    CheckBound(pos);
    pc = pos;
    return 1;
}

int RexxClauses::Goto(String l, int start, int end) {
    for (int i = start; i < end; i++) {
	CheckBound(i);
	if (clauses[i]->IsA(TOK_LABEL)) {
	    if (((LabelClause*)clauses[i])->Matches(l)) {
		pc = i;
		return 1; // success at repositioning!
	    }
	}
    }
    return 0; // could reposition -- label not found
}

int RexxClauses::Find(String l, int start, int end) {
    for (int i = start; i < end; i++) {
	CheckBound(i);
	if (clauses[i]->IsA(TOK_LABEL)) {
	    if (((LabelClause*)clauses[i])->Matches(l)) {
		return i; // success at repositioning!
	    }
	}
    }
    return -1;
}

int RexxClauses::Exists(String l, int start, int end) {
    for (int i = start; i < end; i++) {
	CheckBound(i);
	if (clauses[i]->IsA(TOK_LABEL)) {
	    if (((LabelClause*)clauses[i])->Matches(l)) {
		return 1; // success at repositioning!
	    }
	}
    }
    return 0; // could reposition -- label not found
}


void RexxClauses::SetPC(int p) {
    CheckBound(p);
    pc = p;
}

Clause* RexxClauses::NextClause() {
    pc++;
    CheckBound(pc-1);
    return clauses[pc-1];
}

void RexxClauses::ResetPC() {
    pc = 0;
}

void RexxClauses::AddClause(Clause* cl) {
    CheckBound(freespace);
    if (clauses.fence() == freespace) {
	// expand the array
	clauses.add_high(cl);
    } else {
	clauses[freespace] = cl;
    }
    freespace++;
}




