// -*- C++ -*-
//---------------------
//--- RexxNumber.cc ---
//---------------------
/* 
 * Synopsis: 
 * Implementation Notes:
 * ToDo:
 * 
 * $Id: RexxNumber.cc,v 1.1 92/03/29 09:17:22 rvillari Exp Locker: rvillari $
 * $Log:	RexxNumber.cc,v $
Revision 1.1  92/03/29  09:17:22  rvillari
Initial revision

 *
 *
*/

#include "RexxNumber.h"
#include "stringsupport.h"
#include "Implementation.h"
#include "ArgStr.h"
#include <stream.h>
#include <assert.h>
#include <String.h>
#include <values.h>

const int MaxPow = 11;

static Integer iarray[MaxPow] = {1, Ipow(10, 1), Ipow(10, 2), Ipow(10, 3), Ipow(10, 4), Ipow(10, 5), Ipow(10, 6), Ipow(10, 7), Ipow(10, 8), Ipow(10, 9), Ipow(10, 10)};

inline Integer& fastpow(int i) {
    assert(i >= 0 && i < MaxPow);
    return iarray[i];
}

RexxNumber::RexxNumber(const ArgString& a, int Digits) : digits(0), number(0), exponent(0), state(0) {
    int numbersign = 0; // positive until proven otherwise
    int exponentsign = 0;
    int i = 0;

    int decimalfound = 0;

    while (a[i] == ' ') i++; // skip leading spaces

    if (a[i] == '+') {
	i++;
	while (a[i] == ' ') i++;
    } else if (a[i] == '-') {
	numbersign = 1;
	i++;
	while (a[i] == ' ') i++;
    }

    // decimal needs to be relative to first non-zero thing
    int firstnonzerodigitposition = -1;
    int decimalposition = -1;
    for ( ; i < a.Length(); i++) {
	char c = a[i];
	if (c >= '0' && c <= '9') {
	    if (digits >= Digits) {
		// truncate
		// if (c >= '5') number++;
		// skip to 'e' || 'E' || end of string
		for ( ; i < a.Length(); i++) {
		    c = a[i];
		    if (c == 'e' || c == 'E') break;
		    if (c == '.') decimalposition = i;
		    if (c == ' ')
			while (a[i] == ' ') i++;
//		    exponent++;
//		    digits++;
		}
		break;
	    }
	    number *= 10;
	    number += (c - '0');
	    if (firstnonzerodigitposition == -1 && c > '0') {
		firstnonzerodigitposition = i;
	    }
	    if (firstnonzerodigitposition != -1) digits++;
	} else if (c == '.') {
	    decimalposition = i;
	} else if (c == 'E' || c == 'e') {
	    break;
	} else if (c == ' ') {
	    while (a[i] == ' ') i++;
	}
    }
    
    if (decimalposition == -1) decimalposition = i;
    
    int expo = 0;
    for ( ; i < a.Length(); i++) {
	char c = a[i];
	if (c >= '0' && c <= '9') {
	    expo *= 10;
	    expo += (c - '0');
	} else if (c == '+') {
	} else if (c == '-') {
	    exponentsign = 1;
	} else if (c == ' ') {
	    while (a[i] == ' ') i++;
	}
    }
    exponent += expo;
    if (exponentsign) exponent = -exponent;
    if (numbersign) number.negate();

    if (firstnonzerodigitposition != -1) {
	// has decimal, nonzero
	if (decimalposition > firstnonzerodigitposition) {
	    exponent += - digits + (decimalposition - firstnonzerodigitposition);
	} else {
	    exponent -= (firstnonzerodigitposition - decimalposition) + digits - 1;
	}
    } else {
	number = 0;
	exponent = 0;
    }
}

RexxNumber::RexxNumber(long l) : exponent(0), state(0) {
    number = l;
    updatedigits();
}

RexxNumber::RexxNumber() : number(0), exponent(0), digits(0), state(0) {
}

RexxNumber::RexxNumber(const RexxNumber& r) : number(r.number), exponent(r.exponent), digits(r.digits) {    
}

RexxNumber::~RexxNumber() {
}

void RexxNumber::negate() {
    number.negate();
}

int RexxNumber::sign() const {
    return ::sign(number);
}

void RexxNumber::removetrailingzeros() {
    Integer n = number;
    Integer r, q;
    int zero = 0;

    while (::sign(n) != 0) { // do while number != 0
	divide(n, Integer(10), q, r); // get the remainder
	if (::sign(r) == 0) zero++;
	else break;
	n = q;
    }
    //we should have the number of zeros to remove
    while (zero != 0) {
	number = number / 10;
	zero--;
	digits--;
	exponent++;
    }
}

void RexxNumber::updatedigits(int hint) {
    // hint tells us where we can start looking
//    if (hint == 1) {
//	// maybe we can be better than this
//	hint = (lg(number) / 3) + 1;
//    }
    for (int i = hint; i < 10; i++) {
	if (ucompare(fastpow(i), number) > 0) {
	    digits = i;
	    break;
	}
    }

/*
    digits = 0;
    Integer n = number;
    while (n != 0) {
	n = n / 10;
	digits ++;
    }
*/
}

long RexxNumber::Whole() const {
    static RexxNumber temp;
    temp = *this;
    temp.removetrailingzeros();
    long result;
    int decpos = temp.digits + temp.exponent;
    if (decpos > MaxWholeDigits || decpos < temp.digits) return MAXLONG;
    result = temp.number;
    if (temp.exponent > 0) {
	// we need to add zeros at the end
	result *= fastpow(temp.exponent); // this is safe
    } else {
	// we need to remove some zeros at the end
	result /= fastpow(-temp.exponent);
    }
    return result;
}

int RexxNumber::IsAWhole() const {
    RexxNumber temp = *this;
    temp.removetrailingzeros();
    int decpos = temp.digits + temp.exponent;
    if (decpos > MaxWholeDigits) return 0; // too big
    if (decpos < temp.digits) return 0; // has decimal point
    return 1; // passed all known tests to make it a whole
}

const ArgString& Truncate(const RexxNumber& x, int Digits, int decimalplaces) {
    static ArgString result;
    result = "";
    RexxNumber r = round(x, Digits);
    int decpos = r.digits + r.exponent;
    if (decpos <= 0) {
	// it's a 0.something
	ArgString temp= Itoa(r.number);
	if (temp[0] == '-' && decimalplaces) {
	    result += "-0.";
	    result += (char*)replicate("0", -decpos);
	    result += substring(temp, 1, temp.Length());
	    result = padonright(result, decimalplaces + 3, '0');
	} else {
	    result += "0.";
	    result += (char*)replicate("0", -decpos);
	    result += temp;
	    if (decimalplaces) {
		result = padonright(result, decimalplaces + 2, '0');
	    } else {
		result = "0";
	    }
	}
    } else if (decpos < r.digits) {
	// it's a xx.yy
	ArgString temp = Itoa(r.number);
	if (temp[0] == '-') {
	    result += '-';
	    result += substring(temp, 1, decpos + 1);
	    if (decimalplaces) {
		result += '.';
		result += padonright(substring(temp, decpos + 1, temp.Length()), decimalplaces, '0');
	    }
	} else {
	    result += substring(temp, 0, decpos);
	    if (decimalplaces) {
		result += '.';
		result += padonright(substring(temp, decpos, temp.Length()), decimalplaces, '0');
	    }
	}
    } else {
	// it's a xx000
	result += Itoa(r.number);
	result += (char*)replicate("0", decpos - r.digits);
	if (decimalplaces) {
	    result += '.';
	    result += (char*)replicate("0", decimalplaces);
	}
    }
    return result;
}

const ArgString& Format(const RexxNumber& x, int Digits, int Type, int before, int after, int expp, int expt) {
    static ArgString result;
    result = "";
    RexxNumber r = round(x, Digits);
    int decpos = r.digits + r.exponent;

    // calculate the number of digits after the decimal point
    // and before
    
    int aafter;
    int abefore;

    if (decpos <= 0) {
	abefore = 0;
	aafter = r.digits - decpos;
    } else if (decpos < r.digits) {
	abefore = decpos;
	aafter = r.digits - decpos;
    } else {
	abefore = decpos;
	aafter = 0;
    }

    if (after >= 0 && after < aafter) {
	// we need to round off some digits
	r = round(r, r.digits - (aafter - after));
    }
    return Form(x, Digits, 0); 
}

// type is supposed to have 0 = sci, 1 = eng
const ArgString& Form(const RexxNumber& x, int Digits, int Type) {
    static ArgString result;
    result = "";
    RexxNumber r = round(x, Digits);
    int decpos = r.digits + r.exponent;

    if (r.number == 0) {
	result = "0";
    } else {
	if (decpos < -Digits || decpos > Digits) {
	    // exponential
	    r.removetrailingzeros();
	    decpos = r.digits + r.exponent;
	    ArgString temp = Itoa(r.number);
	    if (temp[0] == '-') {
		result += '-';
		result += substring(temp, 1, 2);
		if (temp.Length() != 2) {
		    result += '.';
		    result += substring(temp, 2, temp.Length());
		}
	    } else {
		result += substring(temp, 0, 1);
		if (temp.Length() != 1) {
		    result += '.';
		    result += substring(temp, 1, temp.Length());
		}
	    }
	    result += 'E';
	    int normexp = decpos - 1;
	    if (normexp > 0) result += '+';
	    result += ArgString((int)normexp);
	} else if (decpos <= 0) {
	    ArgString temp= Itoa(r.number);
	    if (temp[0] == '-') {
		result += "-0.";
		result += (char*)replicate("0", -decpos);
		result += substring(temp, 1, temp.Length());
	    } else {
		result += "0.";
		result += (char*)replicate("0", -decpos);
		result += temp;
	    }
	} else if (decpos <= Digits) {
	    if (decpos >= r.digits) {
		result += Itoa(r.number);
		result += (char*)replicate("0", decpos - r.digits);
	    } else {
		ArgString temp = Itoa(r.number);
		if (temp[0] == '-') {
		    result += '-';
		    result += substring(temp, 1, decpos + 1);
		    result += '.';
		    result += substring(temp, decpos + 1, temp.Length());
		} else {
		    result += substring(temp, 0, decpos);
		    result += '.';
		    result += substring(temp, decpos, temp.Length());
		}
	    }
	}
    }
    // should check this out -- if it's legal to cache it -- especially if we cache it and the user changes DIGITS on us
    result.number = new RexxNumber(r);  // we KNOW that this is a number, and we already have it as a RexxNumber, so we may as well cache this fact!
    return result;
}

RexxNumber& RexxNumber::operator=(const RexxNumber& r) {
    number = r.number;
    exponent = r.exponent;
    digits = r.digits;
    return *this;
}

RexxNumber& RexxNumber::operator=(long l) {
    number = l;
    exponent = 0;
    updatedigits();
    return *this;
}

int RexxNumber::ugreater(const RexxNumber& r) const {
    RexxNumber& x = *this;
    RexxNumber& y = r;

    // comparing x > y -- (*this > r)

    // if they're both zero, ugreater is false
    if (!::sign(x.number) && !::sign(y.number)) return 0;
    // if x == 0, ugreater is false
    else if (!::sign(x.number) && ::sign(y.number)) return 0;
    // if y == 0, ugreater is true
    else if (!::sign(y.number) && ::sign(x.number)) return 1;

    // we got here, so none of the numbers are zero
    // adjust the exponents so they're relative to the 0'th digit
    int adjxexp = x.exponent + x.digits;
    int adjyexp = y.exponent + y.digits;
    if (adjxexp > adjyexp) {
	return 1;
    } else if (adjyexp > adjxexp) {
	return 0;
    } else {
	// the adjusted exponents are equal
	if (x.digits > y.digits) {
	    RexxNumber z = y.number * fastpow(x.digits - y.digits);
	    return (ucompare(x.number, z.number) > 0);
	} else if (y.digits > x.digits) {
	    RexxNumber z = x.number * fastpow(y.digits - x.digits);
	    return (ucompare(z.number, y.number) > 0);
	} else {
	    return (ucompare(x.number, y.number) > 0);
	}
    }

/*
    if (x.digits > y.digits) {
	// x has more digits than y
	y.number *= fastpow(x.digits - y.digits); // safe, never too large
	y.exponent -= x.digits - y.digits;
    } else {
	// y has more digits than x
	y.number /= fastpow(y.digits - x.digits); // safe, never too large
	y.exponent += y.digits - x.digits;
    }
    
    if (x.exponent > y.exponent) return 1;
    else if (x.exponent == y.exponent && ucompare(x.number, y.number) > 0) return 1;
    return 0;
*/
}

int RexxNumber::greater(const RexxNumber& r) const {
    RexxNumber& x = *this;
    RexxNumber& y = r;

    // comparing x > y -- (*this > r)

    // if they're both zero, ugreater is false
    if (::sign(x.number) > ::sign(y.number)) return 1;

    // adjust the exponents so they're relative to the 0'th digit
    int adjxexp = x.exponent + x.digits;
    int adjyexp = y.exponent + y.digits;
    if (adjxexp > adjyexp) {
	if (::sign(x.number) > 0) {
	    return 1;
	} else {
	    return 0;
	}
    } else if (adjyexp > adjxexp) {
	if (::sign(x.number) > 0) {
	    return 0;
	} else {
	    return 1;
	}
    } else {
	// the adjusted exponents are equal
	if (x.digits > y.digits) {
	    RexxNumber z = y.number * fastpow(x.digits - y.digits);
	    return (compare(x.number, z.number) > 0);
	} else if (y.digits > x.digits) {
	    RexxNumber z = x.number * fastpow(y.digits - x.digits);
	    return (compare(z.number, y.number) > 0);
	} else {
	    return (compare(x.number, y.number) > 0);
	}
    }
}

void RexxNumber::Debug() const {
    cout << number << "," << exponent << "," << digits << "\n";
    cout << digits + exponent << "\n";
}

// FRIEND functions

int multiply(const RexxNumber& xx, const RexxNumber& yy, int Digits, RexxNumber& result) {
    result = trunc(xx, Digits + 1);
    static RexxNumber y;

    y = trunc(yy, Digits + 1);

    result.number *= y.number;
    result.exponent += y.exponent;
    // check exponent to make sure it doesn't exceed a certain value
    result.updatedigits();
    return 1;
}

int power(const RexxNumber& xx, const RexxNumber& yy, int Digits, RexxNumber& result) {
    static RexxNumber x, y;

    x = trunc(xx, Digits + 1);
    y = trunc(yy, Digits + 1); // y is whole number

    // normalize y to whole number!
    y = y.Whole();
    int L = y.digits;

    int ysign = sign(y.number);
    if (ysign < 0) y.number.negate();

    int maxbits = lg(y.number);

    // calculate value = x ^ y
    static RexxNumber accumulator;

    accumulator = 1;
    static RexxNumber temp;
    for (int i = maxbits; i >= 0; i--) {
	if (testbit(y.number, i)) {
	    multiply(accumulator, x, Digits + L + 1, temp);
	    accumulator = temp;
	}
	if (i != 0) {
	    multiply(accumulator, accumulator, Digits + L + 1, temp);
	    accumulator = temp;
	}
    }
    if (ysign < 0) {
	divide(1, accumulator, Digits + L + 1, temp);
	accumulator = temp;
    }
    
    accumulator = round(accumulator, Digits);
    accumulator.removetrailingzeros();
    result = accumulator;
    return 1;
}

int divide(const RexxNumber& xx, const RexxNumber& yy, int Digits, RexxNumber& result) {
    static RexxNumber x, y;

    x = trunc(xx, Digits+1);
    y = trunc(yy, Digits+1);

    if (y.sign() == 0) {
	return 0; // overflow (division by zero)
    }

    static RexxNumber quotient;
    // calculating quotient = x / y
    // do while y is greater than x
    while (ucompare(y.number, x.number) > 0) {
	// adjust x so that it is greater than y
	// extend result with 0's
	x.number *= 10;
	x.exponent -= 1;
    }

    // set exponent of quotient
    quotient.exponent = x.exponent - y.exponent;
    Integer q, r = x.number;
    // calculate initial quotient and remainder
    // still need to determine whether calling divide is safe with (r,...,r)!
    Integer remainder;
    divide(r, y.number, q, remainder);
    r = remainder;

    quotient.number = q;
    quotient.updatedigits();

    while (quotient.digits < Digits + 1) {
	r *= 10; // add a zero to remainder
	divide(r, y.number, q, remainder); // divide
	r = remainder;
	quotient.number *= 10; // add quotient to our number
	quotient.number += q;
	quotient.digits++; // add number of digits
	quotient.exponent--;
    }

    quotient = round(quotient, Digits);
    quotient.removetrailingzeros();
    result = quotient;
    return 1;
}

int integerdivide(const RexxNumber& xx, const RexxNumber& yy, int Digits, RexxNumber& qr, RexxNumber& rr) {
    static RexxNumber x, y;

    x = trunc(xx, Digits+1);
    y = trunc(yy, Digits+1);

    if (y.sign() == 0) {
	return 0; // overflow (division by zero)
    }
    if (y.ugreater(x)) {
	qr = RexxNumber(0);
	rr = x;
	rr = round(rr, Digits);
	rr.removetrailingzeros();
	
	if (qr.IsAWhole()) return 1;
	return 0;
    }

    // for sure, x is greater than y

    // calculating quotient = x / y

    static RexxNumber quotient;
    // do while y.number is greater than x.number (adjustment phase)
    while (ucompare(y.number, x.number) > 0) {
	// adjust x so that it is greater than y
	// extend result with 0's
	x.number *= 10;
	x.exponent -= 1;
    }

    // adjust y so that x.exponent - yexponent >= 0
    if (x.exponent - y.exponent < 0) {
	while (y.exponent - x.exponent) {
	    y.number *= 10;
	    y.exponent--;
	    y.digits++;
	}
	/* code above replaces this
	y.number *= fastpow(y.exponent - x.exponent);
	y.exponent -= y.exponent - x.exponent;
	y.digits += y.exponent - x.exponent;
	*/
    }

    // set exponent of quotient

    quotient.exponent = x.exponent - y.exponent;

    Integer q, r = x.number;

    // calculate initial quotient and remainder
    // still need to determine whether calling divide is safe with (r,...,r)!

    Integer remainder;
    divide(r, y.number, q, remainder);
    r = remainder;

    quotient.number = q;
    quotient.updatedigits();

    while (quotient.digits < Digits + 1) {
	if (quotient.exponent == 0) {
	    break;
	}
	r *= 10; // add a zero to remainder
	divide(r, y.number, q, remainder); // divide
	r = remainder;
	quotient.number *= 10; // add quotient to our number
	quotient.number += q;
	quotient.digits++; // add number of digits
	quotient.exponent--;
    }

    // quotient
    quotient = round(quotient, Digits);
    quotient.removetrailingzeros();
    qr = quotient;

    // remainder
    rr.number = r;
    rr.exponent = y.exponent; // need to check if this is RIGHT
    rr.updatedigits();
    rr = round(rr, Digits);
    rr.removetrailingzeros();
    if (qr.IsAWhole()) return 1;
    return 0;
}

int add(const RexxNumber& xx, const RexxNumber& yy, int Digits, RexxNumber& result) {
    static RexxNumber x, y;

    if (yy.ugreater(xx)) {
	x = trunc(yy, Digits+1);
	y = trunc(xx, Digits+1);
    } else if (y.exponent > x.exponent) {
	x = trunc(yy, Digits+1);
	y = trunc(xx, Digits+1);
    } else {
	x = trunc(xx, Digits+1);
	y = trunc(yy, Digits+1);
    }
//    x.Debug();
//    y.Debug();

//    x = trunc(x, Digits+1);
//    y = trunc(y, Digits+1);
    // assume x >= y
//    int nx = x.exponent + x.digits;
//    int ny = y.exponent + y.digits;
    
    // if the difference between x.exponent and y.exponent > 0 (they don't match), and the number of digits of x isn't already at the maximum (Digits + 1), adjust x
    if (x.digits < (Digits + 1) && x.exponent > y.exponent) {
	if (x.exponent - y.exponent > ((Digits+1) - x.digits)) {
	    // we can only expand to Digits + 1
	    x.number *= fastpow((Digits+1) - x.digits); // safe, never too large
	    x.exponent -= ((Digits+1) - x.digits);
	} else {
	    // x.exp - y.exp doesn't require us to expand beyond Digits+1
	    x.number *= fastpow(x.exponent - y.exponent); // safe, never too large
	    x.exponent -= (x.exponent - y.exponent);
	}
	x.updatedigits();
    }
    int xorigdigits = x.digits;

    if (y.exponent > x.exponent) {
	// y is bigger than x, need to remove some zeros
	while (y.exponent > x.exponent && y.sign() != 0) {
	    y.number *= 10;
	    y.exponent--;
	}
	// at the end, exponents should match
    } else if (x.exponent > y.exponent) {
	// y is smaller than x, need to add some zeros
	while (x.exponent > y.exponent) {
	    y.number /= 10;
	    y.exponent++;
	}
    }

    if (::sign(x.number) == ::sign(y.number)) {
	x.number += y.number;
	// if they are the same sign
	x.updatedigits(xorigdigits);
    } else {
	x.number += y.number;
	x.updatedigits();
    }

    if (x.digits > xorigdigits) xorigdigits = x.digits;
    
    // xorigdigits - x.digits is the number of leading zeros
    int leadingzeros = xorigdigits - x.digits;

    // need to do special rounding for add/sub -- due to weird counting of leading zeros
    result =  round(x, Digits - leadingzeros);
    return 1;
}

int subtract(const RexxNumber& xx, const RexxNumber& yy, int Digits, RexxNumber& result) {
    static RexxNumber x, y;

    if (yy.ugreater(xx)) {
	x = trunc(yy, Digits+1);
	y = trunc(xx, Digits+1);
	x.negate();
    } else if (y.exponent > x.exponent) {
	x = trunc(yy, Digits+1);
	y = trunc(xx, Digits+1);
	x.negate();
    } else {
	x = trunc(xx, Digits+1);
	y = trunc(yy, Digits+1);
	y.negate();
    }
//    x.Debug();
//    y.Debug();

//    x = trunc(x, Digits+1);
//    y = trunc(y, Digits+1);
    // assume x >= y
//    int nx = x.exponent + x.digits;
//    int ny = y.exponent + y.digits;
    
    // if the difference between x.exponent and y.exponent > 0 (they don't match), and the number of digits of x isn't already at the maximum (Digits + 1), adjust x
    if (x.digits < (Digits + 1) && x.exponent > y.exponent) {
	if (x.exponent - y.exponent > ((Digits+1) - x.digits)) {
	    // we can only expand to Digits + 1
	    x.number *= fastpow((Digits+1) - x.digits); // safe, never too large
	    x.exponent -= ((Digits+1) - x.digits);
	} else {
	    // x.exp - y.exp doesn't require us to expand beyond Digits+1
	    x.number *= fastpow(x.exponent - y.exponent); // safe, never too large
	    x.exponent -= (x.exponent - y.exponent);
	}
	x.updatedigits();
    }
    int xorigdigits = x.digits;

    if (y.exponent > x.exponent) {
	// y is bigger than x, need to remove some zeros
	while (y.exponent > x.exponent && y.sign() != 0) {
	    y.number *= 10;
	    y.exponent--;
	}
	// at the end, exponents should match
    } else if (x.exponent > y.exponent) {
	// y is smaller than x, need to add some zeros
	while (x.exponent > y.exponent) {
	    y.number /= 10;
	    y.exponent++;
	}
    }

    if (::sign(x.number) == ::sign(y.number)) {
	x.number += y.number;
	// if they are the different signs, the numbers only get bigger
	x.updatedigits(xorigdigits);
    } else {
	x.number += y.number;
	x.updatedigits();
    }

    if (x.digits > xorigdigits) xorigdigits = x.digits;
    
    // xorigdigits - x.digits is the number of leading zeros
    int leadingzeros = xorigdigits - x.digits;

    // need to do special rounding for add/sub -- due to weird counting of leading zeros
    result =  round(x, Digits - leadingzeros);
    return 1;
}

const RexxNumber& trunc(const RexxNumber& xx, int Digits) {
    static RexxNumber x;
    x = xx;
    if (xx.digits > Digits) {
	x = xx;
	x.number /= fastpow(x.digits - Digits);
	x.exponent += (x.digits - Digits);
	x.digits = Digits;
	return x;
    } else {
	return x;
    }
}

const RexxNumber& round(const RexxNumber& xx, int Digits) {
    static RexxNumber x;
    x = xx;
    if (xx.digits > Digits) {
	x = xx;
	if (sign(x.number) < 0)
	    x.number -= 5 * fastpow(x.digits - Digits - 1);
	else
	    x.number += 5 * fastpow(x.digits - Digits - 1);

	x.updatedigits();
	x.number /= fastpow(x.digits - Digits);
	x.exponent += (x.digits - Digits);
	x.digits = Digits;
	return x;
    } else {
	return x;
    }
}

#ifdef TESTNUMBER
main(int argc, char* argv[]) {
    if (argc != 4) abort();
    ArgString x(argv[1]);
    ArgString y(argv[2]);
    int Digits = atoi(argv[3]);
    RexxNumber rx(x, 50);
    RexxNumber ry(y, 50);
    rx.Debug();
    ry.Debug();
    RexxNumber addxy = add(rx, ry, Digits);
    RexxNumber mulxy = multiply(rx, ry, Digits);
    RexxNumber divxy = divide(rx, ry, Digits);
    RexxNumber powxy = power(rx, ry, Digits);
    RexxNumber idivxy, iremxy;
    integerdivide(rx, ry, Digits, idivxy, iremxy);
    ArgString result = Form(addxy, Digits);
    cout << "x + y = (" << (char*)result << ")\n";
    result = Form(mulxy, Digits);
    cout << "x * y = (" << (char*)result << ")\n";
    result = Form(divxy, Digits);
    cout << "x / y = (" << (char*)result << ")\n";
    result = Form(powxy, Digits);
    cout << "x ** y = (" << (char*)result << ")\n";
    result = Form(idivxy, Digits);
    cout << "x % y = (" << (char*)result << ")\n";
    result = Form(iremxy, Digits);
    cout << "x // y = (" << (char*)result << ")\n";
}
#endif
