// -*- C++ -*-
//---------------------
//--- RexxStream.cc ---
//---------------------
/* 
 * Synopsis: 
 * Implementation Notes:
 * ToDo:
 * 
 * $Id: RexxStream.cc,v 1.1 92/03/29 09:17:26 rvillari Exp Locker: rvillari $
 * $Log:	RexxStream.cc,v $
Revision 1.1  92/03/29  09:17:26  rvillari
Initial revision

 *
 *
*/

#include "RexxStream.h"
#include "ArgStr.h"
#include <String.h>
#include <stdio.h>

RexxStream::RexxStream(String& fn) : filename(fn), readposition(0), writeposition(0), lasterror(SNotOpen), file(0), length(0), filestate(FNotOpen) {
}

RexxStream::~RexxStream() {
}

SErrors RexxStream::Error() {
    return lasterror;
}

#define IOERROR(number, count) {				      \
    lasterror = number;							      \
    return count;						      \
}

int RexxStream::isreadaccess() {
    if (filestate == FRead || filestate == FReadWrite) return 1;
    switch (filestate) {
      case FNotOpen : 
	lasterror = SNotOpen;
	break;
      case FWrite :
	lasterror = SNotReadable;
	break;
    };
    return 0;
}

int RexxStream::iswriteaccess() {
    if (filestate == FWrite || filestate == FReadWrite) return 1;
    switch (filestate) {
      case FNotOpen : 
	lasterror = SNotOpen;
	break;
      case FRead :
	lasterror = SNotWriteable;
	break;
    };
    return 0;
}

int RexxStream::isreadwriteaccess() {
    if (filestate == FReadWrite) return 1;
    // set the correct error here
    switch (filestate) {
      case FNotOpen : 
	lasterror = SNotOpen;
	break;
      case FRead :
	lasterror = SNotReadWrite;
	break;
    };
    return 0;
}

int RexxStream::Close() {
    ClearError();
    if (file) {
	fclose(file);
	file = 0;
	filestate = FNotOpen;
    }
    return 1;
}

int RexxStream::Exists() {
    ClearError();
    if (filestate == FNotOpen) {
	// trial open the file
	file = fopen(filename, "r");
	if (file) {
	    fclose(file);
	    file = 0;
	    return 1; // we exist
	} else {
	    return 0;
	}
    } else {
	// for sure it exists
	return 1;
    }
}

int RexxStream::ReadAccess() {
    ClearError();
    if (filestate == FRead || filestate == FReadWrite) return 1;
    if (file) fclose(file);
    file = fopen(filename, "r");
    if (file) {
	filestate = FRead;
	return 1;
    }
    else {
	lasterror = SNotOpen;
	filestate = FNotOpen;
	return 0;
    }
}

// open tries first read/write, then read, then write
int RexxStream::Open() {
    if (ReadWriteAccess()) return 1;
    else if (ReadAccess()) return 1;
    else if (WriteAccess()) return 1;
    else return 0; // failed completely
}

int RexxStream::WriteAccess() {
    ClearError();
    if (filestate == FWrite || filestate == FReadWrite) return 1;
    if (file) fclose(file);
    file = fopen(filename, "w");
    if (file) {
	filestate = FWrite;
	return 1;
    } else {
	lasterror = SNotOpen;
	filestate = FNotOpen;
	return 0;
    }
}

int RexxStream::ReadWriteAccess() {
    ClearError();
    if (filestate == FReadWrite) return 1;
    if (file) fclose(file);
    file = fopen(filename, "r+");
    if (!file) file = fopen(filename, "w+");

    if (!file) {
	filestate = FNotOpen;
	lasterror = SNotOpen;
	return 0;
    }

    if (fseek(file, 0, 2) == -1) {
	fclose(file);
	file = 0;
	filestate = FNotOpen;
	lasterror = SSeekError;
	return 0;
    }
    writeposition = ftell(file);

    filestate = FReadWrite;
    return 1;
}

void RexxStream::ClearError() {
    if (file) clearerr(file);
    lasterror = SNoError;
}

int RexxStream::Linein(ArgString& in, int count, int linenumber) {
    if (lasterror != SNoError) return 0;

    if (!isreadaccess()) return 0;

    // seek to linenumber
    // all errors returned are SeekErrors, because we were trying to seek
    if (linenumber >= 0) {
	if (fseek(file, 0, 0) == -1) IOERROR(SSeekError, 0);
	while (linenumber > 0) {
	    char* ret = fgets(buf, 256, file);
	    if (!ret) {
		IOERROR(SSeekError, 0);
	    }
	    linenumber--;
	}
	readposition = ftell(file);
    }

    // count is 0 or 1
    if (count) {
	if (fseek(file, readposition, 0) == -1) IOERROR(SSeekError, 0);
	char* ret = fgets(buf, 256, file);
	if (!ret) {
	    if (feof(file)) {IOERROR(SEOF, 0);}
	    else IOERROR(SReadError, 0);
	}
	buf[strlen(buf)-1] = '\0';
	in = ArgString(buf);
	readposition = ftell(file); // update the new position
    }
    return count; // success
}

int RexxStream::Lineout(ArgString& out, int linenumber) {
    if (lasterror != SNoError) return 1;

    // at this time, we need isreadwriteaccess to the file

    
    // seek to linenumber
    if (linenumber >= 0) {
	if (!isreadaccess()) return 1; // we need access to do seeks!
	if (fseek(file, 0, 0) == -1) {IOERROR(SSeekError, 1);} // go to the beginning
	while (linenumber > 0) {
	    char* ret = fgets(buf, 256, file); // throw this away
	    if (!ret) {
		IOERROR(SSeekError, 1);
	    }
	    linenumber--;
	}
	writeposition = ftell(file);
    }

    if (!iswriteaccess()) return 1;

    if (fseek(file, writeposition, 0) == -1) IOERROR(SSeekError, 1);
    strncpy(buf, out.String(), out.Length());
    buf[out.Length()] = '\n';
    int ret = fwrite(buf, out.Length()+1, 1, file);
    if (!ret) {
	if (feof(file)) {IOERROR(SEOF, 1);}
	else IOERROR(SWriteError, 1);
    }
    writeposition = ftell(file);
    return 0; // success
}

int RexxStream::Charin(ArgString& in, int count, int charnumber) {
    if (lasterror != SNoError) return 0;

    // need read access here
    if (!isreadaccess()) return 0;

    // seek to charnumber

    if (charnumber >= 0) {
	if (fseek(file, charnumber, 0) == -1) IOERROR(SSeekError, 0); // go to the beginning
	readposition = ftell(file);
    }

    if (count) {
	if (fseek(file, readposition, 0) == -1) IOERROR(SSeekError, 0);
	int ret = fread(buf, 1, count, file);
	if (!ret) {
	    if (feof(file)) {IOERROR(SEOF, 0);}
	    else {IOERROR(SReadError, 0);}
	}
	in = ArgString(buf, ret);
	readposition = ftell(file);
	return ret; // success
    }
    return count; // success
}

int RexxStream::Charout(ArgString& out, int charnumber) {
    if (lasterror != SNoError) return out.Length();

    if (!isreadwriteaccess()) return out.Length();

    if (charnumber >= 0) {
	if (fseek(file, charnumber, 0) == -1) IOERROR(SSeekError, out.Length());
	writeposition = charnumber;
    }

    if (fseek(file, writeposition, 0) == -1) IOERROR(SSeekError, out.Length());
    int ret = fwrite(out.String(), out.Length(), 1, file);
    if (!ret) {
	if (feof(file)) {IOERROR(SEOF, out.Length());}
	else IOERROR(SWriteError, out.Length());
    }
    writeposition = ftell(file); // new writeposition
    return out.Length() - ret; // success
}

ArgString RexxStream::State() {
    switch (filestate) {
      case FNotOpen :
	return ArgString("ERROR");
	break;
      case FRead :
	if (lasterror != SNoError) return ArgString("ERROR");
	if (Chars()) return ArgString("READY");
	else return ArgString("NOTREADY");
	break;
      case FReadWrite :
      case FWrite :
	if (lasterror != SNoError) return ArgString("ERROR");
	if (Chars()) return ArgString("READY");
	else return ArgString("UNKNOWN");
	break;
    };
}

ArgString RexxStream::Description() {
    switch (lasterror) {
      case SNoError :
	return ArgString("OK");
        break;
      case SNotReadable :
	return ArgString("NOT READ");
	break;
      case SNotWriteable :
	return ArgString("NOT WRITE");
	break;
      case SNotReadWrite :
	return ArgString("NOT READ/WRITE");
	break;
      case SSeekError :
	return ArgString("SEEK ERROR");
	break;
      case SReadError :
	return ArgString("READ ERROR");
	break;
      case SWriteError :
	return ArgString("WRITE ERROR");
	break;
      case SNotOpen :
	return ArgString("NOT OPEN");
	break;
      case SEOF :
	return ArgString("END INPUT");
	break;
    };
}

int RexxStream::updatelength() {
    if (lasterror != SNoError) return 0;

    if (!isreadaccess()) return 0;
    if (fseek(file, 0, 2) == -1) IOERROR(SSeekError, 0);
    length = ftell(file); // the length of the file
    return 1;
}

int RexxStream::Chars() {
    if (!updatelength()) return 0;
    return (length - readposition);
}

int RexxStream::Lines() {
    if (!updatelength()) return 0;
    // how many lines to EOF - we need to read the file from the current position to the end and count '\n' -- very inefficient, so we'll take the easy way out for now
    return ((length - readposition) ? 1 : 0);
}

RexxNonSeekableStream::RexxNonSeekableStream(String& fn, FILE* fp, FState fs) : RexxStream(fn) {
    file = fp;
    filestate = fs;
    lasterror = SNoError;
}

int RexxNonSeekableStream::Chars() {
    if (feof(file)) return 0;
    else return 1;
}

int RexxNonSeekableStream::Lines() {
    if (feof(file)) return 0;
    else return 1;
}

int RexxNonSeekableStream::Charin(ArgString& in, int count, int charnumber) {
    if (lasterror != SNoError) return 0;

    // need read access here
    if (!isreadaccess()) return 0;

    // seek to charnumber

    if (charnumber >= 0) {
	IOERROR(SSeekError, 0); // go to the beginning
    }

    if (count) {
	int ret = fread(buf, 1, count, file);
	if (!ret) {
	    if (feof(file)) {IOERROR(SEOF, 0);}
	    else {IOERROR(SReadError, 0);}
	}
	in = ArgString(buf, ret);
	return ret; // success
    }
    return count; // success
}

int RexxNonSeekableStream::Linein(ArgString& in, int count, int linenumber) {
    if (lasterror != SNoError) return 0;

    if (!isreadaccess()) return 0;

    // seek to linenumber
    // all errors returned are SeekErrors, because we were trying to seek
    if (linenumber >= 0) {
	// we are NON-SEEKABLE!
	IOERROR(SSeekError, 0);
    }

    // count is 0 or 1
    if (count) {
	// we are non-seekable
	char* ret = fgets(buf, 256, file);
	if (!ret) {
	    if (feof(file)) {IOERROR(SEOF, 0);}
	    else IOERROR(SReadError, 0);
	}
	buf[strlen(buf)-1] = '\0';
	in = ArgString(buf);
    }
    return count; // success
}

