// -*- C++ -*-
//--------------------
//--- RexxStream.h ---
//--------------------
/*
 * Responsibility: handles persistent rexx streams
 * Role: 
 * Communications: 
 * Bugs & Caveats: 
 * Created: Wed Mar 18  1:46:32 1992 by rvillari
 * Maintainer: 
 * 
 * $Id: RexxStream.h,v 1.1 92/03/29 09:19:24 rvillari Exp Locker: rvillari $
 * $Log:	RexxStream.h,v $
 * Revision 1.1  92/03/29  09:19:24  rvillari
 * Initial revision
 * 
 * 
 * 
*/


#ifndef _RexxStream_h
#pragma once
#define _RexxStream_h 1

#include <stdio.h>
class ArgString;
#include <String.h>

enum SErrors {SNoError, SNotReadable, SNotWriteable, SNotReadWrite, SSeekError, SReadError, SWriteError, SEOF, SNotOpen};

enum FState {FNotOpen, FRead, FWrite, FReadWrite};

class RexxStream {
  protected:
    String filename;
    char buf[256];
    FILE* file;
    FState filestate;
    int readposition;
    int writeposition;
    int length;
    SErrors lasterror;

    int updatelength();
    int iswriteaccess(); 
    int isreadaccess(); // needed for linein, charin
    int isreadwriteaccess(); 
  public:
    RexxStream(String&); // the file name
    void ClearError();

    virtual ArgString State();
    virtual ArgString Description();
    virtual int Linein(ArgString&, int = 1, int = -1); // returnvalue, count, line#
    virtual int Lineout(ArgString&, int = -1); // value, line#
    virtual int Charin(ArgString&, int = 1, int = -1); // returnvalue, count, char#
    virtual int Charout(ArgString&, int = 1); // value, char#
    virtual int Chars(); // how many characters left
    virtual int Lines(); // how many lines left
    int Open(); // tries readwrite, then read, then write
    int WriteAccess(); // ensures we have a writable file
    int ReadAccess(); // ensures that we have a readable file available
    int ReadWriteAccess(); // ensures that we have a wrietable file available
    int Close(); // closes a file
    int Exists(); // checks for existence of file
    SErrors Error(); // what was the last error?
    virtual ~RexxStream();
};

class RexxNonSeekableStream : public RexxStream {
  public:
    RexxNonSeekableStream(String&, FILE*, FState);

    virtual int Chars(); // how many characters left
    virtual int Lines(); // how many lines left
    virtual int Linein(ArgString&, int = 1, int = -1); // returnvalue, count, line#
    virtual int Charin(ArgString&, int = 1, int = -1); // returnvalue, count, char#
};

#endif 



