// This may look like C code, but it is really -*- C++ -*-
/* 
Copyright (C) 1988 Free Software Foundation
    written by Doug Lea (dl@rocky.oswego.edu)
    based on code by Marc Shapiro (shapiro@sor.inria.fr)

This file is part of GNU CC.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY.  No author or distributor
accepts responsibility to anyone for the consequences of using it
or for whether it serves any particular purpose or works at all,
unless he says so in writing.  Refer to the GNU CC General Public
License for full details.

Everyone is granted permission to copy, modify and redistribute
GNU CC, but only under the conditions described in the
GNU CC General Public License.   A copy of this license is
supposed to have been given to you along with GNU CC so you
can know your rights and responsibilities.  It should be in a
file named COPYING.  Among other things, the copyright notice
and this notice must be preserved on all copies.  
*/

#ifndef _RexxTokenpXPlex_h
#ifdef __GNUG__
#pragma once
#pragma interface
#endif
#define _RexxTokenpXPlex_h 1

#include "RexxTokenp.Plex.h"

class RexxTokenpXPlex: public RexxTokenpPlex
{
  RexxTokenpIChunk*       ch;           // cached chunk

  void             make_initial_chunks(int up = 1);

  void             cache(int idx) const;
  void             cache(const RexxTokenp* p) const;

  RexxTokenp*             dopred(const RexxTokenp* p) const;
  RexxTokenp*             dosucc(const RexxTokenp* p) const;

  void             set_cache(const RexxTokenpIChunk* t) const; // logically, 
                                               // not physically const
public:
                   RexxTokenpXPlex();                 // set low = 0;
                                               // fence = 0;
                                               // csize = default

                   RexxTokenpXPlex(int ch_size);      // low = 0; 
                                               // fence = 0;
                                               // csize = ch_size

                   RexxTokenpXPlex(int lo,            // low = lo; 
                            int ch_size);      // fence=lo
                                               // csize = ch_size

                   RexxTokenpXPlex(int lo,            // low = lo
                            int hi,            // fence = hi+1
                            const RexxTokenp  initval,// fill with initval,
                            int ch_size = 0);  // csize= ch_size
                                               // or fence-lo if 0

                   RexxTokenpXPlex(const RexxTokenpXPlex&);
  
  void             operator= (const RexxTokenpXPlex&);

// virtuals


  RexxTokenp&             high_element ();
  RexxTokenp&             low_element ();

  const RexxTokenp&       high_element () const;
  const RexxTokenp&       low_element () const;

  Pix              first() const;
  Pix              last() const;
  void             prev(Pix& ptr) const;
  void             next(Pix& ptr) const;
  int              owns(Pix p) const;
  RexxTokenp&             operator () (Pix p);
  const RexxTokenp&       operator () (Pix p) const;

  int              low() const; 
  int              high() const;
  int              valid(int idx) const;
  void             prev(int& idx) const;
  void             next(int& x) const;
  RexxTokenp&             operator [] (int index);
  const RexxTokenp&       operator [] (int index) const;
    
  int              Pix_to_index(Pix p) const;
  Pix              index_to_Pix(int idx) const;    

  int              can_add_high() const;
  int              can_add_low() const;
  int              full() const;

  int              add_high(const RexxTokenp  elem);
  int              del_high ();
  int              add_low (const RexxTokenp  elem);
  int              del_low ();

  void             fill(const RexxTokenp  x);
  void             fill(const RexxTokenp  x, int from, int to);
  void             clear();
  void             reverse();
  void             append(const RexxTokenpXPlex& a); 
  void             prepend(const RexxTokenpXPlex& a);
    
  int              OK () const; 

};

inline void RexxTokenpXPlex::prev(int& idx) const
{
  --idx;
}

inline void RexxTokenpXPlex::next(int& idx) const
{
  ++idx;
}

inline  int RexxTokenpXPlex::full () const
{
  return 0;
}

inline int RexxTokenpXPlex::can_add_high() const
{
  return 1;
}

inline int RexxTokenpXPlex::can_add_low() const
{
  return 1;
}

inline  int RexxTokenpXPlex::valid (int idx) const
{
  return idx >= lo && idx < fnc;
}

inline int RexxTokenpXPlex::low() const
{
  return lo;
}

inline int RexxTokenpXPlex::high() const
{
  return fnc - 1;
}

inline RexxTokenp& RexxTokenpXPlex:: operator [] (int idx)
{
  if (!ch->actual_index(idx)) cache(idx);
  return *(ch->pointer_to(idx));
}

inline const RexxTokenp& RexxTokenpXPlex:: operator [] (int idx) const
{
  if (!ch->actual_index(idx)) cache(idx);
  return *((const RexxTokenp*)(ch->pointer_to(idx)));
}

inline  RexxTokenp& RexxTokenpXPlex::low_element ()
{
  if (empty()) index_error();
  return *(hd->pointer_to(lo));
}

inline  const RexxTokenp& RexxTokenpXPlex::low_element () const
{
  if (empty()) index_error();
  return *((const RexxTokenp*)(hd->pointer_to(lo)));
}

inline  RexxTokenp& RexxTokenpXPlex::high_element ()
{
  if (empty()) index_error();
  return *(tl()->pointer_to(fnc - 1));
}

inline const RexxTokenp& RexxTokenpXPlex::high_element () const
{
  if (empty()) index_error();
  return *((const RexxTokenp*)(tl()->pointer_to(fnc - 1)));
}

inline  int RexxTokenpXPlex::Pix_to_index(Pix px) const
{
  RexxTokenp* p = (RexxTokenp*)px;
  if (!ch->actual_pointer(p)) cache(p);
  return ch->index_of(p);
}

inline  Pix RexxTokenpXPlex::index_to_Pix(int idx) const
{
  if (!ch->actual_index(idx)) cache(idx);
  return (Pix)(ch->pointer_to(idx));
}

inline Pix RexxTokenpXPlex::first() const
{
  return Pix(hd->RexxTokenpIChunk::first_pointer());
}

inline Pix RexxTokenpXPlex::last() const
{
  return Pix(tl()->RexxTokenpIChunk::last_pointer());
}

inline void RexxTokenpXPlex::prev(Pix& p) const
{
  Pix q = Pix(ch->RexxTokenpIChunk::pred((RexxTokenp*) p));
  p = (q == 0)? Pix(dopred((const RexxTokenp*) p)) : q;
}

inline void RexxTokenpXPlex::next(Pix& p) const
{
  Pix q = Pix(ch->RexxTokenpIChunk::succ((RexxTokenp*) p));
  p = (q == 0)? Pix(dosucc((const RexxTokenp*)p)) : q;
}

inline RexxTokenp& RexxTokenpXPlex:: operator () (Pix p)
{
  return *((RexxTokenp*)p);
}

inline const RexxTokenp& RexxTokenpXPlex:: operator () (Pix p) const
{
  return *((const RexxTokenp*)p);
}

inline void RexxTokenpXPlex::set_cache(const RexxTokenpIChunk* t) const
{
  ((RexxTokenpXPlex*)(this))->ch = (RexxTokenpIChunk*)t;
}

#endif
