#include "RexxTokens.h"
#include "Clause.h"
#include "Express.h"
#include "ArgStr.h"
#include "rexxparse.tok.h"
#include "stringsupport.h"

RexxToken::RexxToken(int t) : type(t), secondary(0) {}

RexxToken::~RexxToken() {
}

ArgString* RexxToken::Value() {
    return &value;
}

int RexxToken::Type() {
    return type;
}

int RexxToken::Secondary() {
    return secondary;
}

void RexxToken::Set(int t, int s, ArgString v) {
    type = t;
    secondary = s;
    value = v;
}

RexxTokens::RexxTokens() : pc(0), tokens(MinTokens), finalpc(-1) {
    tokens.fill(0); // fill it all with null
}

RexxTokens::~RexxTokens() {}

void RexxTokens::Dump() {
    ResetPC();
    while (NextToken().Type()) {}
    ResetPC();
}

void RexxTokens::CheckBound(int i) {
    if (i < 0) {
	cout << "RexxTokens::CheckBound() negative!\n";
	abort();
    }
/*
  else if (i > MaxTokens) {
	cout << "RexxTokens::CheckBound() > MaxTokens!\n";
	abort();
    }
*/
}

void RexxTokens::Backup() {
    pc--;
    CheckBound(pc);
}

RexxToken& RexxTokens::NextToken() {
    pc++;
    CheckBound(pc-1);
    return *tokens[pc-1];
}

RexxToken& RexxTokens::operator[](int progcount) {
//    CheckBound(progcount);
    static RexxToken unknowntoken(TOK_UNKNOWN);
    if (progcount < 0 || progcount > finalpc) return unknowntoken;
    return *tokens[progcount];
}

void RexxTokens::ResetPC() {
    pc = 0;
}

int RexxTokens::LastTokenType() {
    if ((pc-1) < 0) return TOK_UNKNOWN;
    CheckBound(pc-1);
    return tokens[pc-1]->Type();
}

void RexxTokens::AddToken(int type, int secondary, ArgString a) {
    CheckBound(pc);
    
    if (tokens.fence() == pc) {
	// expand the array
	tokens.add_high(new RexxToken);
    }
    tokens[pc]->Set(type, secondary, a);
    if (type == 0) finalpc = pc;
    pc++;
}

void RexxTokens::AddToken(int type, int secondary, char* string) {
    CheckBound(pc);

    if (tokens.fence() == pc) {
	// expand the array
	tokens.add_high(new RexxToken);
    }
    tokens[pc]->Set(type, secondary, ArgString(string));
    if (type == 0) finalpc = pc;

    pc++;
}

void RexxTokens::AddToken(int type, char* string) {
    CheckBound(pc);
    if (tokens.fence() == pc) {
	// expand the array
	tokens.add_high(new RexxToken);
    }
    tokens[pc]->Set(type, 0, ArgString(string));
    if (type == 0) finalpc = pc;

    pc++;
}

void RexxTokens::AddToken(int type) {
    CheckBound(pc);
    if (tokens.fence() == pc) {
	// expand the array
	tokens.add_high(new RexxToken);
    }
    tokens[pc]->Set(type, 0, ArgString());
    if (type == 0) finalpc = pc;

    pc++;
}

void RexxTokens::clear() {
    // need to grow our plex
    if (tokens.fence() == 0) {
	// expand the array
	tokens.add_high(new RexxToken);
    }
    
    tokens[0]->Set(0, 0, ArgString());
    finalpc = -1;
    
    ResetPC();
}


