#include "SymTab.h"
#include "ArgStr.h"
#include "Name.Symbolp.VHMap.h"
#include <String.h>
#include "Symbol.h"
#include "Glob.h"
#include "stringsupport.h"

extern Glob glob;

//static void KillWhite(String& s) {
//    s.gsub(RXwhite,"");
//}

SymbolTable::SymbolTable() : table(0) {
}

SymbolTable::~SymbolTable() {
    for (Pix i = table.first(); i != 0; table.next(i)) {
	if (table.contents(i)->IsOwnedBy(this)) {
	    delete table.contents(i);
	}
    }
}

const Symbol* SymbolTable::GetSymbol(String& s, const Symbol* def) const {
    SymbolTable* lthis = (SymbolTable*)this;
    Symbol* symbol = lthis->table[s];

    if (!symbol) {
	if (!def) symbol = new Symbol(this);
	else symbol = (Symbol*)def;
	lthis->table[s] = (Symbol*) symbol;
    }

    if (symbol->IsValid()) {
	glob.UninitializedAccess = 0;
    } else {
	symbol->Write(ArgString(s), 0);
	glob.UninitializedAccess = 1;
    }

    return symbol;
}

const Symbol* SymbolTable::GetStemSymbol(String& s, const Symbol* def) const {
    SymbolTable* lthis = (SymbolTable*)this;
    Symbol* symbol = lthis->table[s];

    if (!symbol) {
	if (!def) symbol = new Stem(this);
	else symbol = (Symbol*)def;
	lthis->table[s] = (Symbol*) symbol;
    }

    if (symbol->IsValid()) {
	glob.UninitializedAccess = 0;
    } else {
	glob.UninitializedAccess = 1;
	symbol->Write(ArgString(s), 0);
    }

    return symbol;
}

String& SymbolTable::Expand(String& s) const {
    static String t1, t2, res;
    res = "";
    t1 = s + "|";  // 
    while (1) {
	t2 = t1.before(Regex("[.|]"));
	if (t2 != "" && issscsymbol(ArgString((char*)t2))) {
	    ArgString& temp = Read(t2);
	    res = res + temp.String();
	} else {
	    res = res + t2; // probably some constant
	}
//	cout << t2 << "\n";
	t1 = t1.after(".");
	if (t1 == "") break;
	res = res + ".";
    }
    return res;
}

const Symbol* SymbolTable::GetCompoundSymbol(String& s, const Symbol* def) const {
//    KillWhite(s);
    // break up and evaluate...
    static String stem, t1;

    stem = s.through("."); // we got the stem part
    
    t1 = s.after(".");

    String& res = Expand(t1);

    const Stem* temp = (Stem*)GetStemSymbol(stem); // ensure that we have the stem in memory
    Symbol* symbol = (Symbol*)temp->GetSymbol(res, def);
    if (!symbol->IsValid()) {
	glob.UninitializedAccess = 1;
	if (temp->IsValid()) {
	    symbol->Write(temp->Read(), 0);
	} else {
	    res = stem.before(".") + res;
	    symbol->Write(ArgString(res), 0);
	}
    } else {
	glob.UninitializedAccess = 0;
    }
    return symbol;
}

void SymbolTable::Drop(String& s) {
//    KillWhite(s);
    Symbol* symbol = (Symbol*)GetSymbol(s);
    if (symbol) symbol->Drop();
    else abort();
}

void SymbolTable::DropStem(String& s) {
//    KillWhite(s);
    Symbol* symbol = (Symbol*)GetStemSymbol(s);
    if (symbol) symbol->Drop();
    else abort();
/*
    if (table.contains(s)) {
	table[s]->Drop();
    }
*/
}

void SymbolTable::DropCompound(String& s) {
//    KillWhite(s);
    Symbol* symbol = (Symbol*)GetCompoundSymbol(s);
    symbol->Drop();
}

const ArgString& SymbolTable::Read(String& s, const Symbol* def) const {
//    KillWhite(s);
    const Symbol* symbol = GetSymbol(s, def);
    if (!symbol) abort();

    return symbol->Read();
}

const ArgString& SymbolTable::ReadStem(String& s, const Symbol* def) const {
//    KillWhite(s);
    SymbolTable* lthis = (SymbolTable*)this;

    const Symbol* symbol = GetStemSymbol(s, def);
    if (!symbol) abort();

    return symbol->Read();
}

const ArgString& SymbolTable::ReadCompound(String& s, const Symbol* def) const {
    //    KillWhite(s);
    const Symbol* symbol = GetCompoundSymbol(s, def);

    if (!symbol) abort();

    return symbol->Read();
}

void SymbolTable::Write(String& s, const ArgString& value) {
//    KillWhite(s);
    Symbol* symbol = (Symbol*)GetSymbol(s);
    if (!symbol) abort();
    
    symbol->Write(value);
}

void SymbolTable::WriteStem(String& s, const ArgString& value) {
//    KillWhite(s);
    Symbol* symbol = (Symbol*)GetStemSymbol(s);
    if (!symbol) abort();

    symbol->Write(value);
}

void SymbolTable::WriteCompound(String& s, const ArgString& value) {
//    KillWhite(s);
    Symbol* symbol = (Symbol*)GetCompoundSymbol(s);

    symbol->Write(value);
}


