#include "SymTabs.h"
#include "Symbol.h"
#include "SymbolTablep.SLStack.h"
#include "SymTab.h"
#include "ArgStr.h"
#include <String.h>
#include "stringsupport.h"
#include "Errors.h"
#include "support.h"

SymbolTables::SymbolTables() {}

SymbolTables::~SymbolTables() {}

const Symbol* SymbolTables::GetSymbolAny(const String& s, const Symbol* def) const {
    if (!s.contains(".")) return GetSymbol(s, def);
    else if (s.contains(".", s.length() - 1) && s.freq(".") == 1) return GetStemSymbol(s, def);
    else if (s.contains(".")) return GetCompoundSymbol(s, def);
    abort();
}

const Symbol* SymbolTables::GetSymbol(const String& s, const Symbol* def) const {
    SymbolTables* lthis = (SymbolTables*)this;
    return lthis->tables.top()->GetSymbol(s, def);
}

const Symbol* SymbolTables::GetStemSymbol(const String& s, const Symbol* def) const {
    SymbolTables* lthis = (SymbolTables*)this;
    return lthis->tables.top()->GetStemSymbol(s, def);
}

const Symbol* SymbolTables::GetCompoundSymbol(const String& s, const Symbol* def) const {
    SymbolTables* lthis = (SymbolTables*)this;
    return lthis->tables.top()->GetCompoundSymbol(s, def);
}

void SymbolTables::DropAny(const String& s) {
    // this assumes a proper symbol
	// need to add check for this later on
    if (!s.contains(".")) return DropSimple(s);
    else if (s.contains(".", s.length() - 1) && s.freq(".") == 1) return DropStem(s);
    else if (s.contains(".")) return DropCompound(s);
    abort();
}

void SymbolTables::DropSimple(const String& s) {
    return tables.top()->Drop(s);
}

void SymbolTables::DropStem(const String& s) {
    return tables.top()->DropStem(s);
}

void SymbolTables::DropCompound(const String& s) {
    return tables.top()->DropCompound(s);
}




void SymbolTables::PushSymbolTable(SymbolTable* s) {
    tables.push(((s == 0) ? new SymbolTable : s));
}

SymbolTable* SymbolTables::PopSymbolTable() {
    return tables.pop();
}


SymbolTable* SymbolTables::TopSymbolTable() {
    return tables.top();
}


int SymbolTables::Expose(const String& s) {
    String vars[MaxExposes];
    // expose the symbols in s
    int count = split(s, vars, 100, RXwhite);
    for (int i = 0; i < count; i++) {
	if (!issscsymbol(ArgString((char*)vars[i]))) {
	    SetError(ESymbolExpected);
	    return 0;
	}
	if (iscompoundsymbol(ArgString((char*)vars[i]))) {
	    static String stem, tail;
	    stem = vars[i].through(".");
	    tail = vars[i].after(".");

	    String& expanded = tables.top()->Expand(tail);
	    SymbolTable* top = PopSymbolTable();

	    Stem* stemp = (Stem*)GetStemSymbol(stem);
	    const Symbol* def = stemp->GetSymbol(expanded);

	    PushSymbolTable(top); // push it back
	    stemp = (Stem*)GetStemSymbol(stem);
	    stemp->GetSymbol(expanded, def);
	} else {
	    SymbolTable* top = PopSymbolTable(); // get the top to expose the next
	    const Symbol* def = GetSymbolAny(vars[i]);
	    PushSymbolTable(top); // put it back
	    ReadAny(vars[i], def);
	}
    }
    return 1;
}

int SymbolTables::Drop(const String& s) {
    String vars[MaxDrops];
    // drop the symbols in s
    int count = split(s, vars, 100, RXwhite);
    for (int i = 0; i < count; i++) {
	if (!issscsymbol(ArgString((char*)vars[i]))) {
	    SetError(ESymbolExpected);
	    return 0;
	}
	DropAny(vars[i]); // drop the symbol
    }
    return 1;
}
