#ifndef _SYMBOLTABLES_H
#pragma once
#define _SYMBOLTABLES_H

#include "SymbolTablep.SLStack.h"
#include "SymTab.h"

class String;
class ArgString;
class Symbol;

class SymbolTables {
  private:
    SymbolTablepSLStack tables;
  public:
    SymbolTables();
    ~SymbolTables();

    const Symbol* GetSymbolAny(const String&, const Symbol* = 0) const;
    const Symbol* GetSymbol(const String&, const Symbol* = 0) const;
    const Symbol* GetStemSymbol(const String&, const Symbol* = 0) const;
    const Symbol* GetCompoundSymbol(const String&, const Symbol* = 0) const;

    inline const ArgString& ReadAny(const String&, const Symbol* = 0) const; // this is slow!  it has to think
    inline const ArgString& Read(const String&, const Symbol* = 0) const;
    inline const ArgString& ReadStem(const String&, const Symbol* = 0) const;
    inline const ArgString& ReadCompound(const String&, const Symbol* = 0) const;

    void DropAny(const String&);
    void DropSimple(const String&);
    void DropStem(const String&);
    void DropCompound(const String&);

    inline void WriteAny(const String&, const ArgString&);
    inline void Write(const String&, const ArgString&);
    inline void WriteStem(const String&, const ArgString&);
    inline void WriteCompound(const String&, const ArgString&);

    int Expose(const String&);
    int Drop(const String&);

    void PushSymbolTable(SymbolTable* = 0);
    SymbolTable* PopSymbolTable();
    SymbolTable* TopSymbolTable();
};

inline const ArgString& SymbolTables::ReadAny(const String& s, const Symbol* def) const {
    // this assumes a proper symbol
	// need to add check for this later on
    if (!s.contains(".")) return Read(s, def);
    else if (s.contains(".", s.length() - 1) && s.freq(".") == 1) return ReadStem(s, def);
    else if (s.contains(".")) return ReadCompound(s, def);
    abort();
}

inline const ArgString& SymbolTables::Read(const String& s, const Symbol* def) const {
    SymbolTables* lthis = (SymbolTables*)this;
    return lthis->tables.top()->Read(s, def);
}

inline const ArgString& SymbolTables::ReadStem(const String& s, const Symbol* def) const {
    SymbolTables* lthis = (SymbolTables*)this;
    return lthis->tables.top()->ReadStem(s, def);
}

inline const ArgString& SymbolTables::ReadCompound(const String& s, const Symbol* def) const {
    SymbolTables* lthis = (SymbolTables*)this;
    return lthis->tables.top()->ReadCompound(s, def);
}

inline void SymbolTables::WriteAny(const String& s, const ArgString& a) {
    // this assumes a proper symbol
    // need to add check for this later on
    if (!s.contains(".")) Write(s, a);
    else if (s.contains(".", s.length() - 1) && s.freq(".") == 1) WriteStem(s, a);
    else if (s.contains(".")) WriteCompound(s, a);
    else abort();
}

inline void SymbolTables::Write(const String& s, const ArgString& a) {
    tables.top()->Write(s, a);
}

inline void SymbolTables::WriteStem(const String& s, const ArgString& a) {
    tables.top()->WriteStem(s, a);
}

inline void SymbolTables::WriteCompound(const String& s, const ArgString& a) {
    tables.top()->WriteCompound(s, a);
}

#endif


