#ifndef _SYMBOL_H
#pragma once
#define _SYMBOL_H

class String;

#include "Name.Symbolp.VHMap.h"
#include "ArgStr.h"

class SymbolTable;

class Symbol {
  private:
    int state;
    ArgString value;
    const void* owner; // the owner of this symbol
  public:
    Symbol(const void*); // register the owner with me
    Symbol(const ArgString&, const void*);
    inline int IsOwnedBy(const void*) const;
    virtual ~Symbol();
    virtual void Drop();
    inline void Write(const ArgString&, int = 1);
    inline const ArgString& Read() const;
    inline int IsValid() const ;
};

inline int Symbol::IsValid() const {
    return state;
}

inline const ArgString& Symbol::Read() const {
    return value;
}

inline void Symbol::Write(const ArgString& v, int s) {
    value = v;
    state = s; // we are valid/invalid
}

inline int Symbol::IsOwnedBy(const void* s) const {
    return s == owner;
}

// Stem class

class Stem : public Symbol {
  private:
    NameSymbolpVHMap table;
  public:
    Stem(const void*);
    Stem(const ArgString&, const void*);
    virtual ~Stem();
    virtual void Drop();
    void Drop(const String&);
    inline const Symbol* GetSymbol(const String&, const Symbol* = 0) const;
    inline const ArgString& Read(const String&, const Symbol* = 0) const;
    inline Write(const String&, const ArgString&, int = 1);
};

inline const ArgString& Stem::Read(const String& s, const Symbol* def) const {
    const Symbol* symbol = GetSymbol(s, def);

    if (!symbol) abort();

    return symbol->Read();
}

inline const Symbol* Stem::GetSymbol(const String& s, const Symbol* def) const {
    Stem* lthis = (Stem*)this;
    const Symbol* symbol = lthis->table[s];

    if (!symbol) {
	if (def == 0) symbol = new Symbol(this);
	else symbol = def;
	lthis->table[s] = (Symbol*)symbol;
    }

    return symbol;
}

inline Stem::Write(const String& s, const ArgString& value, int st) {
    Symbol* symbol = (Symbol*)GetSymbol(s);

    if (!symbol) abort();

    symbol->Write(value, st);
}

#endif

