// This may look like C code, but it is really -*- C++ -*-
/* 
Copyright (C) 1988 Free Software Foundation
    written by Doug Lea (dl@rocky.oswego.edu)

This file is part of GNU CC.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY.  No author or distributor
accepts responsibility to anyone for the consequences of using it
or for whether it serves any particular purpose or works at all,
unless he says so in writing.  Refer to the GNU CC General Public
License for full details.

Everyone is granted permission to copy, modify and redistribute
GNU CC, but only under the conditions described in the
GNU CC General Public License.   A copy of this license is
supposed to have been given to you along with GNU CC so you
can know your rights and responsibilities.  It should be in a
file named COPYING.  Among other things, the copyright notice
and this notice must be preserved on all copies.  
*/


#ifndef _SymbolTablepSLList_h
#ifdef __GNUG__
#pragma once
#pragma interface
#endif
#define _SymbolTablepSLList_h 1

#include <Pix.h>
#include "SymbolTablep.defs.h"

#ifndef _SymbolTablepSLListNode_h
#define _SymbolTablepSLListNode_h 1

struct SymbolTablepSLListNode
{
  SymbolTablepSLListNode*         tl;
  SymbolTablep                    hd;
                         SymbolTablepSLListNode();
                         SymbolTablepSLListNode(SymbolTablep  h, SymbolTablepSLListNode* t = 0);
                         ~SymbolTablepSLListNode();
};

inline SymbolTablepSLListNode::SymbolTablepSLListNode() {}

inline SymbolTablepSLListNode::SymbolTablepSLListNode(SymbolTablep  h, SymbolTablepSLListNode* t) :hd(h), tl(t) {}

inline SymbolTablepSLListNode::~SymbolTablepSLListNode() {}

typedef SymbolTablepSLListNode* SymbolTablepSLListNodePtr;

#endif


class SymbolTablepSLList
{
protected:
  SymbolTablepSLListNode*        last;

public:
                        SymbolTablepSLList();
                        SymbolTablepSLList(SymbolTablepSLList& a);
                        ~SymbolTablepSLList();

  SymbolTablepSLList&            operator = (SymbolTablepSLList& a);

  int                   empty();
  int                   length();

  void                  clear();

  Pix                   prepend(SymbolTablep  item);
  Pix                   append(SymbolTablep  item);

  void                  join(SymbolTablepSLList&);

  Pix                   prepend(SymbolTablepSLListNode*);
  Pix                   append(SymbolTablepSLListNode*);

  SymbolTablep&                  operator () (Pix p);
  Pix                   first();
  void                  next(Pix& p);
  int                   owns(Pix p);
  Pix                   ins_after(Pix p, SymbolTablep  item);
  void                  del_after(Pix p);

  SymbolTablep&                  front();
  SymbolTablep&                  rear();
  SymbolTablep                   remove_front();
  int                   remove_front(SymbolTablep& x);
  void                  del_front();

  void                  error(const char* msg);
  int                   OK();
};


inline SymbolTablepSLList::~SymbolTablepSLList()
{
  clear();
}

inline SymbolTablepSLList::SymbolTablepSLList()
{
  last = 0;
}

inline int SymbolTablepSLList::empty()
{
  return last == 0;
}


inline Pix SymbolTablepSLList::first()
{
  return (last == 0)? 0 : Pix(last->tl);
}

inline void SymbolTablepSLList::next(Pix& p)
{
  p = (p == 0 || p == last)? 0 : Pix(((SymbolTablepSLListNode*)(p))->tl);
}

inline SymbolTablep& SymbolTablepSLList::operator () (Pix p)
{
  if (p == 0) error("null Pix");
  return ((SymbolTablepSLListNode*)(p))->hd;
}

inline SymbolTablep& SymbolTablepSLList::front()
{
  if (last == 0) error("front: empty list");
  return last->tl->hd;
}

inline SymbolTablep& SymbolTablepSLList::rear()
{
  if (last == 0) error("rear: empty list");
  return last->hd;
}


#endif
