// -*- C++ -*-
//-------------------
//--- Variable.cc ---
//-------------------
/* 
 * Synopsis: 
 * Implementation Notes:
 * ToDo:
 * 
 * $Id$
 * $Log$
 *
 *
*/

#include "Variable.h"
#include "Glob.h"
#include "stringsupport.h"

extern Glob glob;

// Variable

Variable::Variable() : symboltable(0) {
}

SymbolTable* Variable::CurrentTable() {
    return symboltable;
}

int Variable::HasChanged() {
    SymbolTable* top = glob.SymbolTables.TopSymbolTable();
    if (symboltable != top) {
	// we changed symboltables, so we need to update ours
	symboltable = top;
	return 1;
    } else {
	return 0;
    }
}

// fixed variable (contradiction in terms)

FixedVariable::FixedVariable(String& n) : name((char*)n) {
}

ArgString& FixedVariable::Read() {
    return name;
}

void FixedVariable::Write(ArgString& a) {
    abort();
}

// Simple variable

SimpleVariable::SimpleVariable(String& n) : name(n), symbol(0) {
}

ArgString& SimpleVariable::Read() {
    if (HasChanged()) {
	symbol = (Symbol*)CurrentTable()->GetSymbol(name);
	return symbol->Read();
    } else {
	if (symbol->IsValid()) {
	    glob.UninitializedAccess = 0;
	} else {
	    symbol->Write(ArgString(name), 0);
	    glob.UninitializedAccess = 1;
	}
	return symbol->Read();
    }
}

void SimpleVariable::Write(ArgString& a) {
    if (HasChanged()) {
	symbol = (Symbol*)CurrentTable()->GetSymbol(name);
	symbol->Write(a);
    } else {
	symbol->Write(a);
    }
}

// Stem variable

StemVariable::StemVariable(String& n) : name(n), stem(0) {
}

ArgString& StemVariable::Read() {
    if (HasChanged()) {
	stem = (Stem*)CurrentTable()->GetStemSymbol(name);
	return stem->Read();
    } else {
	if (stem->IsValid()) {
	    glob.UninitializedAccess = 0;
	} else {
	    stem->Write(ArgString(name), 0);
	    glob.UninitializedAccess = 1;
	}
	return stem->Read();
    }
}

void StemVariable::Write(ArgString& a) {
    if (HasChanged()) {
	stem = (Stem*)CurrentTable()->GetStemSymbol(name);
	stem->Write(a);
    } else {
	stem->Write(a);
    }
}

CompoundVariable::CompoundVariable(String& n) : stem(0), variables(0) {
    static String varnames[50]; // make this a constant later
    static String t1;
    static const Regex regex("[.|]");

    nodes = 0;
    stemname = n.through("."); // stem part
    t1 = n.after(".") + "|"; // '|' is invalid in a variable, so this is safe
    while (1) {
	varnames[nodes] = t1.before(regex);
	t1 = t1.after(".");
	nodes++;
	if (t1 == "") break;
    }

    // create our Variables here
    variables = new Variable*[nodes];

    for (int j = 0; j < nodes; j++) {
	if (issscsymbol((char*)varnames[j])) {
	    variables[j] = new SimpleVariable(varnames[j]);
	} else {
	    variables[j] = new FixedVariable(varnames[j]);
	}
    }
}

String& CompoundVariable::Expand() {
    static String result;

    result = "";

    int nodesminusone = nodes - 1;
    for (int i = 0; i < nodesminusone; i++) {
	result += (char*)variables[i]->Read();
	result += ".";
    }
    result += (char*)variables[i]->Read(); // this is for the last one!

    return result;
}

ArgString& CompoundVariable::Read() {
    String& expanded = Expand();
    if (HasChanged()) {
	stem = (Stem*)CurrentTable()->GetStemSymbol(stemname);
    }

    // stem should be valid
    Symbol* symbol = (Symbol*)stem->GetSymbol(expanded);
    if (!symbol->IsValid()) {
	glob.UninitializedAccess = 1;
	if (stem->IsValid()) {
	    symbol->Write(stem->Read());
	} else {
	    String fullname = stemname + expanded;
	    symbol->Write(ArgString((char*)fullname));
	}
    } else {
	glob.UninitializedAccess = 0;
    }

    return symbol->Read();
}

void CompoundVariable::Write(ArgString& a) {
    if (HasChanged()) {
	stem = (Stem*)CurrentTable()->GetStemSymbol(stemname);
    }
    String& expanded = Expand();
    stem->Write(expanded, a);
}


