/* Rexx benchmarks */

LOOPCOUNT=10000
SHORTLOOP=1000
  log='benchmark.log'                                         /* ets */
  call stream log, 'c', 'b'    /* open for read/write */
  "rm" log ">/dev/null 2>&1"                       /* rav */
  call lineout log,,                                          /* ets */
      '                  ELAPSED       VIRT CPU      TOT CPU' /* ets */
  Say '                  ELAPSED       VIRT CPU      TOT CPU'
  call dobench 'magnifier';
/*  call dobench 'forloop';
  call dobench 'whileloop';
  call dobench 'repeatloop';
  call dobench 'literalassign';
  call dobench 'memoryaccess';
  call dobench 'realarithmetic';
  call dobench 'realalgebra';
  call dobench 'vector';
  call dobench 'equalif';
  call dobench 'unequalif'; */
  call dobench 'noparameters';
  call dobench 'values';
  call dobench 'reference';
  call dobench 'command';
  call lineout log                                            /* ets */
  exit;
dobench: routine=arg(1)
   junk = time('Reset');
   bvi=0                                                      /* ets */
   bto=0

   select
     when routine='magnifier'        then     call magnifier
     when routine='forloop'          then     call forloop
     when routine='whileloop'        then     call whileloop
     when routine='repeatloop'       then     call repeatloop
     when routine='literalassign'    then     call literalassign
     when routine='memoryaccess'     then     call memoryaccess
     when routine='realarithmetic'   then     call realarithmetic
     when routine='realalgebra'      then     call realalgebra
     when routine='vector'           then     call vector
     when routine='equalif'          then     call equalif
     when routine='unequalif'        then     call unequalif
     when routine='noparameters'     then     call noparameters
     when routine='values'           then     call values
     when routine='reference'        then     call reference
     when routine='command'          then     call command
   end
/**/
   elap = time('Elapsed');
   evi=0                                                        /* ets */
   eto = 0
   result = left(routine,12) format(elap,6,6),
                             format(evi-bvi,6,6),
                             format(eto-bto,6,6)
   say result
/* 'EXECIO 1 DISKW BENCHMAR LOG A (VAR  RESULT' */              /* ets */
   call lineout log,result                                      /* ets */
   return

/* Rexx Benchmarks. Adapted from the Pascal routines published in
                    PC World, Vol. 5, No. 5, May 1982 pages 141-142.
   Pascal version by E. J. W. West, IBM Hursley. 4/30/1982.
   C code by Joseph Kesselman, IBM Kingston, 12/26/1985.
   Rexx version by Alex MARTELLI at ROMESC, 10/10/88.
      K.Hansjakob changed INTERPRET to SELECT, 11/88.
      B.Moser changed IOX -> EXECIO, sin/cos -> STATE command, 10/91.
*/

/* Null loop, to serve as an overhead measurement for the following
   tests. The name of the routine comes from the use of the outer
   LOOPCOUNT-loop to magnify later measurements to a measurable size.
*/
magnifier: do k=1 to LOOPCOUNT; end; return;

/* Nested FOR loop test. */
forloop: do k=1 to LOOPCOUNT; do j=1 to 10; end; end; return;

/* Compare WHILE against above FOR results.  */
whileloop: do k = 1 to LOOPCOUNT;
    j = 1; do while j<=10; j=j+1; end;
    end; return;

/* Now to measure the REPEAT, or test-at-the-end, variation. */
repeatloop: do k = 1 to LOOPCOUNT;
    j = 1; do until j>10; j=j+1; end;
    end; return;

/* Speed of Literal assignment is an easy one, given the above */
literalassign: do k = 1 to LOOPCOUNT;
    do j=1 to 10; l=0; end;
    end; return;

/* Compare a variable fetch with the above literal */
memoryaccess: do k = 1 to LOOPCOUNT;
    do j= 1 to 10; l=j; end;
    end; return;

/* Floating-point math.  */
realarithmetic: do k = 1 to LOOPCOUNT;
    x = k/2.0*3.0+4.0-5.0;
    end; return;

/* Like the previous test, but use variable rather than constants */
realalgebra: do k = 1 to LOOPCOUNT;
    x = k/k + k - k;
    end; return;

/* Another common operation:  Array indexing.  */
vector:
    matrix.0 = 0;
    do k = 1 to LOOPCOUNT
        do j = 1 to 10; jm1=j-1; matrix.j = matrix.jm1; end;
    end; return;

/* This is a comparison standard for the following test.  It measures
how fast an IF statement runs if the two branches are taken an equal
number of times. */
equalif: do k = 1 to LOOPCOUNT
    do j = 1 to 10; if j<6 then l=1; else l=0; end;
    end; return;

/* Now let's see what happens if we disrupt the symmetry of the above
test by taking one branch more than the other.  This measures relative
speed of the two branches, and may indicate which should be used for the
more commonly-taken decision */
unequalif: do k = 1 to LOOPCOUNT
    do j = 1 to 10; if j<2 then l=1; else l=0; end;
    end; return;

/* Procedure-calling overhead is next on the list.  Simplest case is
where there are no parameters passed.  Note that none5 modifies a global
variable; we'll refer back to this variable in later routines. */
noparameters: global=0; do k=1 to LOOPCOUNT; call none1; end; return;

none1: call none2; return;
none2: call none3; return;
none3: call none4; return;
none4: call none5; return;
none5: global=1; return;

/* now with an argument */
values: global=0; do k=1 to LOOPCOUNT; call value1 k; end; return;

value1: arg i; call value2 i; return
value2: arg i; call value3 i; return
value3: arg i; call value4 i; return
value4: arg i; call value5 i; return
value5: arg i; global = i; return

/*call-by-reference: oops, NO support for this in Rexx ! so using name*/
reference: global=0; do k=1 to LOOPCOUNT; call refer1 'k'; end; return

refer1: arg nmi; call refer2 nmi; return
refer2: arg nmi; call refer3 nmi; return
refer3: arg nmi; call refer4 nmi; return
refer4: arg nmi; call refer5 nmi; return
refer5: arg nmi; global=value(nmi); return

/* Finally, we measure the performance of a pure host command.
   Note that this is slow enough that we have to decrease the
   magnification (k) counter or we'd be here all day. */
command: do k=1 to SHORTLOOP

/*  'STATE BENCHMAR EXEC'  */
     "test /"
end; return;

pst: procedure                                                 /* ets */
     call popen 'ps'                                           /* ets */
     pull h                                                    /* ets */
     pos=wordpos("TIME",h)                                     /* ets */
     time=""                                                   /* ets */
     do while queued()>0                                       /* ets */
        pull l                                                 /* ets */
        if wordpos("RXX",l)>0                                  /* ets */
        then do                                                /* ets */
             if time<>""                                       /* ets */
             then do                                           /* ets */
                  say "more than one rexx program running!"    /* ets */
                  exit 4                                       /* ets */
                  end                                          /* ets */
             else time=word(l,pos)                             /* ets */
             end                                               /* ets */
        end                                                    /* ets */
     if time=""                                                /* ets */
     then do                                                   /* ets */
          say "cannot find my process!"                        /* ets */
          exit 4                                               /* ets */
          end                                                  /* ets */
     parse var time m ":" s                                    /* ets */
     return m*60+s                                                /* ets */










