#include "Glob.h"
#include "RexxTokens.h"
#include "ArgStr.h"
#include "Express.h"
#include "rexxparse.tok.h"
#include <stream.h>

extern Glob glob;

static int ppindex;

void addtoken(int t, int s = 0,  ArgString a = "") {
    // this adds concat if we have a symbol followed by a string
    // or a string followed by a symbol
    if ((glob.Tokens2->LastTokenType() == TOK_STRING_LITERAL &&
	 (t == TOK_SYMBOL || t == TOK_FIXED)) || 
	((glob.Tokens2->LastTokenType() == TOK_SYMBOL || 
	  glob.Tokens2->LastTokenType() == TOK_FIXED) &&
	 t == TOK_STRING_LITERAL))
	glob.Tokens2->AddToken(TOK_CONCAT);
    glob.Tokens2->AddToken(t, s, a);
}

void backupifblank() {
    if (glob.Tokens2->LastTokenType() == TOK_BLANK) glob.Tokens2->Backup();
}

void skipspace() {
    while ((*glob.Tokens)[ppindex].Type() == TOK_BLANK) {
	ppindex++;
    }
}

void load(int);
int match(int);
int match(int, int);
int match(int, int, int);
void skiptokens(int);
int tokentype(int);

#define settype(t) (*glob.Tokens)[ppindex].Set(t, 0, "")

// this returns current token and advances one
RexxToken& nexttoken() {
    int saw_blank = 0;
    int temp;
    if ((*glob.Tokens)[ppindex].Type() == TOK_BLANK)
	saw_blank = 1;

    // if there is a space, skip over multiple spaces
    skipspace();
    
    switch (tokentype(0)) {
      case TOK_LPAREN :
	// we want to register that we saw a blank before '('
	skiptokens(1);
	// skip over lparen
	if (saw_blank) {
	    ppindex--;
	    settype(TOK_LPAREN);
	    ppindex--;
	    settype(TOK_BLANK);
	} else {
	    ppindex--;
	    settype(TOK_LPAREN);
	}
	break;
      case TOK_RPAREN :
	// don't do anything special here -- we'll return TOK_RPAREN
	break;
      case TOK_COLON :
	skiptokens(1);
	ppindex--;
	settype(TOK_COLON);
	break;
      case TOK_AND :
	load(1);
	if (match(TOK_AND)) {
	    skiptokens(2);
	    ppindex--;
	    settype(TOK_XOR);
	} else {
	    skiptokens(1);
	    settype(TOK_AND);
	}
	break;
      case TOK_NOT :
	load(2);
	if (match(TOK_EQUAL, TOK_EQUAL)) {
	    skiptokens(3);
	    ppindex--;
	    settype(TOK_EXACT_NOT_EQUAL);
	} else if (match(TOK_LESS_THAN, TOK_LESS_THAN)) {
	    skiptokens(3);
	    ppindex--;
	    settype(TOK_EXACT_GREATER_EQUAL);
	} else if (match(TOK_GREATER_THAN, TOK_GREATER_THAN)) {
	    skiptokens(3);
	    ppindex--;
	    settype(TOK_EXACT_LESS_EQUAL);
	} else if (match(TOK_EQUAL)) {
	    skiptokens(2);
	    ppindex--;
	    settype(TOK_NOT_EQUAL);
	} else if (match(TOK_LESS_THAN)) {
	    skiptokens(2);
	    ppindex--;
	    settype(TOK_GREATER_EQUAL);
	} else if (match(TOK_GREATER_THAN)) {
	    skiptokens(2);
	    ppindex--;
	    settype(TOK_LESS_EQUAL);
	} else {
	    skiptokens(1);
	    ppindex--;
	    settype(TOK_NOT);
	}
	break;
      case TOK_OR :
	load(1);
	if (match(TOK_OR)) {
	    skiptokens(2);
	    ppindex--;
	    settype(TOK_CONCAT);
	} else {
	    skiptokens(1);
	    ppindex--;
	    settype(TOK_OR);
	}
	break;
      case TOK_LESS_THAN :
	load(2);
	if (match(TOK_LESS_THAN, TOK_EQUAL)) {
	    skiptokens(3);
	    ppindex--;
	    settype(TOK_EXACT_LESS_EQUAL);
	} else if (match(TOK_LESS_THAN)) {
	    skiptokens(2);
	    ppindex--;
	    settype(TOK_EXACT_LESS_THAN);
	} else if (match(TOK_GREATER_THAN)) {
	    skiptokens(2);
	    ppindex--;
	    settype(TOK_NOT_EQUAL);
	} else if (match(TOK_EQUAL)) {
	    skiptokens(2);
	    ppindex--;
	    settype(TOK_LESS_EQUAL);
	} else {
	    skiptokens(1);
	    ppindex--;
	    settype(TOK_LESS_THAN);
	}
	break;
      case TOK_GREATER_THAN :
	load(2);
	if (match(TOK_GREATER_THAN, TOK_EQUAL)) {
	    skiptokens(3);
	    ppindex--;
	    settype(TOK_EXACT_GREATER_EQUAL);
	} else if (match(TOK_GREATER_THAN)) {
	    skiptokens(2);
	    ppindex--;
	    settype(TOK_EXACT_GREATER_THAN);
	} else if (match(TOK_EQUAL)) {
	    skiptokens(2);
	    ppindex--;
	    settype(TOK_GREATER_EQUAL);
	} else if (match(TOK_LESS_THAN)) {
	    skiptokens(2);
	    ppindex--;
	    settype(TOK_NOT_EQUAL);
	} else {
	    skiptokens(1);
	    ppindex--;
	    settype(TOK_GREATER_THAN);
	}
	break;
      case TOK_EQUAL :
	load(1);
	if (match(TOK_EQUAL)) {
	    skiptokens(2);
	    ppindex--;
	    settype(TOK_EXACT_EQUAL);
	} else {
	    skiptokens(1);
	    ppindex--;
	    settype(TOK_EQUAL);
	}
	break;
      case TOK_PLUS :
	skiptokens(1);
	ppindex--;
	settype(TOK_PLUS);
	break;
      case TOK_INT_DIV :
	skiptokens(1);
	ppindex--;
	settype(TOK_INT_DIV);
	break;
      case TOK_MINUS :
	skiptokens(1);
	ppindex--;
	settype(TOK_MINUS);
	break;
      case TOK_REAL_DIV :
	load(1);
	if (match(TOK_REAL_DIV)) {
	    skiptokens(2);
	    ppindex--;
	    settype(TOK_REMAINDER);
	} else {
	    skiptokens(1);
	    ppindex--;
	    settype(TOK_REAL_DIV);
	}
	break;
      case TOK_MULT :
	load(1);
	if (match(TOK_MULT)) {
	    skiptokens(2);
	    ppindex--;
	    settype(TOK_EXP);
	} else {
	    skiptokens(1);
	    ppindex--;
	    settype(TOK_MULT);
	}
	break;
      case TOK_COMMA :
	skiptokens(1);
	ppindex--;
	settype(TOK_COMMA);
	break;
      case TOK_NEWLINE :
	// don't do anything
	break;
      default:
	// otherwise, just return the token
	if (saw_blank) {
	    ppindex--;
	}
	break;
    };

    ppindex++;
    return (*glob.Tokens)[ppindex-1];
}

// returns the type of the token i + ppindex (skipping spaces)
int tokentype(int i) {
    int old = ppindex;

    skipspace();
    for (int j = 0; j < i; j++) {
	ppindex++;
	skipspace();
    }
    int type = (*glob.Tokens)[ppindex].Type();
    ppindex = old;
    return type;
}

RexxToken& token(int i) {
    int old = ppindex;

    skipspace();
    for (int j = 0; j < i; j++) {
	ppindex++;
	skipspace();
    }
    RexxToken* tok = &((*glob.Tokens)[ppindex]);
    ppindex = old;
    return *tok;
}

void skiptokens(int i) {
    skipspace();
    for (int j = 0; j < i; j++) {
	ppindex++;
	skipspace();
    }
}

void fixupcommas() {
    int i = 0;
    while ((*glob.Tokens)[i].Type()) {
//	printf("%d\n", (*glob.Tokens)[i].Type());
	if ((*glob.Tokens)[i].Type() == TOK_COMMA) {
	    int save_comma_loc = i;
	    i++; // skip over TOK_COMMA
	    while ((*glob.Tokens)[i].Type() == TOK_BLANK) i++; // skip over spaces
	    if ((*glob.Tokens)[i].Type() == TOK_NEWLINE &&
		(*glob.Tokens)[i].Secondary() > 0) {
		(*glob.Tokens)[i].Set(TOK_BLANK, 0, ""); // convert newline into tok_blank
		(*glob.Tokens)[save_comma_loc].Set(TOK_BLANK, 0, ""); // convert comma into tok_blank
	    } else {
		i--;
	    }
	}
	i++;
    }
}

static int isakeyword(int type) {
    switch (type) {
      case TOK_ADDRESS :
      case TOK_VALUE :
      case TOK_ARG :
      case TOK_BREAK :
      case TOK_CALL :

      case TOK_ON :
      case TOK_OFF :
      case TOK_NAME :

      case TOK_FAILURE :
      case TOK_ERROR :
      case TOK_HALT :
      case TOK_NOTREADY :

      case TOK_DO :
      case TOK_TO :
      case TOK_BY :
      case TOK_FOR :
      case TOK_FOREVER :
      case TOK_UNTIL :
      case TOK_WHILE :

      case TOK_DROP :
      case TOK_SAY :
      case TOK_ELSE :
      case TOK_END :
      case TOK_EXIT :
      case TOK_IF :
      case TOK_WHEN :
      case TOK_THEN :
      case TOK_INTERPRET :
      case TOK_ITERATE :
      case TOK_LEAVE :
      case TOK_NOP :
      case TOK_NUMERIC :

      case TOK_DIGITS :
      case TOK_FUZZ :
      case TOK_FORM :
      case TOK_SCIENTIFIC :
      case TOK_ENGINEERING :

      case TOK_OPTIONS :
      case TOK_OTHERWISE :
      case TOK_PARSE :
	
      case TOK_LINEIN :
      case TOK_EXTERNAL :
      case TOK_SOURCE :
      case TOK_VAR :
      case TOK_VERSION :
      case TOK_WITH :
      case TOK_UPPER :

      case TOK_PROCEDURE :
      case TOK_EXPOSE :

      case TOK_PULL :
      case TOK_PUSH :
      case TOK_QUEUE :
      case TOK_RETURN :
      case TOK_SELECT :

      case TOK_SIGNAL :
      case TOK_SYNTAX :
      case TOK_NOVALUE :

      case TOK_TRACE :
	return 1;
      default:
	return 0;
    };
}

static int isasymbol(int type) {
    return (type == TOK_SYMBOL || type == TOK_FIXED || isakeyword(type));
}

static int isaninstruction(int type) {
    switch (type) {
      case TOK_ADDRESS :
      case TOK_ARG :
      case TOK_BREAK :
      case TOK_CALL :
      case TOK_DO :
      case TOK_DROP :
      case TOK_SAY :
      case TOK_ELSE :
      case TOK_END :
      case TOK_EXIT :
      case TOK_IF :
      case TOK_WHEN :
      case TOK_THEN :
      case TOK_INTERPRET :
      case TOK_ITERATE :
      case TOK_LEAVE :
      case TOK_NOP :
      case TOK_NUMERIC :
      case TOK_OPTIONS :
      case TOK_OTHERWISE :
      case TOK_PARSE :
      case TOK_PROCEDURE :
      case TOK_PULL :
      case TOK_PUSH :
      case TOK_QUEUE :
      case TOK_RETURN :
      case TOK_SELECT :
      case TOK_SIGNAL :
      case TOK_TRACE :
	return 1;
      default:
	return 0;
    };
}

static int isalabel(int type1, int type2) {
    return (isasymbol(type1) && (type2 == TOK_COLON));
}

static int isanassignment(int type1, int type2, int type3) {
    return ((isasymbol(type1) && type1 != TOK_FIXED) && (type2 == TOK_EQUAL) && (type3 != TOK_EQUAL));
}

void processasexpression() {
    RexxToken* tok;
    tok = &(nexttoken());
    while (tok->Type() != TOK_NEWLINE) {
	if (isakeyword(tok->Type())) {
	    addtoken(TOK_SYMBOL, TOK_SIMPLE, *tok->Value());
	} else {
	    addtoken(tok->Type(), tok->Secondary(), *tok->Value());
	}
	tok = &(nexttoken());
    }
    addtoken(tok->Type(), tok->Secondary(), *tok->Value());
}

void processasifexpression() {
    // special version of copyexpression, looking for tok_then
    RexxToken* tok;
    tok = &(nexttoken());
    while (tok->Type() != TOK_NEWLINE) {
	if (isakeyword(tok->Type())) {
	    if (tok->Type() == TOK_THEN) {
		backupifblank();
		addtoken(TOK_NEWLINE);
		addtoken(TOK_THEN);
		addtoken(TOK_NEWLINE);
		skipspace();
		return;
	    }
	    addtoken(TOK_SYMBOL, TOK_SIMPLE, *tok->Value());
	} else {
	    addtoken(tok->Type(), tok->Secondary(), *tok->Value());
	}
	tok = &(nexttoken());
    }
    addtoken(tok->Type(), tok->Secondary(), *tok->Value());
}

void processasparseexpression() {
    // special version of copyexpression, looking for tok_with
    RexxToken* tok;
    tok = &(nexttoken());
    while (tok->Type() != TOK_NEWLINE) {
	if (isakeyword(tok->Type())) {
	    if (tok->Type() == TOK_WITH) {
		backupifblank();
		addtoken(TOK_WITH);
		skipspace();
	    } else {
		addtoken(TOK_SYMBOL, TOK_SIMPLE, *tok->Value());
	    }
	} else {
	    addtoken(tok->Type(), tok->Secondary(), *tok->Value());
	}
	tok = &(nexttoken());
    }
    addtoken(tok->Type(), tok->Secondary(), *tok->Value());
}

void processasdoexpression() {
    // special version of copyexpression, looking for TO, BY, FOR, FOREVER, WHILE, UNTIL
    RexxToken* tok;
    tok = &(nexttoken());
    while (tok->Type() != TOK_NEWLINE) {
	if (isakeyword(tok->Type())) {
	    if (tok->Type() == TOK_TO) {
		backupifblank();
		addtoken(TOK_TO);
		skipspace();
	    } else if (tok->Type() == TOK_BY) {
		backupifblank();
		addtoken(TOK_BY);
		skipspace();
	    } else if (tok->Type() == TOK_FOR) {
		backupifblank();
		addtoken(TOK_FOR);
		skipspace();
	    } else if (tok->Type() == TOK_FOREVER) {
		backupifblank();
		addtoken(TOK_FOREVER);
		skipspace();
	    } else if (tok->Type() == TOK_WHILE) {
		backupifblank();
		addtoken(TOK_WHILE);
		skipspace();
	    } else if (tok->Type() == TOK_UNTIL) {
		backupifblank();
		addtoken(TOK_UNTIL);
		skipspace();
	    } else {
		addtoken(TOK_SYMBOL, TOK_SIMPLE, *tok->Value());
	    }
	} else {
	    addtoken(tok->Type(), tok->Secondary(), *tok->Value());
	}
	tok = &(nexttoken());
    }
    addtoken(tok->Type(), tok->Secondary(), *tok->Value());
}

static int mt1, mt2, mt3, mt4;

void load(int n) {
    switch (n) {
      case 4 :
	mt4 = tokentype(4);
      case 3 :
	mt3 = tokentype(3);
      case 2 :
	mt2 = tokentype(2);
      case 1 :
	mt1 = tokentype(1);
    };
}

int match(int t1) {
    return mt1 == t1;
}

int match(int t1, int t2) {
    return mt1 == t1 && mt2 == t2;
}

int match(int t1, int t2, int t3) {
    return mt1 == t1 && mt2 == t2 && mt3 == t3;
}

int match(int t1, int t2, int t3, int t4) {
    return mt1 == t1 && mt2 == t2 && mt3 == t3 &&
	mt4 == t4;
}


//  always call load() before doing match!
void processasinstruction(int instruction) {
    switch (instruction) {
      case TOK_ADDRESS :
	load(1);
	if (match(TOK_VALUE)) {
	    addtoken(TOK_ADDRESS_VALUE);
	} else if (match(TOK_STRING_LITERAL) ||
	    isasymbol(tokentype(1))) {
	    addtoken(TOK_ADDRESS_SS);
	} else {
	    addtoken(TOK_ADDRESS);
	}
	skiptokens(1);
	processasexpression();
	break;
      case TOK_ARG :
	addtoken(TOK_ARG);
	skiptokens(1);
	processasexpression();
	break;
      case TOK_BREAK :
	addtoken(TOK_BREAK);
	skiptokens(1);
	processasexpression();
	break;
      case TOK_CALL :
	load(3);
	if (match(TOK_ON, TOK_ERROR, TOK_NAME)) {
	    addtoken(TOK_CALL_ON_ERROR_NAME);
	    skiptokens(4);
	} else if (match(TOK_ON, TOK_FAILURE, TOK_NAME)) {
	    addtoken(TOK_CALL_ON_FAILURE_NAME);
	    skiptokens(4);
	} else if (match(TOK_ON, TOK_HALT, TOK_NAME)) {
	    addtoken(TOK_CALL_ON_HALT_NAME);
	    skiptokens(4);
	} else if (match(TOK_ON, TOK_NOTREADY, TOK_NAME)) {
	    addtoken(TOK_CALL_ON_NOTREADY_NAME);
	    skiptokens(4);

	} else if (match(TOK_ON, TOK_ERROR)) {
	    addtoken(TOK_CALL_ON_ERROR);
	    skiptokens(3);
	} else if (match(TOK_ON, TOK_FAILURE)) {
	    addtoken(TOK_CALL_ON_FAILURE);
	    skiptokens(3);
	} else if (match(TOK_ON, TOK_HALT)) {
	    addtoken(TOK_CALL_ON_HALT);
	    skiptokens(3);
	} else if (match(TOK_ON, TOK_NOTREADY)) {
	    addtoken(TOK_CALL_ON_NOTREADY);
	    skiptokens(3);

	} else if (match(TOK_OFF, TOK_ERROR)) {
	    addtoken(TOK_CALL_OFF_ERROR);
	    skiptokens(3);
	} else if (match(TOK_OFF, TOK_FAILURE)) {
	    addtoken(TOK_CALL_OFF_FAILURE);
	    skiptokens(3);
	} else if (match(TOK_OFF, TOK_HALT)) {
	    addtoken(TOK_CALL_OFF_HALT);
	    skiptokens(3);
	} else if (match(TOK_OFF, TOK_NOTREADY)) {
	    addtoken(TOK_CALL_OFF_NOTREADY);
	    skiptokens(3);
	} else {
	    addtoken(TOK_CALL);
	    skiptokens(1);
	}
	processasexpression();

	break;
      case TOK_DO :
	addtoken(TOK_DO);
	skiptokens(1);
	processasdoexpression();
	break;
      case TOK_DROP :
	addtoken(TOK_DROP);
	skiptokens(1);
	processasexpression();
	break;
      case TOK_SAY :
	addtoken(TOK_SAY);
	skiptokens(1);
	processasexpression();
	break;
      case TOK_ELSE :
	addtoken(TOK_ELSE);
	skiptokens(1);
	load(1);
	if (!match(TOK_NEWLINE))
	    addtoken(TOK_NEWLINE);
	break;
      case TOK_END :
	addtoken(TOK_END);
	skiptokens(1);
	processasexpression();
	break;
      case TOK_EXIT :
	addtoken(TOK_EXIT);
	skiptokens(1);
	processasexpression();
	break;
      case TOK_IF :
	addtoken(TOK_IF);
	skiptokens(1);
	processasifexpression();
	break;
      case TOK_WHEN :
	addtoken(TOK_WHEN);
	skiptokens(1);
	processasifexpression();
	break;
      case TOK_THEN :
	addtoken(TOK_THEN);
	skiptokens(1);
	load(1);
	if (!match(TOK_NEWLINE))
	    addtoken(TOK_NEWLINE);
	break;
      case TOK_INTERPRET :
	addtoken(TOK_INTERPRET);
	skiptokens(1);
	processasexpression();
	break;
      case TOK_ITERATE :
	addtoken(TOK_ITERATE);
	skiptokens(1);
	processasexpression();
	break;
      case TOK_LEAVE :
	addtoken(TOK_LEAVE);
	skiptokens(1);
	processasexpression();
	break;
      case TOK_NOP :
	addtoken(TOK_NOP);
	skiptokens(1);
	processasexpression();
	break;
      case TOK_NUMERIC :
	load(2);
	if (match(TOK_FORM, TOK_SCIENTIFIC)) {
	    addtoken(TOK_NUMERIC_FORM_SCIENTIFIC);
	    skiptokens(3);
	} else if (match(TOK_FORM, TOK_ENGINEERING)) {
	    addtoken(TOK_NUMERIC_FORM_ENGINEERING);
	    skiptokens(3);
	} else if (match(TOK_FORM, TOK_VALUE)) {
	    addtoken(TOK_NUMERIC_FORM_VALUE);
	    skiptokens(3);
	} else if (match(TOK_FORM)) {
	    addtoken(TOK_NUMERIC_FORM_VALUE);
	    skiptokens(2);
	} else if (match(TOK_DIGITS)) {
	    addtoken(TOK_NUMERIC_DIGITS);
	    skiptokens(2);
	} else if (match(TOK_FUZZ)) {
	    addtoken(TOK_NUMERIC_FUZZ);
	    skiptokens(2);
	} else {
	    addtoken(TOK_NUMERIC);
	}
	processasexpression();
	break;
      case TOK_OPTIONS :
	addtoken(TOK_OPTIONS);
	skiptokens(1);
	processasexpression();
	break;
      case TOK_OTHERWISE :
	addtoken(TOK_OTHERWISE);
	skiptokens(1);
	load(1);
	if (!match(TOK_NEWLINE))
	    addtoken(TOK_NEWLINE);
	break;
      case TOK_PARSE :
	load(2);
	if (match(TOK_UPPER, TOK_ARG)) {
	    addtoken(TOK_PARSE_UPPER_ARG);
	    skiptokens(3);
	} else if (match(TOK_UPPER, TOK_LINEIN)) {
	    addtoken(TOK_PARSE_UPPER_LINEIN);
	    skiptokens(3);
	} else if (match(TOK_UPPER, TOK_EXTERNAL)) {
	    addtoken(TOK_PARSE_UPPER_EXTERNAL);
	    skiptokens(3);
	} else if (match(TOK_UPPER, TOK_PULL)) {
	    addtoken(TOK_PARSE_UPPER_PULL);
	    skiptokens(3);
	} else if (match(TOK_UPPER, TOK_SOURCE)) {
	    addtoken(TOK_PARSE_UPPER_SOURCE);
	    skiptokens(3);
	} else if (match(TOK_UPPER, TOK_VALUE)) {
	    addtoken(TOK_PARSE_UPPER_VALUE);
	    skiptokens(3);
	    processasparseexpression();
	    break;
	} else if (match(TOK_UPPER, TOK_VAR)) {
	    addtoken(TOK_PARSE_UPPER_VAR);
	    skiptokens(3);
	} else if (match(TOK_UPPER, TOK_VERSION)) {
	    addtoken(TOK_PARSE_UPPER_VERSION);
	    skiptokens(3);
	} else if (match(TOK_ARG)) {
	    addtoken(TOK_PARSE_ARG);
	    skiptokens(2);
	} else if (match(TOK_LINEIN)) {
	    addtoken(TOK_PARSE_LINEIN);
	    skiptokens(2);
	} else if (match(TOK_EXTERNAL)) {
	    addtoken(TOK_PARSE_EXTERNAL);
	    skiptokens(2);
	} else if (match(TOK_PULL)) {
	    addtoken(TOK_PARSE_PULL);
	    skiptokens(2);
	} else if (match(TOK_SOURCE)) {
	    addtoken(TOK_PARSE_SOURCE);
	    skiptokens(2);
	} else if (match(TOK_VALUE)) {
	    addtoken(TOK_PARSE_VALUE);
	    skiptokens(2);
	    processasparseexpression();
	    break;
	} else if (match(TOK_VAR)) {
	    addtoken(TOK_PARSE_VAR);
	    skiptokens(2);
	} else if (match(TOK_VERSION)) {
	    addtoken(TOK_PARSE_VERSION);
	    skiptokens(2);
	} else {
	    addtoken(TOK_PARSE);
	    skiptokens(1);
	}
	processasexpression();
	break;
      case TOK_PROCEDURE :
	load(1);
	if (match(TOK_EXPOSE)) {
	    addtoken(TOK_PROCEDURE_EXPOSE);
	    skiptokens(2);
	} else {
	    addtoken(TOK_PROCEDURE);
	    skiptokens(1);
	}
	processasexpression();
	break;
      case TOK_PULL :
	addtoken(TOK_PULL);
	skiptokens(1);
	processasexpression();
	break;
      case TOK_PUSH :
	addtoken(TOK_PUSH);
	skiptokens(1);
	processasexpression();
	break;
      case TOK_QUEUE :
	addtoken(TOK_QUEUE);
	skiptokens(1);
	processasexpression();
	break;
      case TOK_RETURN :
	addtoken(TOK_RETURN);
	skiptokens(1);
	processasexpression();
	break;
      case TOK_SELECT :
	addtoken(TOK_SELECT);
	skiptokens(1);
	processasexpression();
	break;
      case TOK_SIGNAL :
	load(3);
	if (match(TOK_ON, TOK_ERROR, TOK_NAME)) {
	    addtoken(TOK_SIGNAL_ON_ERROR_NAME);
	    skiptokens(4);
	} else if (match(TOK_ON, TOK_FAILURE, TOK_NAME)) {
	    addtoken(TOK_SIGNAL_ON_FAILURE_NAME);
	    skiptokens(4);
	} else if (match(TOK_ON, TOK_HALT, TOK_NAME)) {
	    addtoken(TOK_SIGNAL_ON_HALT_NAME);
	    skiptokens(4);
	} else if (match(TOK_ON, TOK_NOTREADY, TOK_NAME)) {
	    addtoken(TOK_SIGNAL_ON_NOTREADY_NAME);
	    skiptokens(4);
	} else if (match(TOK_ON, TOK_NOVALUE, TOK_NAME)) {
	    addtoken(TOK_SIGNAL_ON_NOVALUE_NAME);
	    skiptokens(4);
	} else if (match(TOK_ON, TOK_SYNTAX, TOK_NAME)) {
	    addtoken(TOK_SIGNAL_ON_SYNTAX_NAME);
	    skiptokens(4);

	} else if (match(TOK_ON, TOK_ERROR)) {
	    addtoken(TOK_SIGNAL_ON_ERROR);
	    skiptokens(3);
	} else if (match(TOK_ON, TOK_FAILURE)) {
	    addtoken(TOK_SIGNAL_ON_FAILURE);
	    skiptokens(3);
	} else if (match(TOK_ON, TOK_HALT)) {
	    addtoken(TOK_SIGNAL_ON_HALT);
	    skiptokens(3);
	} else if (match(TOK_ON, TOK_NOTREADY)) {
	    addtoken(TOK_SIGNAL_ON_NOTREADY);
	    skiptokens(3);
	} else if (match(TOK_ON, TOK_NOVALUE)) {
	    addtoken(TOK_SIGNAL_ON_NOVALUE);
	    skiptokens(3);
	} else if (match(TOK_ON, TOK_SYNTAX)) {
	    addtoken(TOK_SIGNAL_ON_SYNTAX);
	    skiptokens(3);

	} else if (match(TOK_OFF, TOK_ERROR)) {
	    addtoken(TOK_SIGNAL_OFF_ERROR);
	    skiptokens(3);
	} else if (match(TOK_OFF, TOK_FAILURE)) {
	    addtoken(TOK_SIGNAL_OFF_FAILURE);
	    skiptokens(3);
	} else if (match(TOK_OFF, TOK_HALT)) {
	    addtoken(TOK_SIGNAL_OFF_HALT);
	    skiptokens(3);
	} else if (match(TOK_OFF, TOK_NOTREADY)) {
	    addtoken(TOK_SIGNAL_OFF_NOTREADY);
	    skiptokens(3);
	} else if (match(TOK_OFF, TOK_NOVALUE)) {
	    addtoken(TOK_SIGNAL_OFF_NOVALUE);
	    skiptokens(3);
	} else if (match(TOK_OFF, TOK_SYNTAX)) {
	    addtoken(TOK_SIGNAL_OFF_SYNTAX);
	    skiptokens(3);

	} else if (match(TOK_VALUE)) {
	    addtoken(TOK_SIGNAL_VALUE);
	    skiptokens(2);
	} else {
	    addtoken(TOK_SIGNAL);
	    skiptokens(1);
	}
	processasexpression();
	break;
      case TOK_TRACE :
	load(1);
	if (match(TOK_VALUE)) {
	    addtoken(TOK_TRACE_VALUE);
	} else if (match(TOK_STRING_LITERAL) ||
	    isasymbol(tokentype(1))) {
	    addtoken(TOK_TRACE_SS);
	} else {
	    addtoken(TOK_TRACE_VALUE);
	}
	skiptokens(1);
	processasexpression();
	break;
    };
}

void rexx_postprocess() {
    fixupcommas(); // fixes up commas followed by real newlines into space
    ppindex = 0;
    while ((*glob.Tokens)[ppindex].Type()) {
	skipspace();
	if (isalabel(tokentype(0), tokentype(1))) {
	    RexxToken* a = &(token(0));
	    addtoken(TOK_LABEL, a->Secondary(), *a->Value());
	    // only add a newline if one is not there already
	    load(1);
	    if (!match(TOK_NEWLINE))
		addtoken(TOK_NEWLINE);
	    skiptokens(2);
	} else if (isanassignment(tokentype(0), tokentype(1), tokentype(2))) {
	    RexxToken* a = &(token(0));
	    addtoken(TOK_ASSIGNMENT, a->Secondary(), *a->Value());
	    addtoken(TOK_EQUAL);
	    skiptokens(2);
	    processasexpression();
	} else if (isaninstruction(tokentype(0))) {
	    processasinstruction(tokentype(0));
	} else {
	    // process everything as an expression
	    processasexpression();
	}
    }
    addtoken(0);
}
