%{
#define YYDEBUG_LEXER_TEXT (yytext) /* our lexer loads this up each time.
				     We are telling the graphical debugger
				     where to find the spelling of the 
				     tokens.*/

/* interface with flex: should be in header file */

#include <stdio.h>
#include <stdlib.h>
//#include <stream.h>
#include "Clause.h"
#include "Express.h"
#include "Expressionp.SLList.h"
#include "InternalData.h"

extern Clause* clause;
extern FILE*	yyin;
extern int	yyerror(char*);
extern int      yylex(void);
int		yyeof = 0;
//int parsererror = 0;

extern char* yytext;
%}
%left	TOK_OR			TOK_XOR

%left	TOK_AND			

%left	TOK_EQUAL		TOK_GREATER_EQUAL	TOK_GREATER_THAN
	TOK_LESS_EQUAL		TOK_LESS_THAN
        TOK_NOT_EQUAL
	TOK_EXACT_EQUAL         TOK_EXACT_NOT_EQUAL
    	TOK_EXACT_GREATER_EQUAL TOK_EXACT_GREATER_THAN
	TOK_EXACT_LESS_EQUAL    TOK_EXACT_LESS_THAN

%left   TOK_BLANK               TOK_CONCAT

%left	TOK_MINUS		TOK_PLUS

%left	TOK_INT_DIV		TOK_REMAINDER		TOK_REAL_DIV
	TOK_MULT

%left   NEGATIVE                POSITIVE

%right	TOK_EXP			

%left	TOK_NOT

%token  TOK_COMMA               TOK_LPAREN
%token  TOK_RPAREN              TOK_NEWLINE

%token <symbol> TOK_STRING_LITERAL

%token <symbol> TOK_LABEL
%token TOK_COLON

%token <symbol> TOK_SYMBOL

%token  <symbol> TOK_FIXED

%token  TOK_STEM
%token  TOK_COMPOUND
%token  TOK_SIMPLE

%token <symbol> TOK_ASSIGNMENT

%token TOK_NULLCLAUSE
%token TOK_ERRORCLAUSE
%token TOK_COMMANDCLAUSE

    TOK_LINENUMBER

    TOK_ADDRESS
    TOK_ADDRESS_SS
    TOK_ADDRESS_VALUE

    TOK_VALUE

    TOK_ARG

    TOK_BREAK

    TOK_CALL

    TOK_ON
    TOK_OFF
    TOK_FAILURE
    TOK_ERROR
    TOK_HALT
    TOK_NOTREADY

    TOK_CALL_ON_FAILURE
    TOK_CALL_ON_ERROR
    TOK_CALL_ON_HALT
    TOK_CALL_ON_NOTREADY

    TOK_CALL_OFF_FAILURE
    TOK_CALL_OFF_ERROR
    TOK_CALL_OFF_HALT
    TOK_CALL_OFF_NOTREADY

    TOK_NAME

    TOK_CALL_ON_FAILURE_NAME
    TOK_CALL_ON_ERROR_NAME
    TOK_CALL_ON_HALT_NAME
    TOK_CALL_ON_NOTREADY_NAME


    TOK_DO
    TOK_TO
    TOK_BY
    TOK_FOR
    TOK_FOREVER
    TOK_WHILE
    TOK_UNTIL

    TOK_DROP

    TOK_SAY
    
    TOK_ELSE

    TOK_END

    TOK_EXIT
    TOK_EOF

    TOK_IF
    TOK_WHEN
    TOK_THEN

    TOK_INTERPRET
    
    TOK_ITERATE

    TOK_LEAVE

    TOK_NOP

    TOK_DIGITS
    TOK_FUZZ
    TOK_FORM
    TOK_SCIENTIFIC
    TOK_ENGINEERING

    TOK_NUMERIC
    TOK_NUMERIC_DIGITS
    TOK_NUMERIC_FUZZ
    TOK_NUMERIC_FORM_SCIENTIFIC
    TOK_NUMERIC_FORM_ENGINEERING
    TOK_NUMERIC_FORM_VALUE

    TOK_OPTIONS

    TOK_OTHERWISE

    TOK_LINEIN
    TOK_EXTERNAL
    TOK_SOURCE
    TOK_VALUE

    TOK_VAR
    TOK_VERSION

    TOK_PARSE
    TOK_PARSE_ARG
    TOK_PARSE_LINEIN
    TOK_PARSE_EXTERNAL
    TOK_PARSE_PULL
    TOK_PARSE_SOURCE
    TOK_PARSE_VALUE
    TOK_WITH
    TOK_PARSE_VAR
    TOK_PARSE_VERSION
    TOK_PARSE_UPPER_ARG
    TOK_PARSE_UPPER_LINEIN
    TOK_PARSE_UPPER_EXTERNAL
    TOK_PARSE_UPPER_NUMERIC
    TOK_PARSE_UPPER_PULL
    TOK_PARSE_UPPER_SOURCE
    TOK_PARSE_UPPER_VALUE
    TOK_PARSE_UPPER_VAR
    TOK_PARSE_UPPER_VERSION

    TOK_EXPOSE

    TOK_PROCEDURE
    TOK_PROCEDURE_EXPOSE

    TOK_PULL

    TOK_PUSH

    TOK_QUEUE
    
    TOK_RETURN

    TOK_SELECT


    TOK_SYNTAX
    TOK_NOVALUE

    TOK_SIGNAL
    TOK_SIGNAL_ON_FAILURE
    TOK_SIGNAL_ON_ERROR
    TOK_SIGNAL_ON_HALT
    TOK_SIGNAL_ON_NOTREADY
    TOK_SIGNAL_ON_NOVALUE
    TOK_SIGNAL_ON_SYNTAX

    TOK_SIGNAL_OFF_FAILURE
    TOK_SIGNAL_OFF_ERROR
    TOK_SIGNAL_OFF_HALT
    TOK_SIGNAL_OFF_NOTREADY
    TOK_SIGNAL_OFF_NOVALUE
    TOK_SIGNAL_OFF_SYNTAX

    TOK_SIGNAL_ON_FAILURE_NAME
    TOK_SIGNAL_ON_ERROR_NAME
    TOK_SIGNAL_ON_HALT_NAME
    TOK_SIGNAL_ON_NOTREADY_NAME
    TOK_SIGNAL_ON_NOVALUE_NAME
    TOK_SIGNAL_ON_SYNTAX_NAME

    TOK_SIGNAL_VALUE

    TOK_TRACE
    TOK_TRACE_SS
    TOK_TRACE_VALUE

    TOK_UPPER

    TOK_UNKNOWN

%type <expression> expression_any any_function_call

%type <expressionlist> optional_expression_list

%type <symbol> symbol_or_string

%start clause

%union{
    struct {
	ArgString* string;
	int secondary;
    } symbol;
    ArgString* string;
    Expression* expression;
    ExpressionpSLList* expressionlist;
}

%%

clause : /* eof */ {clause = 0; YYACCEPT;}
       | error TOK_NEWLINE {
		if (clause) {
			delete clause;
		}
		clause = new ErrorClause;
		YYACCEPT;
	}
       | null_clause TOK_NEWLINE {clause = new NullClause; YYACCEPT;}
       | label_clause TOK_NEWLINE {YYACCEPT;}
       | assignment_clause TOK_NEWLINE {YYACCEPT;}
       | instruction_clause TOK_NEWLINE {YYACCEPT;}
       | command_clause TOK_NEWLINE {YYACCEPT;}
    ;


/*
obl :
   | TOK_BLANK
    ;
*/

bl : TOK_BLANK
    ;

null_clause :
    ;

label_clause : TOK_LABEL {clause = new LabelClause(*($1.string));}
    ;

assignment_clause : TOK_ASSIGNMENT TOK_EQUAL expression_any {
    clause = new AssignmentClause(*($1.string), $1.secondary, $3);
    }
    ;

command_clause : expression_any {clause = new CommandClause; ((CommandClause*)clause)->SetExpression($1);}
    ;

instruction_clause : {clause = new AddressClause;} address_instruction
                   | {clause = new ParseArgClause;} arg_instruction
                   | {clause = new BreakClause;} break_instruction
                   | {clause = new CallClause;} call_instruction
                   | {clause = new DoClause;} do_instruction
                   | {clause = new DropClause;} drop_instruction
                   | {clause = new ElseClause;} else_instruction
                   | {clause = new EndClause;} end_instruction
                   | {clause = new ExitClause;} exit_instruction
                   | {clause = new IfClause;} if_instruction
                   | {clause = new InterpretClause;} interpret_instruction
                   | {clause = new IterateClause;} iterate_instruction
                   | {clause = new LeaveClause;} leave_instruction
                   | {clause = new NopClause;} nop_instruction
                   | {clause = new NumericClause;} numeric_instruction
                   | {clause = new OptionsClause;} options_instruction
                   | {clause = new OtherwiseClause; }otherwise_instruction
                   | {/* see parse_instruction */} parse_instruction
                   | {clause = new ProcedureClause;} procedure_instruction
                   | {clause = new ParsePullClause;} pull_instruction
                   | {clause = new PushClause;} push_instruction
                   | {clause = new QueueClause;} queue_instruction
                   | {clause = new ReturnClause;} return_instruction
                   | {clause = new SayClause;} say_instruction
                   | {clause = new SelectClause;} select_instruction
                   | {clause = new SignalClause;} signal_instruction
                   | {clause = new ThenClause;} then_instruction
                   | {clause = new TraceClause;} trace_instruction
                   | {clause = new WhenClause;} when_instruction
    ;

/*
conditional_clause : assignment_clause
    | instruction_clause
    | command_clause
    ;
*/


/* Address instruction */
address_instruction : TOK_ADDRESS
                    | TOK_ADDRESS_SS symbol_or_string {((AddressClause*)clause)->SetExpression(new LiteralExpression(*($2.string)));}
                    | TOK_ADDRESS_SS symbol_or_string expression_any {((AddressClause*)clause)->SetEnvironment(*($2.string)); ((AddressClause*)clause)->SetExpression($3); ((AddressClause*)clause)->SetCommand();}
                    | TOK_ADDRESS_SS symbol_or_string bl expression_any {((AddressClause*)clause)->SetEnvironment(*($2.string)); ((AddressClause*)clause)->SetExpression($4); ((AddressClause*)clause)->SetCommand();}
                    | TOK_ADDRESS_VALUE expression_any {((AddressClause*)clause)->SetExpression($2);}
                    | TOK_ADDRESS expression_any {((AddressClause*)clause)->SetExpression($2);}
    ;

symbol_or_string : TOK_SYMBOL {$$ = $1;}
    | TOK_STRING_LITERAL {$$ = $1;}
    ;

/* arg instruction */
arg_instruction : TOK_ARG optional_template_list
    ;

optional_template_list : /* null */
        | TOK_BLANK
	| template_list
        | TOK_BLANK template_list
	| optional_template_list TOK_COMMA {((ParseClause*)clause)->NextTemplate();} template_list
	;

obl : /* null */
	| TOK_BLANK
	;

template_list : TOK_SYMBOL {
	((ParseClause*)clause)->AddPattern(*($1.string), PVariable);
	}
    | TOK_STRING_LITERAL {
	ArgString as = (*($1.string));
	((ParseClause*)clause)->AddPattern(as, PString);
	}
    | TOK_FIXED {
	ArgString as = (*($1.string));
	String s = as.String();
	if (s == ".")
		((ParseClause*)clause)->AddPattern(as, PPeriod);
	else
		((ParseClause*)clause)->AddPattern(as, PAbsoluteInteger);
	} 
    | TOK_EQUAL TOK_FIXED {
	ArgString as = (*($2.string));
	((ParseClause*)clause)->AddPattern(as, PAbsoluteInteger);
	} 
    | TOK_PLUS TOK_FIXED {
	ArgString as = (*($2.string));
	((ParseClause*)clause)->AddPattern(as, PPositiveInteger);
	} 
    | TOK_MINUS TOK_FIXED {
	ArgString as = (*($2.string));
	((ParseClause*)clause)->AddPattern(as, PNegativeInteger);
	}
    | TOK_LPAREN TOK_SYMBOL TOK_RPAREN {
	((ParseClause*)clause)->AddPattern(*($2.string), PStringReference);
	}
    | TOK_EQUAL TOK_LPAREN TOK_SYMBOL TOK_RPAREN {
	((ParseClause*)clause)->AddPattern(*($3.string), PAbsoluteReference);
	}
    | TOK_PLUS TOK_LPAREN TOK_SYMBOL TOK_RPAREN {
	((ParseClause*)clause)->AddPattern(*($3.string), PPositiveReference);
	}
    | TOK_MINUS TOK_LPAREN TOK_SYMBOL TOK_RPAREN {
	((ParseClause*)clause)->AddPattern(*($3.string), PNegativeReference);
	}
    | template_list obl TOK_SYMBOL {
	((ParseClause*)clause)->AddPattern(*($3.string), PVariable);
	}
    | template_list obl TOK_STRING_LITERAL {
	ArgString as = (*($3.string));
	((ParseClause*)clause)->AddPattern(as, PString);
	}
    | template_list obl TOK_FIXED {
	ArgString as = (*($3.string));
	String s = as.String();
	if (s == ".")
		((ParseClause*)clause)->AddPattern(as, PPeriod);
	else
		((ParseClause*)clause)->AddPattern(as, PAbsoluteInteger);
	} 
    | template_list obl TOK_EQUAL TOK_FIXED {
	ArgString as = (*($4.string));
	((ParseClause*)clause)->AddPattern(as, PAbsoluteInteger);
	}
    | template_list obl TOK_PLUS TOK_FIXED {
	ArgString as = (*($4.string));
	((ParseClause*)clause)->AddPattern(as, PPositiveInteger);
	}
    | template_list obl TOK_MINUS TOK_FIXED {
	ArgString as = (*($4.string));
	((ParseClause*)clause)->AddPattern(as, PNegativeInteger);
	}
    | template_list obl TOK_LPAREN TOK_SYMBOL TOK_RPAREN {
	((ParseClause*)clause)->AddPattern(*($4.string), PStringReference);
	}
    | template_list obl TOK_EQUAL TOK_LPAREN TOK_SYMBOL TOK_RPAREN {
	((ParseClause*)clause)->AddPattern(*($5.string), PAbsoluteReference);
	}
    | template_list obl TOK_PLUS TOK_LPAREN TOK_SYMBOL TOK_RPAREN {
	((ParseClause*)clause)->AddPattern(*($5.string), PPositiveReference);
	}
    | template_list obl TOK_MINUS TOK_LPAREN TOK_SYMBOL TOK_RPAREN {
	((ParseClause*)clause)->AddPattern(*($5.string), PNegativeReference);
	}
    ;

/* Break instruction */
break_instruction : TOK_BREAK
    ;

/* call instruction */
call_instruction : TOK_CALL TOK_SYMBOL optional_expression_list {
	((CallClause*)clause)->SetExpression(new FunctionExpression(*($2.string), $3));
	}
    | TOK_CALL TOK_SYMBOL TOK_BLANK optional_expression_list {
	((CallClause*)clause)->SetExpression(new FunctionExpression(*($2.string), $4));
	}
    | TOK_CALL TOK_STRING_LITERAL optional_expression_list {
	((CallClause*)clause)->SetExpression(new FunctionExpression(*($2.string), $3, 1));
	}
    | TOK_CALL TOK_STRING_LITERAL TOK_BLANK optional_expression_list {
	((CallClause*)clause)->SetExpression(new FunctionExpression(*($2.string), $4, 1));
	}
    | TOK_CALL TOK_SYMBOL TOK_LPAREN optional_expression_list TOK_RPAREN {
	((CallClause*)clause)->SetExpression(new FunctionExpression(*($2.string), $4));
	}
    | TOK_CALL TOK_STRING_LITERAL TOK_LPAREN optional_expression_list TOK_RPAREN {
	((CallClause*)clause)->SetExpression(new FunctionExpression(*($2.string), $4, 1));
	}
    | TOK_CALL_ON_FAILURE {
	((CallClause*)clause)->TurnOn(SigFailure);
	}
    | TOK_CALL_ON_ERROR {
	((CallClause*)clause)->TurnOn(SigError);
	}
    | TOK_CALL_ON_HALT {
	((CallClause*)clause)->TurnOn(SigHalt);
	}
    | TOK_CALL_ON_NOTREADY {
	((CallClause*)clause)->TurnOn(SigNotReady);
	}
    | TOK_CALL_OFF_FAILURE {
	((CallClause*)clause)->TurnOff(SigFailure);
	}
    | TOK_CALL_OFF_ERROR {
	((CallClause*)clause)->TurnOff(SigError);
	}
    | TOK_CALL_OFF_HALT {
	((CallClause*)clause)->TurnOff(SigHalt);
	}
    | TOK_CALL_OFF_NOTREADY {
	((CallClause*)clause)->TurnOff(SigNotReady);
	}
    | TOK_CALL_ON_FAILURE_NAME TOK_SYMBOL {
	((CallClause*)clause)->TurnOn(SigFailure, *($2.string));
	}
    | TOK_CALL_ON_ERROR_NAME TOK_SYMBOL {
	((CallClause*)clause)->TurnOn(SigError, *($2.string));
	}
    | TOK_CALL_ON_HALT_NAME TOK_SYMBOL {
	((CallClause*)clause)->TurnOn(SigHalt, *($2.string));
	}
    | TOK_CALL_ON_NOTREADY_NAME TOK_SYMBOL {
	((CallClause*)clause)->TurnOn(SigNotReady, *($2.string));
	}
    ;

/* do instruction */
do_instruction : TOK_DO
    | TOK_DO do_repetitor
    | TOK_DO do_conditional
    | TOK_DO do_repetitor do_conditional
    ;

do_repetitor : TOK_FOREVER {((DoClause*)clause)->SetForever();}
    | TOK_SYMBOL TOK_EQUAL expression_any do_opt_to do_opt_by do_opt_for {((DoClause*)clause)->SetSymbol(*($1.string), $1.secondary); ((DoClause*)clause)->SetFrom($3);}
    | expression_any {((DoClause*)clause)->SetFor($1);}
    ;

do_opt_to : /* optional */
    | TOK_TO expression_any {((DoClause*)clause)->SetTo($2);}
    ;

do_opt_by : /* optional */
    | TOK_BY expression_any {((DoClause*)clause)->SetBy($2);}
    ;

do_opt_for : /* optional */
    | TOK_FOR expression_any {((DoClause*)clause)->SetFor($2);}
    ;

do_conditional : TOK_WHILE expression_any {((DoClause*)clause)->SetWhile($2);}
    | TOK_UNTIL expression_any {((DoClause*)clause)->SetUntil($2);}
    ;

/*
do_instruction : TOK_DO
    | TOK_DO TOK_SYMBOL TOK_EQUAL expression_any {((DoClause*)clause)->SetSymbol(*($2.string), $2.secondary); ((DoClause*)clause)->SetFrom($4);}
    | TOK_DO TOK_SYMBOL TOK_EQUAL expression_any do_argument_list {((DoClause*)clause)->SetSymbol(*($2.string), $2.secondary); ((DoClause*)clause)->SetFrom($4);}
    | TOK_DO do_argument_list
    | TOK_DO expression_any {((DoClause*)clause)->SetFor($2);}
    | TOK_DO expression_any do_argument_list {((DoClause*)clause)->SetFor($2);}
    ;

do_argument_list : TOK_FOREVER {((DoClause*)clause)->SetForever();}
    | TOK_TO expression_any {((DoClause*)clause)->SetTo($2);}
    | TOK_BY expression_any {((DoClause*)clause)->SetBy($2);}
    | TOK_FOR expression_any {((DoClause*)clause)->SetFor($2);}
    | TOK_WHILE expression_any {((DoClause*)clause)->SetWhile($2);}
    | TOK_UNTIL expression_any {((DoClause*)clause)->SetUntil($2);}
    | do_argument_list TOK_FOREVER {((DoClause*)clause)->SetForever();}
    | do_argument_list TOK_FOR expression_any {((DoClause*)clause)->SetFor($3);}
    | do_argument_list TOK_TO expression_any {((DoClause*)clause)->SetTo($3);}
    | do_argument_list TOK_BY expression_any {((DoClause*)clause)->SetBy($3);}
    | do_argument_list TOK_WHILE expression_any {((DoClause*)clause)->SetWhile($3);}
    | do_argument_list TOK_UNTIL expression_any {((DoClause*)clause)->SetUntil($3);}
    ;
*/

/* drop instruction */
drop_instruction : TOK_DROP drop_variable_list
    ;

drop_variable_list : TOK_SYMBOL {
	((DropClause*)clause)->AddVariable(*($1.string));
	}
    | TOK_LPAREN TOK_SYMBOL TOK_RPAREN {
	((DropClause*)clause)->AddVariableRef(*($2.string));
	}
    | drop_variable_list bl TOK_SYMBOL {
	((DropClause*)clause)->AddVariable(*($3.string));
	}
    | drop_variable_list bl TOK_LPAREN TOK_SYMBOL TOK_RPAREN {
	((DropClause*)clause)->AddVariableRef(*($4.string));
	}
    ;

/* else instruction */
else_instruction : TOK_ELSE
    ;

separator_list : TOK_NEWLINE
    | TOK_BLANK
    | separator_list TOK_NEWLINE
    | separator_list TOK_BLANK
    ;

/* end instruction */
end_instruction : TOK_END
    | TOK_END TOK_SYMBOL {((EndClause*)clause)->SetVariable(*($2.string));}
    ;

/* exit instruction */
exit_instruction : TOK_EXIT
    | TOK_EXIT expression_any {((ExitClause*)clause)->SetExpression($2);}
    ;

/* if instruction */
if_instruction : TOK_IF expression_any {((IfClause*)clause)->SetExpression($2);}
    ;

/* interpret instruction */
interpret_instruction : TOK_INTERPRET expression_any {((InterpretClause*)clause)->SetExpression($2);}
    ;

/* iterate instruction */
iterate_instruction : TOK_ITERATE
    | TOK_ITERATE TOK_SYMBOL {((IterateClause*)clause)->SetVariable(*($2.string));}
    ;

/* leave instruction */
leave_instruction : TOK_LEAVE
    | TOK_LEAVE TOK_SYMBOL {((LeaveClause*)clause)->SetVariable(*($2.string));}
    ;

/* nop instruction */
nop_instruction : TOK_NOP
    ;

/* numeric instruction */
numeric_instruction : TOK_NUMERIC_DIGITS expression_any {((NumericClause*)clause)->SetDigits($2);}
    | TOK_NUMERIC_FUZZ expression_any {((NumericClause*)clause)->SetFuzz($2);}
    | TOK_NUMERIC_FORM_SCIENTIFIC {((NumericClause*)clause)->SetForm(new LiteralExpression(ArgString("SCIENTIFIC")));}
    | TOK_NUMERIC_FORM_ENGINEERING {((NumericClause*)clause)->SetForm(new LiteralExpression(ArgString("ENGINEERING")));}
    | TOK_NUMERIC_FORM_VALUE expression_any {((NumericClause*)clause)->SetForm($2);}
    ;

/* options instruction */
options_instruction : TOK_OPTIONS expression_any {((OptionsClause*)clause)->SetExpression($2);}
    ;

/* otherwise instruction */
otherwise_instruction : TOK_OTHERWISE
    ;

/* parse instruction */
parse_instruction : {clause = new ParseArgClause;} TOK_PARSE_ARG optional_template_list
    | {clause = new ParseLineinClause;} TOK_PARSE_LINEIN optional_template_list
    | {clause = new ParseExternalClause;} TOK_PARSE_EXTERNAL optional_template_list
    | {clause = new ParsePullClause;} TOK_PARSE_PULL optional_template_list
    | {clause = new ParseSourceClause;} TOK_PARSE_SOURCE optional_template_list
    | {clause = new ParseValueClause;} TOK_PARSE_VALUE expression_any TOK_WITH optional_template_list {((ParseValueClause*)clause)->SetExpression($3);}
    | {clause = new ParseVarClause;} TOK_PARSE_VAR TOK_SYMBOL optional_template_list {((ParseVarClause*)clause)->SetVariable(*($3.string), $3.secondary);}
    | {clause = new ParseVersionClause;} TOK_PARSE_VERSION optional_template_list
    | {clause = new ParseArgClause;} TOK_PARSE_UPPER_ARG optional_template_list {((ParseArgClause*)clause)->Upper();}
    | {clause = new ParseLineinClause;} TOK_PARSE_UPPER_LINEIN optional_template_list {((ParseLineinClause*)clause)->Upper();}
    | {clause = new ParseExternalClause;} TOK_PARSE_UPPER_EXTERNAL optional_template_list {((ParseExternalClause*)clause)->Upper();}
    | {clause = new ParsePullClause;} TOK_PARSE_UPPER_PULL optional_template_list {((ParsePullClause*)clause)->Upper();}
    | {clause = new ParseSourceClause;} TOK_PARSE_UPPER_SOURCE optional_template_list {((ParseSourceClause*)clause)->Upper();}
    | {clause = new ParseValueClause;} TOK_PARSE_UPPER_VALUE expression_any TOK_WITH optional_template_list {((ParseValueClause*)clause)->SetExpression($3); ((ParseValueClause*)clause)->Upper();}
    | {clause = new ParseVarClause;} TOK_PARSE_UPPER_VAR TOK_SYMBOL optional_template_list {((ParseVarClause*)clause)->SetVariable(*($3.string), $3.secondary); ((ParseVarClause*)clause)->Upper();}
    | {clause = new ParseVersionClause;} TOK_PARSE_UPPER_VERSION optional_template_list {((ParseVersionClause*)clause)->Upper();}
    ;

/* procedure instruction */
procedure_instruction : TOK_PROCEDURE
    | TOK_PROCEDURE_EXPOSE procedure_variable_list
    ;

procedure_variable_list : TOK_SYMBOL {
	((ProcedureClause*)clause)->AddVariable(*($1.string));
	}
    | TOK_LPAREN TOK_SYMBOL TOK_RPAREN {
	((ProcedureClause*)clause)->AddVariableRef(*($2.string));
	}
    | procedure_variable_list bl TOK_SYMBOL {
	((ProcedureClause*)clause)->AddVariable(*($3.string));
	}
    | procedure_variable_list bl TOK_LPAREN TOK_SYMBOL TOK_RPAREN {
	((ProcedureClause*)clause)->AddVariableRef(*($4.string));
	}
    ;



/* pull instruction */
/* optional_template_list fills in the necessary details for this instruction */
pull_instruction : TOK_PULL optional_template_list
    ;

/* push instruction */
push_instruction : TOK_PUSH
    | TOK_PUSH expression_any {((PushClause*)clause)->SetExpression($2);}
    ;

/* queue instruction */
queue_instruction : TOK_QUEUE
    | TOK_QUEUE expression_any {((QueueClause*)clause)->SetExpression($2);}
    ;

/* return instruction */
return_instruction : TOK_RETURN
    | TOK_RETURN expression_any {((ReturnClause*)clause)->SetExpression($2);}
    ;

/* say instruction */
say_instruction : TOK_SAY {((SayClause*)clause)->SetExpression(0);}
    | TOK_SAY expression_any {((SayClause*)clause)->SetExpression($2);}
    ;

/* select_instruction */
select_instruction : TOK_SELECT
    ;

/* signal instruction */
signal_instruction : TOK_SIGNAL TOK_SYMBOL {
	((SignalClause*)clause)->SetExpression(new LiteralExpression(*($2.string)));
	}
    | TOK_SIGNAL TOK_STRING_LITERAL {
	((SignalClause*)clause)->SetExpression(new LiteralExpression(*($2.string)));
	}
    | TOK_SIGNAL_VALUE expression_any {
	((SignalClause*)clause)->SetExpression($2);
	}
    | TOK_SIGNAL_ON_ERROR_NAME TOK_SYMBOL {
	((SignalClause*)clause)->TurnOn(SigError, *($2.string));
	}
    | TOK_SIGNAL_ON_FAILURE_NAME TOK_SYMBOL {
	((SignalClause*)clause)->TurnOn(SigFailure, *($2.string));
	}
    | TOK_SIGNAL_ON_HALT_NAME TOK_SYMBOL {
	((SignalClause*)clause)->TurnOn(SigHalt, *($2.string));
	}
    | TOK_SIGNAL_ON_NOTREADY_NAME TOK_SYMBOL  {
	((SignalClause*)clause)->TurnOn(SigNotReady, *($2.string));
	}
    | TOK_SIGNAL_ON_NOVALUE_NAME TOK_SYMBOL {
	((SignalClause*)clause)->TurnOn(SigNoValue, *($2.string));
	}
    | TOK_SIGNAL_ON_SYNTAX_NAME TOK_SYMBOL {
	((SignalClause*)clause)->TurnOn(SigSyntax, *($2.string));
	}

    | TOK_SIGNAL_ON_ERROR {
	((SignalClause*)clause)->TurnOn(SigError, ArgString());
	}
    | TOK_SIGNAL_ON_FAILURE {
	((SignalClause*)clause)->TurnOn(SigFailure, ArgString());
	}
    | TOK_SIGNAL_ON_HALT {
	((SignalClause*)clause)->TurnOn(SigHalt, ArgString());
	}
    | TOK_SIGNAL_ON_NOTREADY {
	((SignalClause*)clause)->TurnOn(SigNotReady, ArgString());
	}
    | TOK_SIGNAL_ON_NOVALUE {
	((SignalClause*)clause)->TurnOn(SigNoValue, ArgString());
	}
    | TOK_SIGNAL_ON_SYNTAX {
	((SignalClause*)clause)->TurnOn(SigSyntax, ArgString());
	}

    | TOK_SIGNAL_OFF_ERROR {
	((SignalClause*)clause)->TurnOff(SigHalt);
	}
    | TOK_SIGNAL_OFF_FAILURE {
	((SignalClause*)clause)->TurnOff(SigFailure);
	}
    | TOK_SIGNAL_OFF_HALT {
	((SignalClause*)clause)->TurnOff(SigHalt);
	}
    | TOK_SIGNAL_OFF_NOTREADY {
	((SignalClause*)clause)->TurnOff(SigNotReady);
	}
    | TOK_SIGNAL_OFF_NOVALUE {
	((SignalClause*)clause)->TurnOff(SigNoValue);
	}
    | TOK_SIGNAL_OFF_SYNTAX {
	((SignalClause*)clause)->TurnOff(SigSyntax);
	}
    ;

trace_instruction : TOK_TRACE_SS symbol_or_string {((TraceClause*)clause)->SetExpression(new LiteralExpression(*($2.string)));}
    | TOK_TRACE_VALUE expression_any {((TraceClause*)clause)->SetExpression($2);}
    ;

/* then instruction */
then_instruction : TOK_THEN
    ;

/* upper instruction
upper_instruction : TOK_UPPER variable_list
    ;
*/

/* WHEN INSTRUCTION */
when_instruction : TOK_WHEN expression_any {((WhenClause*)clause)->SetExpression($2);}
    ;


expression_any : error {$$ = new ErrorExpression(EInvalidExpression);}
/*
	| error TOK_RPAREN {$$ = new ErrorExpression(EUnexpectedCommaOrRightparen);}
	| TOK_LPAREN error {$$ = new ErrorExpression(EUnmatchedLeftparenInExpression);}
*/
           | TOK_SYMBOL {
                   switch($1.secondary) {
                     case TOK_SIMPLE :
		       $$ = new SimpleSymExpression(*($1.string)); break;
                     case TOK_COMPOUND : 
                       $$ = new CompoundSymExpression(*($1.string)); break;
                     case TOK_STEM :
		       $$ = new StemSymExpression(*($1.string)); break;
		   };
                 }
           | TOK_STRING_LITERAL {$$ = new LiteralExpression(*($1.string));}
	   | TOK_FIXED {$$ = new LiteralExpression(*($1.string));}
           | TOK_LPAREN expression_any TOK_RPAREN {$$ = $2;}
           | any_function_call {$$ = $1;}
           | TOK_NOT expression_any {$$ = new NotExpression($2);}
           | expression_any TOK_EXP expression_any {$$ = new ExponentExpression($1, $3);}
           | TOK_MINUS expression_any %prec NEGATIVE {$$ = new NegativeExpression($2);}
           | TOK_PLUS expression_any %prec POSITIVE {$$ = new PositiveExpression($2);}
           | expression_any TOK_MULT expression_any {$$ = new MultiplyExpression($1, $3);}
           | expression_any TOK_INT_DIV expression_any {$$ = new IntDivideExpression($1, $3);}
           | expression_any TOK_REMAINDER expression_any {$$ = new RemainderExpression($1, $3);}
           | expression_any TOK_REAL_DIV expression_any {$$ = new DivideExpression($1, $3);}
           | expression_any TOK_MINUS expression_any {$$ = new SubtractExpression($1, $3);}
           | expression_any TOK_PLUS expression_any {$$ = new AddExpression($1, $3);}
           | expression_any TOK_BLANK expression_any {$$ = new BlankExpression($1, $3);}
           | expression_any TOK_CONCAT expression_any {$$ = new ConcatExpression($1, $3);}
           | expression_any TOK_EQUAL expression_any {$$ = new EqualExpression($1, $3);}
           | expression_any TOK_NOT_EQUAL expression_any {$$ = new NotEqualExpression($1, $3);}
           | expression_any TOK_LESS_EQUAL expression_any {$$ = new LessEqualExpression($1, $3);}
           | expression_any TOK_GREATER_EQUAL expression_any {$$ = new GreaterEqualExpression($1, $3);}
           | expression_any TOK_LESS_THAN expression_any {$$ = new LessThanExpression($1, $3);}
           | expression_any TOK_GREATER_THAN expression_any {$$ = new GreaterThanExpression($1, $3);}
           | expression_any TOK_EXACT_EQUAL expression_any {$$ = new ExactEqualExpression($1, $3);}
           | expression_any TOK_EXACT_NOT_EQUAL expression_any {$$ = new ExactNotEqualExpression($1, $3);}
           | expression_any TOK_EXACT_LESS_EQUAL expression_any {$$ = new ExactLessEqualExpression($1, $3);}
           | expression_any TOK_EXACT_GREATER_EQUAL expression_any {$$ = new ExactGreaterEqualExpression($1, $3);}
           | expression_any TOK_EXACT_LESS_THAN expression_any {$$ = new ExactLessThanExpression($1, $3);}
           | expression_any TOK_EXACT_GREATER_THAN expression_any {$$ = new ExactGreaterThanExpression($1, $3);}
           | expression_any TOK_AND expression_any {$$ = new AndExpression($1, $3);}
           | expression_any TOK_OR expression_any {$$ = new OrExpression($1, $3);}
           | expression_any TOK_XOR expression_any {$$ = new XorExpression($1, $3);}
    ; 

any_function_call : TOK_SYMBOL TOK_LPAREN optional_expression_list
    TOK_RPAREN {
	$$ = new FunctionExpression(*($1.string), $3);
    }
    | TOK_STRING_LITERAL TOK_LPAREN optional_expression_list TOK_RPAREN {
	$$ = new FunctionExpression(*($1.string), $3, 1); // skip search
    }
    ;

/*
optional_expression_list :
    | expression_list
    ;
*/

optional_expression_list : optional_expression_list TOK_COMMA expression_any {
    $$->append($3);}
                | optional_expression_list TOK_COMMA {$$->append((Expression*)0);}
                | expression_any {$$ = new ExpressionpSLList; $$->append($1);}
                | /* null expression */ {$$ = new ExpressionpSLList; $$->append((Expression*)0);}
    ;

%%

int
yyerror(char* string)
{
/*
  char buf[200];

  parsererror = 1;
*/
  /*
	cout << string << " at line " << glob.LineNumber << "\n";
  */
  return 1;
}



