%{
/* - remove comments
   - remove newlines in strings
   - convert ';' into '\n'
*/
#include "MemoryFile.h"

#include "Glob.h"
extern Glob glob; // our globals are here

// stage 1 is input, glob.Stage2 is output
#define yylex pp1_lex
#undef YY_INPUT
#define YY_INPUT(buf, result, max_size) {int c = glob.Stage1.input(); result = (c==EOF) ? YY_NULL : (buf[0] = c, 1);}

#define DONE {BEGIN 0; yyrestart(0); yyterminate();}

%}

%x comment dquote squote

digit	[0-9]
char    [a-zA-Z]
dot     [.]
other   [!?$_]
any     [0-9a-zA-Z.!?$_]
nodot   [0-9a-zA-Z!?$_]
ws      [\t ]*
special [,)(~+\-*/%|=><&^:]

%%

%{
/* handle strings */
%}

^"#".*  {
    	/* skip this stuff -- non standard!  needed for unix */
	if (glob.LineNumber != 1) {
            glob.pp1_InvalidCharacter = 1;
            DONE;
        }
	}

[\t ]+                  {
                        glob.Stage2.output(' ');
		        }

\"               {BEGIN dquote;
                  glob.Stage2.output('"');
	         }

<dquote>\"\"     {glob.Stage2.output('"'); glob.Stage2.output('"');
	         }

<dquote>[^"\n]*  {int i;
                  for (i = 0; i < strlen(yytext); i++) glob.Stage2.output(yytext[i]);
                 }

<dquote>\n       {glob.pp1_StringNotTerminated = 1; /*printf("<dquote>\n"); */DONE;}

<dquote>\"       {BEGIN 0; glob.Stage2.output('"');}

\'               {BEGIN squote;
                  glob.Stage2.output('\'');
	         }

<squote>\'\'     {glob.Stage2.output('\''); glob.Stage2.output('\'');
	         }

<squote>[^'\n]*  {int i;
                  for (i = 0; i < strlen(yytext); i++) glob.Stage2.output(yytext[i]);
                 }

<squote>\n       {glob.pp1_StringNotTerminated = 1; /*printf("<squote>\n");*/ DONE;}

<squote>\'       {BEGIN 0; glob.Stage2.output('\'');}

%{
/* convert '\n' and ';' into newlines */
%}

{ws}"\n"{ws}            {glob.LineNumber++;
			glob.Stage2.output('\n');
			char buff[32];
			sprintf(buff, "#%d\n", glob.LineNumber);
			glob.Stage2.output(buff);
		        }

{ws}";"{ws}             {glob.Stage2.output('\n');
		        }


%{
/* comment processing */
%}

"/*"	{
	glob.NestingLevel = 1;
	BEGIN comment;
}
<comment>"*/"		{ if (--glob.NestingLevel == 0) BEGIN 0; }
<comment>"/*"		{ ++glob.NestingLevel; }
<comment>\n		;
<comment>[^*/\n]*	;
<comment>[*/]		;

<<EOF>> {
	if (glob.NestingLevel) {
		glob.pp1_CommentNotTerminated = 1;
                DONE;
	}
	yyrestart(0);
	yyterminate();
	}

{any}|{special}                   {glob.Stage2.output(yytext[0]);}

.                       {printf("Bad character %c\n", yytext[0]); glob.pp1_InvalidCharacter = 1;  DONE;}

%%


/*
main() {
    yylex();
}
*/






