%{

/* notes :
*/

/*
#define YYSTYPE char*
*/

#include "ArgStr.h"
#include <String.h>
#include "Express.h"
#include "RexxTokens.h"
#include "rexxparse.tok.h"
#include "MemoryFile.h"

#include "Glob.h"

#include "stringsupport.h"

extern Glob glob;

#define DONE {glob.Tokens->AddToken(TOK_NEWLINE, glob.LineNumber, ""); glob.Tokens->AddToken(0); BEGIN 0; yyrestart(0); yyterminate();}

#define yylex rexx_yylex
#undef YY_INPUT
#define YY_INPUT(buf, result, max_size) {int c = glob.Stage1.input(); result = (c==EOF) ? YY_NULL : (buf[0] = c, 1);}

    
    static char string[1024];
    static int reserved = 0;

/* scan for reserved tokens in mid-line area */

#define RES_IF_WHEN 1
#define RES_DO 2
#define RES_PARSE 3

%}

%x comment dquote squote

digit	[0-9]
char    [a-zA-Z]
dot     [.]
other   [!?$_]
any     [0-9a-zA-Z.!?$_]
nodot   [0-9a-zA-Z!?$_]
ws      [\t ]*
bl      [\t ]+

%%

^"#".*  {
    	/* skip this stuff -- non standard!  needed for unix */
	if (glob.LineNumber != 1) {
            glob.pp1_InvalidCharacter = 1;
            DONE;
        }
	}

%{
/* convert '\n' and ';' into newlines */
%}

"\n"            {
                         glob.Tokens->AddToken(TOK_NEWLINE, glob.LineNumber, "");
			glob.LineNumber++;
		        }

";"             {glob.Tokens->AddToken(TOK_NEWLINE);
		        }


%{
/* comment processing */
%}

"/*"	{
	glob.NestingLevel = 1;
	BEGIN comment;
}
<comment>"*/"		{ if (--glob.NestingLevel == 0) BEGIN 0; }
<comment>"/*"		{ ++glob.NestingLevel; }
<comment>\n		;
<comment>[^*/\n]*	;
<comment>[*/]		;

<<EOF>> {
	if (glob.NestingLevel) {
		glob.pp1_CommentNotTerminated = 1;
                DONE;
	}
	DONE;
	}

%{
/* operators and stuff */
%}

[\t ]+                  {
                        /*printf("(blank)\n");*/
                        glob.Tokens->AddToken(TOK_BLANK);
		        }

\"               {BEGIN dquote;
		  strcpy(string, "");
	         }

<dquote>\"\"     {
		  strcat(string, "\"");
	         }

<dquote>[^"\n]*  {
                  strcat(string, yytext);
                 }

<dquote>\n       {glob.pp1_StringNotTerminated = 1; /*printf("<dquote>\n"); */DONE;}

<dquote>\" {
	BEGIN 0;
        /*printf("(string - %s)\n", string);*/
        glob.Tokens->AddToken(TOK_STRING_LITERAL, 0, string);
	}

<dquote>\"[xX]/[^0-9a-zA-Z.!?$_] {
        BEGIN 0;
	/*printf("(x/b string - %s)\n", string);*/
        // we need to massage this into a hex string
	ArgString s(string);
	if (!validhex(s)) {
		glob.rexx_InvalidHexOrBinString = 1;
		DONE;
	}
        glob.Tokens->AddToken(TOK_STRING_LITERAL, 1, packhex(s));
	}

<dquote>\"[bB]/[^0-9a-zA-Z.!?$_] {
        BEGIN 0;
	/*printf("(x/b string - %s)\n", string);*/
        // we need to massage this into a binary or hex string
	ArgString s(string);
	if (!validbin(s)) {
		glob.rexx_InvalidHexOrBinString = 1;
		DONE;
	}

        glob.Tokens->AddToken(TOK_STRING_LITERAL, 2, packbin(s));
	}

\' {
	BEGIN squote;
	strcpy(string, "");
	}

<squote>\'\' {
	strcat(string, "\'");
	}

<squote>[^'\n]* {
        strcat(string, yytext);
        }

<squote>\n       {glob.pp1_StringNotTerminated = 1; /*printf("<squote>\n");*/ DONE;}

<squote>\' {
	BEGIN 0;
	/*printf("(string - %s)\n", string);*/
	glob.Tokens->AddToken(TOK_STRING_LITERAL, 0, string);
	}

<squote>\'[xX]/[^0-9a-zA-Z.!?$_] {
	BEGIN 0;
	/*printf("(x/b string - %s)\n", string);*/
	ArgString s(string);
	if (!validhex(s)) {
		glob.rexx_InvalidHexOrBinString = 1;
		DONE;
	}

        glob.Tokens->AddToken(TOK_STRING_LITERAL, 1, packhex(s));
	}

<squote>\'[bB]/[^0-9a-zA-Z.!?$_] {
	BEGIN 0;
	/*printf("(x/b string - %s)\n", string);*/
	ArgString s(string);
	if (!validbin(s)) {
		glob.rexx_InvalidHexOrBinString = 1;
		DONE;
	}

        glob.Tokens->AddToken(TOK_STRING_LITERAL, 2, packbin(s));
	}


","     {
	/*		printf("(comma)\n");*/
	glob.Tokens->AddToken(TOK_COMMA);
	}

")"     {/*printf("(rparen)\n");*/ glob.Tokens->AddToken(TOK_RPAREN);}

"("     {/*printf("(lparen)\n");*/ glob.Tokens->AddToken(TOK_LPAREN);}

"^"     {/*printf("(not)\n");*/ glob.Tokens->AddToken(TOK_NOT);}

"\\"    {/*printf("(not)\n");*/ glob.Tokens->AddToken(TOK_NOT);}

"+"     {/*printf("(plus)\n");*/ glob.Tokens->AddToken(TOK_PLUS);}

"-"     {/*printf("(minus)\n");*/ glob.Tokens->AddToken(TOK_MINUS);}

"*"            {/*printf("(multiplication)\n");*/ glob.Tokens->AddToken(TOK_MULT);}

"/"            {/*printf("(division)\n");*/ glob.Tokens->AddToken(TOK_REAL_DIV);}

"%"            {/*printf("(int div)\n");*/ glob.Tokens->AddToken(TOK_INT_DIV);}

"="            {/*printf("(equal)\n");*/ glob.Tokens->AddToken(TOK_EQUAL);}

">"            {/*printf("(great)\n");*/ glob.Tokens->AddToken(TOK_GREATER_THAN);}

"<"            {/*printf("(less)\n");*/ glob.Tokens->AddToken(TOK_LESS_THAN);}

"&"            {/*printf("(and)\n");*/ glob.Tokens->AddToken(TOK_AND);}

"|"            {/*printf("(or)\n");*/ glob.Tokens->AddToken(TOK_OR);}

":"	glob.Tokens->AddToken(TOK_COLON, yytext);

"address"       glob.Tokens->AddToken(TOK_ADDRESS, 0, "ADDRESS");

"value"		glob.Tokens->AddToken(TOK_VALUE, 0, "VALUE");

"arg"		glob.Tokens->AddToken(TOK_ARG, 0, "ARG");

"break"         glob.Tokens->AddToken(TOK_BREAK, 0, "BREAK");

"call"          glob.Tokens->AddToken(TOK_CALL, 0, "CALL");

"on"		glob.Tokens->AddToken(TOK_ON, 0, "ON");

"off"		glob.Tokens->AddToken(TOK_OFF, 0, "OFF");

"error"		glob.Tokens->AddToken(TOK_ERROR, 0, "ERROR");

"failure"	glob.Tokens->AddToken(TOK_FAILURE, 0, "FAILURE");

"halt"		glob.Tokens->AddToken(TOK_HALT, 0, "HALT");

"notready"	glob.Tokens->AddToken(TOK_NOTREADY, 0, "NOTREADY");

"name"		glob.Tokens->AddToken(TOK_NAME, 0, "NAME");

"do"  		glob.Tokens->AddToken(TOK_DO, 0, "DO");

"to"		glob.Tokens->AddToken(TOK_TO, 0, "TO");

"by"		glob.Tokens->AddToken(TOK_BY, 0, "BY");

"for"		glob.Tokens->AddToken(TOK_FOR, 0, "FOR");

"forever"	glob.Tokens->AddToken(TOK_FOREVER, 0, "FOREVER");

"while"		glob.Tokens->AddToken(TOK_WHILE, 0, "WHILE");

"until"		glob.Tokens->AddToken(TOK_UNTIL, 0, "UNTIL");


"drop"		glob.Tokens->AddToken(TOK_DROP, 0, "DROP");

"say"		glob.Tokens->AddToken(TOK_SAY, 0, "SAY");

"else"		glob.Tokens->AddToken(TOK_ELSE, 0, "ELSE");

"end"		glob.Tokens->AddToken(TOK_END, 0, "END");

"exit"		glob.Tokens->AddToken(TOK_EXIT, 0, "EXIT");

"if"		glob.Tokens->AddToken(TOK_IF, 0, "IF");

"then"		glob.Tokens->AddToken(TOK_THEN, 0, "THEN");

"interpret"	glob.Tokens->AddToken(TOK_INTERPRET, 0, "INTERPRET");

"iterate"	glob.Tokens->AddToken(TOK_ITERATE, 0, "ITERATE");

"leave"		glob.Tokens->AddToken(TOK_LEAVE, 0, "LEAVE");

"nop"		glob.Tokens->AddToken(TOK_NOP, 0, "NOP");

"numeric"	glob.Tokens->AddToken(TOK_NUMERIC, 0, "NUMERIC");

"digits"	glob.Tokens->AddToken(TOK_DIGITS, 0, "DIGITS");

"fuzz"		glob.Tokens->AddToken(TOK_FUZZ, 0, "FUZZ");

"form"		glob.Tokens->AddToken(TOK_FORM, 0, "FORM");

"scientific"	glob.Tokens->AddToken(TOK_SCIENTIFIC, 0, "SCIENTIFIC");

"engineering"	glob.Tokens->AddToken(TOK_ENGINEERING, 0, "ENGINEERING");

"value"		glob.Tokens->AddToken(TOK_VALUE, 0, "VALUE");

"options"	glob.Tokens->AddToken(TOK_OPTIONS, 0, "OPTIONS");

"otherwise"	glob.Tokens->AddToken(TOK_OTHERWISE, 0, "OTHERWISE");

"parse"		glob.Tokens->AddToken(TOK_PARSE, 0, "PARSE");

"arg"		glob.Tokens->AddToken(TOK_ARG, 0, "ARG");

"linein"	glob.Tokens->AddToken(TOK_LINEIN, 0, "LINEIN");

"external"	glob.Tokens->AddToken(TOK_EXTERNAL, 0, "EXTERNAL");

"numeric"	glob.Tokens->AddToken(TOK_NUMERIC, 0, "NUMERIC");

"pull"		glob.Tokens->AddToken(TOK_PULL, 0, "PULL");

"source"	glob.Tokens->AddToken(TOK_SOURCE, 0, "SOURCE");

"with"		glob.Tokens->AddToken(TOK_WITH, 0, "WITH");

"var"		glob.Tokens->AddToken(TOK_VAR, 0, "VAR");

"version"	glob.Tokens->AddToken(TOK_VERSION, 0, "VERSION");

"upper"		glob.Tokens->AddToken(TOK_UPPER, 0, "UPPER");

"procedure"	glob.Tokens->AddToken(TOK_PROCEDURE, 0, "PROCEDURE");

"expose"	glob.Tokens->AddToken(TOK_EXPOSE, 0, "EXPOSE");

"push"		glob.Tokens->AddToken(TOK_PUSH, 0, "PUSH");

"queue"		glob.Tokens->AddToken(TOK_QUEUE, 0, "QUEUE");

"return"	glob.Tokens->AddToken(TOK_RETURN, 0, "RETURN");

"select"	glob.Tokens->AddToken(TOK_SELECT, 0, "SELECT");

"signal"	glob.Tokens->AddToken(TOK_SIGNAL, 0, "SIGNAL");

"novalue"	glob.Tokens->AddToken(TOK_NOVALUE, 0, "NOVALUE");

"syntax"	glob.Tokens->AddToken(TOK_SYNTAX, 0, "SYNTAX");

"trace"		glob.Tokens->AddToken(TOK_TRACE, 0, "TRACE");

"when"		glob.Tokens->AddToken(TOK_WHEN, 0, "WHEN");

%{
/* fixed, simple, stem, compound processing */
%}

({digit}|{dot}){any}* {
        String s(yytext);
        s.upcase();
        strcpy(string,(char*)s);
        glob.Tokens->AddToken(TOK_FIXED, 1, string);
	}

({char}|{other}){nodot}* {
        String s(yytext);
        s.upcase();
        strcpy(string,(char*)s);
        glob.Tokens->AddToken(TOK_SYMBOL, TOK_SIMPLE, string);
	}

({char}|{other}){nodot}*{dot} {
        String s(yytext);
        s.upcase();
        strcpy(string,(char*)s);
        glob.Tokens->AddToken(TOK_SYMBOL, TOK_STEM, string);
	}

({char}|{other}){nodot}*{dot}{any}+ {
        String s(yytext);
        s.upcase();
        strcpy(string,(char*)s);
        glob.Tokens->AddToken(TOK_SYMBOL, TOK_COMPOUND, string);
	}

.       {
        glob.pp1_InvalidCharacter = 1;
        DONE;
        }

%%

 
