#include "stringsupport.h"
#include <String.h>
#include "ArgStr.h"
#include "math.h"
#include <assert.h>

int validnumber(const ArgString& a) {
    const char* s = (const char*)a;
    int length = a.Length();

    // we need to make sure that this is a null-terminated string!
    assert(s[length] == '\0');

    int i = 0;
    while (s[i] == ' ') i++; // skip spaces before the sign
    if (s[i] == '+' || s[i] == '-') i++;
    while (s[i] == ' ') i++; // skip spaces after the sign

    if (s[i] >= '0' && s[i] <= '9') {
	// parses (digits.digits), (digits.)
	while (s[i] >= '0' && s[i] <= '9') i++; // skip over digits
	if (s[i] == '.') i++;
	while (s[i] >= '0' && s[i] <= '9') i++; // skip over digits
    } else if (s[i] == '.') {
	i++;
	if (s[i] >= '0' && s[i] <= '9') {
	    // parses (.digits)
	    while (s[i] >= '0' && s[i] <= '9') i++; // skip over digits
	} else {
	    // parsed (.) only
	    return 0;
	} 
    } else {
	// no digit or '.'
	return 0;
    }

    if (s[i] == 'e' || s[i] == 'E') {
	// parses ([eE]{+-}digits)
	i++;
	if (s[i] == '+' || s[i] == '-') i++;
	if (s[i] >= '0' && s[i] <= '9') {
	    while (s[i] >= '0' && s[i] <= '9') i++; // skip over digits
	} else {
	    return 0;
	}
    }

    while (s[i] == ' ')  i++; // skip over trailing spaces

    if (i == length) return 1; // we've parsed the whole string
    return 0; // we haven't parsed the whole string
}

ArgString space(ArgString& a) {
    ArgString string = stripfront(striprear(a));
    ArgString result("");

    int readingspaces = 0;
    for (int i = 0; i < string.Length(); i++) {
	if (readingspaces && string[i] == ' ') {
	    // we're reading spaces and we got a space
	    // do nothing
	} else if (readingspaces && string[i] != ' ') {
	    // we're reading spaces and we got a non-space
	    readingspaces = 0;
	    result += " ";
	} else if (!readingspaces && string[i] != ' ') {
	    // we're reading words and we got a non-space
	} else if (!readingspaces && string[i] == ' ') {
	    // we're reading words and we got a space
	    readingspaces = 1;
	}

	if (!readingspaces) {
	    result += ArgString(string[i]);
	}
    }
    return result;
}

int split(ArgString& string, ArgString words[], int maxn) {
    int index = 0;
    int wordnumber = -1;
    int readingspaces = 1;
    ArgString word = "";
    for (int i = 0; i < string.Length(); i++) {
	if (readingspaces && string[i] == ' ') {
	    // we're reading spaces and we got a space
	    // do nothing
	} else if (readingspaces && string[i] != ' ') {
	    // we're reading spaces and we got a non-space
	    wordnumber++;
	    readingspaces = 0;
	} else if (!readingspaces && string[i] != ' ') {
	    // we're reading words and we got a non-space
	    // build the word
	    word += string[i];
	} else if (!readingspaces && string[i] == ' ') {
	    // we're reading words and we got a space
	    words[wordnumber] = word;
	    word = "";
	    readingspaces = 1;
	}
    }
    return wordnumber;
}

ArgString xrange(unsigned char start, unsigned char end) {
    ArgString result;
    for (unsigned char i = start; i != end; i++)
	result += (char)i;
    result += (char)i;
    return result;
}

ArgString upper(ArgString& a) {
    ArgString result(a);
    for (int i = 0; i < result.Length(); i++) {
	if (result[i] >= 'a' && result[i] <= 'z')
	    result[i] = 'A' + result[i] - 'a';
    }
    return result;
}

ArgString lower(ArgString& a) {
    ArgString result(a);
    for (int i = 0; i < result.Length(); i++) {
	if (result[i] >= 'A' && result[i] <= 'Z')
	    result[i] = 'a' + result[i] - 'A';
    }
    return result;
}

ArgString stripfront(ArgString& a, char pad) {
    int length = a.Length();
    for (int i = 0; i < length; i++)
	if (a[i] != pad) break;
    if (i < length) {
	return substring(a, i, length);
    } else {
	return "";
    }
}

ArgString striprear(ArgString& a, char pad) {
    int length = a.Length();
    for (int i = length - 1; i >= 0; i--)
	if (a[i] != pad) break;
    if (i >= 0) {
	return substring(a, 0, i + 1);
    } else {
	return "";
    }
}

ArgString ud2x(ArgString& a) {
    ArgString result("");

    long l = a.Integer();
    char c;
    while (1) {
	if (l < 16) {
	    result += ArgString(chartohex((char)l));
	    break;
	}
	c = chartohex((char) (l % 16));
	result += ArgString(c);
	l = l / 16;
    }
    return reverse(result);
}

ArgString ud2c(ArgString& a) {
    ArgString result("");
    long l = a.Integer();
    char c;
    while (1) {
	if (l < 256) {
	    result += ArgString((char)l);
	    break;
	}
	c = (char) (l % 256);
	result += ArgString(c);
	l = l / 256;
    }
    return reverse(result);
}

ArgString reverse(ArgString& a) {
    ArgString result(a);
    int length = a.Length();
    for (int i = 0; i < length; i++) {
	result[i] = a[length - i - 1];
    }
    return result;
}

// returns substring from start to end (excluding end)
ArgString substring(ArgString& a, int s, int e) {
    if (e - s < 0) return "";
    if (s >= a.Length()) return "";
    if (e > a.Length()) e = a.Length();
    ArgString result("", e - s);
    for (int i = 0; i < (e - s); i++) {
	result[i] = a[i + s];
    }
    return result;
}

int validhex(ArgString& a) {
    const char* cp = a.String();
    // no leading or trailing spaces
    if (cp[0] == ' ' || cp[0] == '\t' || cp[strlen(cp)-1] == ' ' || cp[strlen(cp)-1] == '\t') return 0;
    
    // check the groupings
    // first group is allowed to be odd
    int count = 0;
    int groupnumber = 1;
    
    while (*cp) {
	// skip any spaces
	while (*cp && (*cp == ' ' || *cp == '\t')) cp++;
	while (*cp && *cp != ' ' && *cp != '\t') {
	    if ((*cp >= 'a' && *cp <= 'f') || (*cp >= 'A' && *cp <= 'F') ||
		(*cp >= '0' && *cp <= '9')) {
		count++;
	    } else {
		return 0; // invalid character in hex string
	    }
	    cp++;
	}
	if (groupnumber != 1 && (count % 2) != 0) return 0;
	groupnumber++;
	count = 0;
    }
    return 1;
}

int validbin(ArgString& a) {
    const char* cp = a.String();
    // no leading or trailing spaces
    if (cp[0] == ' ' || cp[0] == '\t' || cp[strlen(cp)-1] == ' ' || cp[strlen(cp)-1] == '\t') return 0;
    
    // check the groupings
    // first group is allowed to be odd
    int count = 0;
    int groupnumber = 1;
    
    while (*cp) {
	// skip any spaces
	while (*cp && (*cp == ' ' || *cp == '\t')) cp++;
	while (*cp && *cp != ' ' && *cp != '\t') {
	    if (*cp >= '0' && *cp <= '1') {
		count++;
	    } else {
		return 0; // invalid character in bin string
	    }
	    cp++;
	}
	if (groupnumber != 1 && (count % 4) != 0) return 0;
	groupnumber++;
	count = 0;
    }
    return 1;
}



// takes h & l and returns character
char hextochar(char h, char l) {
    unsigned  char result = 0x0;
    if (h >= '0' && h <= '9') result = result | (h - '0');
    else if (h >= 'A' && h <= 'F') result = result | (h - 'A' + 10);
    result = result << 4;
    if (l >= '0' && l <= '9') result = result | (l - '0');
    else if (l >= 'A' && l <= 'F') result = result | (l - 'A' + 10);
    return result;
}

char hextochar(char l) {
    unsigned  char result = 0x0;
    if (l >= '0' && l <= '9') result = result | (l - '0');
    else if (l >= 'A' && l <= 'F') result = result | (l - 'A' + 10);
    return result;
}

char bintohex(char b3, char b2, char b1, char b0) {
    char result = bintochar('0', '0', '0', '0', b3, b2, b1, b0);
    result = chartohex(result);
    return result;
}

char chartohex(char c) {
    // only lower 4 bits are significant
    c = c & 0x0f;
    if (c >= 0 && c <= 9) return ('0' + c);
    else return ('A' + (c - 10));
}

// takes b7...b0 and returns character
// brute force!
char bintochar(char b7, char b6, char b5, char b4, char b3, char b2, char b1, char b0) {
    unsigned  char result = 0x0;
    result = result | (b7 - '0');
    result = result << 1;

    result = result | (b6 - '0');
    result = result << 1;

    result = result | (b5 - '0');
    result = result << 1;

    result = result | (b4 - '0');
    result = result << 1;

    result = result | (b3 - '0');
    result = result << 1;

    result = result | (b2 - '0');
    result = result << 1;

    result = result | (b1 - '0');
    result = result << 1;

    result = result | (b0 - '0');

    return result;
}

ArgString packhex(ArgString& a) {
    if (!validhex(a)) return ArgString(); // invalid hex
    String temp = a.String();
    temp.upcase(); // all upper case
    temp.gsub(RXwhite, ""); // no more spaces
    if ((temp.length() % 2) != 0) temp = "0" + temp; // add an extra '0'
    ArgString result("", temp.length()/2); // make us an argstring of the correct length
    int index = 0;
    for (int i = 0; i < temp.length()/2; i++) {
	result[i] = hextochar(temp[index], temp[index+1]);
	index += 2;
    }
    return result;
}

ArgString packbin(ArgString& a) {
    if (!validbin(a)) return ArgString(); // invalid bin
    String temp = a.String();
    temp.upcase(); // all upper case
    temp.gsub(RXwhite, ""); // no more spaces
    if ((temp.length() % 4) != 0) {
	temp = replicate("0", 4 - (temp.length() % 4)) + temp; // add extra 0's
    }
    if ((temp.length() % 8) != 0) {
	temp = "0000" + temp;
    }

    ArgString result("", temp.length()/8); // make us an argstring of the correct length
    int index = 0;
    for (int i = 0; i < temp.length()/8; i++) {
	result[i] = bintochar(temp[index], temp[index+1], temp[index+2], temp[index+3],
		      temp[index+4], temp[index+5], temp[index+6], temp[index+7]);
	index += 8;
    }
    return result;
}

// pad on right
ArgString padonright(ArgString& orig, int newlength, char pad) {
    ArgString result(orig.String(), newlength);
    for (int i = orig.Length(); i < newlength; i++)
	result[i] = pad;
    return result;
}

// pad on left
ArgString padonleft(ArgString& orig, int newlength, char pad) {
    ArgString result("", newlength);
    int numpads;
    int copystart;
    if (newlength < orig.Length()) {
	// we need to strip some stuff
	numpads = 0;
	copystart = orig.Length() - newlength;
    } else {
	numpads = newlength - orig.Length();
	copystart = 0;
    }

    for (int i = 0; i < numpads; i++) {
	result[i] = pad;
    }
    int copylength = newlength - numpads;
    for (i = 0; i < copylength; i++) {
	result[numpads + i] = orig[i + copystart];
    }
    return result;
}

ArgString twocomplement(ArgString& string) {
    int length = string.Length();
    ArgString result(string);
    for (int i = 0; i < length; i++) {
	result[i] = (unsigned char)string[i] ^ 0xff;
    }
    for (i = length - 1; i >= 0; i++) {
	if ((unsigned char)result[i] == 0xff) {
	    // carry over
	    result[i]++;
	} else {
	    result[i]++;
	    break;
	}
    }
    return result;
}

ArgString xtwocomplement(ArgString& string) {
    int length = string.Length();
    ArgString result(string);
    for (int i = 0; i < length; i++) {
	result[i] = chartohex(15 - hextochar(string[i]));
    }
    for (i = length - 1; i >= 0; i++) {
	if (result[i] == 'F') {
	    // carry over
	    result[i] = '0';
	} else {
	    result[i] = chartohex(hextochar(result[i] + 1));
	    break;
	}
    }
    return result;
}

ArgString c2d(ArgString& string, int n) {
    long result = 0;
    ArgString positive;
    // check sign
    ArgString temp = padonleft(string, n, '\0');
    if ((unsigned char)temp[0] & 0x80) {
	// negative
	positive = twocomplement(temp);
	positive = uc2d(positive);
	result = 0 - positive.Integer();
    } else {
	// positive
	positive = temp;
	positive = uc2d(positive);
	result = positive.Integer();
    }
    
    return ArgString((int)result);
}

ArgString uc2d(ArgString& string) {
    int length = string.Length();
    long result = 0;
    
    ArgString revstring = reverse(string);

    long multiplier = 1;
    for (int i = 0; i < length; i++) {
	result = result + (unsigned char)(revstring[i]) * multiplier;
	multiplier = multiplier * 256;
    }
    return ArgString((int)result);
}

int isalphanumeric(ArgString& string) {
    int length = string.Length();
    if (!length) return 0;
    for (int i = 0; i < length; i++) {
	if (!((string[i] >= 'a' && string[i] <= 'z') ||
	    (string[i] >= 'A' && string[i] <= 'Z') ||
		(string[i] >= '0' && string[i] <= '9'))) return 0;
    }
    return 1;
}

int islower(ArgString& string) {
    int length = string.Length();
    if (!length) return 0;
    for (int i = 0; i < length; i++) {
	if (!(string[i] >= 'a' && string[i] <= 'z')) return 0;
    }
    return 1;
}

int ismixed(ArgString& string) {
    int length = string.Length();
    if (!length) return 0;
    for (int i = 0; i < length; i++) {
	if (!((string[i] >= 'a' && string[i] <= 'z') ||
	    (string[i] >= 'A' && string[i] <= 'Z')
	      )) return 0;
    }
    return 1;
}

int issscsymbol(ArgString& string) {
    if (!issymbol(string)) return 0;
    // fixed symbol
    if (string[0] == '.' || (string[0] >= '0' && string[0] <= '9'))
	return 0;
    return 1;
}

int issimplesymbol(ArgString& string) {
    if (!issscsymbol(string)) return 0;
    // detect simple symbol
    int length = string.Length();
    for (int i = 0; i < length; i++) {
	if (string[i] == '.') return 0; // no periods allowed
    }
    return 1;
}

int isstemsymbol(ArgString& string) {
    if (!issscsymbol(string)) return 0;
    // detect stem symbol
    int length = string.Length() - 1;
    for (int i = 0; i < length; i++) {
	if (string[i] == '.') return 0;
    }
    if (string[i] == '.') return 1; // last one needs to be a period
    return 0;
}

int iscompoundsymbol(ArgString& string) {
    if (!issscsymbol(string)) return 0;
    // detect compound symbol
    int length = string.Length() - 1;
    int numperiods = 0;
    for (int i = 0; i < length; i++) {
	if (string[i] == '.') numperiods++;
    }
    if (numperiods >= 1) return 1;
    return 0;
}



int issymbol(ArgString& string) {
    int length = string.Length();
    if (!length) return 0;
    // check if it's a number
    if (string.IsANumber()) return 1;

    for (int i = 0; i < length; i++) {
	// next character may be in [a-zA-Z!?$_.]
	if (!((string[i] >= 'a' && string[i] <= 'z') ||
	    (string[i] >= 'A' && string[i] <= 'Z') ||
		(string[i] >= '0' && string[i] <= '9') ||
	      (string[i] == '!') || (string[i] == '?') || 
	      (string[i] == '?') || (string[i] == '$') ||
	      (string[i] == '_') || (string[i] == '.'))) return 0;
    }
    return 1;
}

int isupper(ArgString& string) {
    int length = string.Length();
    if (!length) return 0;
    for (int i = 0; i < length; i++) {
	if (!(string[i] >= 'A' && string[i] <= 'Z')) return 0;
    }
    return 1;
}

