#include "support.h"

#include "InternalEnv.h"
#include "InternalData.h"
#include "RexxClauses.h"
#include "Express.h"
#include "Expressionp.SLList.h"
//#include <stream.h>
#include "Glob.h"
#include "RexxStream.h"

extern Glob glob;

void SetElapsedTime(long sec, long micro) {
    glob.Environments.GetData()->ElapsedSeconds = sec;
    glob.Environments.GetData()->ElapsedMicroSeconds = micro;
}

void GetElapsedTime(long& sec, long& micro) {
    sec = glob.Environments.GetData()->ElapsedSeconds;
    micro = glob.Environments.GetData()->ElapsedMicroSeconds;
}

int GetStart() {
    return glob.Environments.GetData()->StartPC;
}

int GetEnd() {
    return glob.Environments.GetData()->EndPC;
}

void SetStart(int s) {
    glob.Environments.GetData()->StartPC = s;
}

void SetEnd(int e) {
    glob.Environments.GetData()->EndPC = e;
}

int GetSourcecodeLines() {
    return glob.Environments.GetData()->Source->lines();
}

char* GetSourcecode(int l) {
    return glob.Environments.GetData()->Source->getline(l);
}

int SourcecodeLoaded() {
    return glob.Environments.GetData()->SourceLoaded;
}

void SetSourcecode(MemoryFile& m) {
    glob.Environments.GetData()->Source = new MemoryFile;
    (*glob.Environments.GetData()->Source) = m; // copy it
    glob.Environments.GetData()->SourceOwner = 1;
    glob.Environments.GetData()->SourceLoaded = 1;
}

// some global stuff 

void SetDigits(int d) {
    glob.Environments.GetData()->Digits = d;
}

int GetDigits(void) {
    return glob.Environments.GetData()->Digits;
}

void SetForm(int f) {
    glob.Environments.GetData()->Form = f;
}

int GetForm(void) {
    return glob.Environments.GetData()->Form;
}

void SetFuzz(int f) {
    glob.Environments.GetData()->Fuzz = f;
}

int GetFuzz(void) {
    return glob.Environments.GetData()->Fuzz;
}

ArgString GetDefaultAddress() {
    return glob.DefaultAddress;
}

void PrepareLexers() {
    glob.Stage2.clear();
    glob.Stage3.clear();
    glob.Tokens->clear();
    glob.Tokens2->clear();
    glob.LineNumber = 1;
    glob.NestingLevel = 0;
    glob.pp1_CommentNotTerminated = 0;
    glob.pp1_StringNotTerminated = 0;
    glob.pp1_InvalidCharacter = 0;
    glob.rexx_InvalidHexOrBinString = 0;
}

ArgString GetLinein() {
/*    char buf[128];
    cin.getline(buf, 128);
    buf[strlen(buf) - 1] = '\0'; // kill terminator
*/
    ArgString ret;
    glob.Streams["stdin"]->Linein(ret);
    return ret;
}

ArgString GetPull() {
    if (glob.GlobalQueue.Enqueued()) return glob.GlobalQueue.Deq();
    return GetLinein();
}

ArgString GetExternal() {
    if (glob.GlobalQueue.Enqueued()) return glob.GlobalQueue.Deq();
    // return null string if there is nothing there
    return ArgString("");
}

ArgString GetVersion() {
    return glob.Version;
}

void SetSource(ArgString s) {
    glob.Environments.GetData()->source = s;
}

ArgString GetSource() {
    return glob.Environments.GetData()->source;
}

// support functions for doing internalenv stuff

void SetCurrentAddress(ArgString a) {
    glob.Environments.GetData()->currentaddress = a;
}

void SetPreviousAddress(ArgString a) {
    glob.Environments.GetData()->previousaddress = a;
}

ArgString& GetCurrentAddress() {
    return glob.Environments.GetData()->currentaddress;
}

ArgString& GetPreviousAddress() {
    return glob.Environments.GetData()->previousaddress;
}

// this is used in external functions and commands
// we need to do this so that we don't carry over signals from an external command or function
void ClearPendingSignals() {
    if (!glob.SignalIsPending) return;

    for (Signals i = SigSignal; i <= SigNotReady; i++) {
	LowerSignal(i);
    }
}

// has side effect of raisingsignal that is pending and setting errornumber
int PendingSignal() {
    // cycle through all the signals and see if any has been raised
    for (Signals i = SigSignal; i <= SigNotReady; i++) {
	if (SignalPending(i)) {
	    if (!SignalDelayed(i)) {
		RaiseSignal(i);
		SetErrorNumberForSignal(i);
		return 1;
	    }
	}
    }
    return 0;
}

// processes one pending signal only (priority is given to lower signal numbers)
int ProcessPendingSignals() {
    while (1) {
	// pendingsignal does some sort of priority scheme
	if (glob.SignalIsPending && PendingSignal()) {
	    if (ActionIsSignal()) {
		// we're only supposed to process CALLs
//		cout << "in processpendingsignals -- sigaction caught\n";
		return 0; // let someone else process this
	    } else if (ActionIsCall()) {
		// process this signal as a CALL
		// delay the signal
		SetSignalDelayed(glob.RaisedSignal);
		// lower the raisedsignal, coz we're going to handle the call
		LowerSignal();
		// save the delayed signal
		Signals delayedsignal = glob.RaisedSignal;
		FunctionExpression* exp = new FunctionExpression(GetSignalLabel(glob.RaisedSignal), new ExpressionpSLList);
		exp->SetCall(); // we are a CALL function so don't raise signals if you don't get a return value!
		SetSignalBeingProcessed(glob.RaisedSignal);
		ArgString a = exp->Evaluate();
		delete exp;
		SetSignalBeingProcessed(SigSignal);
		// reactivate the delayed signal if it is still delayed
		if (SignalDelayed(delayedsignal)) SetSignalReTurnOn(delayedsignal);
		if (!a.IsValid()) {
		    if (IsReturn()) {
			// this error is OK
		    } else if (IsSignal() && ActionIsCall()) {
//			cout << "issignal() && actioniscall()\n";
			// this is ok too, let's stay in the loop
		    } else {
			// anything else is fatal and let's get out of here
			return 0;
		    }
		}
	    }
	} else {
	    return 1;
	}
    }
}

ArgString& GetSignalLabel(Signals s) {
    return glob.Environments.GetData()->SignalLabel[s];
}

void HasSymbolTable(int i) {
    glob.Environments.GetData()->HasSymbolTable = i;
}

int HasSymbolTable() {
    return glob.Environments.GetData()->HasSymbolTable;
}

// return reference to arguments[int]
ArgString& Args(int i ) {
    if (i < 0 || i > (MaxArguments - 1)) {
	SetError(EInterpretationError);
    }
    return glob.Environments.GetData()->Arguments[i];
}

int& ArgsOmitted(int i ) {
    if (i < 0 || i > (MaxArguments - 1)) {
	SetError(EInterpretationError);
    }
    return glob.Environments.GetData()->ArgumentsOmitted[i];
}

/*
void SetNumArguments(int i) {
    glob.Environments.GetData()->NumArguments = i;
}
*/

int GetNumArguments() {
    // this is the Cowlishaw def. of numargs
    int max = -1;
    for (int i = 0; i < 10; i++)
	if (!glob.Environments.GetData()->ArgumentsOmitted[i]) max = i;
	
    return max + 1;
}

int SignalPending(Signals s) {
    return glob.SignalPending[s];
}

int SignalOn(Signals s) {
    return glob.Environments.GetData()->SignalState[s] == On;
}

int SignalOff(Signals s) {
    return glob.Environments.GetData()->SignalState[s] == Off;
}

int SignalDelayed(Signals s) {
    return glob.Environments.GetData()->SignalState[s] == Delayed;
}

ArgString& GetSignalLabel() {
    return glob.Environments.GetData()->SignalLabel[glob.RaisedSignal];
}

int SignalDelayed() {
    return glob.Environments.GetData()->SignalState[glob.RaisedSignal] == Delayed;
}

int SignalOn() {
    return glob.Environments.GetData()->SignalState[glob.RaisedSignal] == On;
}

int SignalOff() {
    return glob.Environments.GetData()->SignalState[glob.RaisedSignal] == Off;
}

int LowerSignal() {
    LowerSignal(glob.RaisedSignal);
}

int ActionIsCall() {
    return glob.Environments.GetData()->SignalAction[glob.RaisedSignal] == ACall;
}

int ActionIsSignal() {
    return glob.Environments.GetData()->SignalAction[glob.RaisedSignal] == ASignal;
}

Signals GetSignalBeingProcessed() {
    return glob.Environments.GetData()->SignalBeingProcessed;
}

void SetSignalBeingProcessed(Signals s) {
    glob.Environments.GetData()->SignalBeingProcessed = s;
}

ArgString GetSignalDescription(Signals s) {
    return glob.Environments.GetData()->SignalDescription[s];
}

void SetSignalDescription(Signals s, ArgString a) {
    glob.Environments.GetData()->SignalDescription[s] = a;
}

void SetSignalOff(Signals s) {
    glob.Environments.GetData()->SignalState[s] = Off;
    glob.Environments.GetData()->SignalAction[s] = ASignal; // default action is to signal
}

void SetSignalDelayed(Signals s) {
    glob.Environments.GetData()->SignalState[s] = Delayed;
}

// this should only be used to re-turn-on something
//   i.e. use SetSignalOn(Signals, ...) in clauses
void SetSignalReTurnOn(Signals s) {
    glob.Environments.GetData()->SignalState[s] = On;
}

void SetSignalPending(Signals s) {
    glob.SignalPending[s] = 1;
    glob.SignalIsPending = 1;
}

void SetSignalOn(Signals s , ArgString a, Actions actions) {
    glob.Environments.GetData()->SignalAction[s] = actions;
    glob.Environments.GetData()->SignalState[s] = On;
    if (!a.IsValid()) {
	switch (s) {
	  case SigSignal : 
	    abort(); // should never get here
	    break;
	  case SigNoValue : a = ArgString("NOVALUE");
	    break;
	  case SigSyntax : a = ArgString("SYNTAX");
	    break;
	  case SigHalt : a = ArgString("HALT");
	    break;
	  case SigError : a = ArgString("ERROR");
	    break;
	  case SigFailure : a = ArgString("FAILURE");
	    break;
	  case SigNotReady : a = ArgString("NOTREADY");
	    break;
	};
    }
    glob.Environments.GetData()->SignalLabel[s] = a;
}

void LowerSignal(Signals s) {
    glob.SignalPending[s] = 0;
    for (Signals i = SigSignal; i <= SigNotReady; i++) {
	if (SignalPending(i)) {
	    glob.SignalIsPending = 1;
	    return;
	}
    }
    glob.SignalIsPending = 0;
}

void RaiseSignal(Signals s) {
    glob.ReasonForReturn = ReasonSignal;
    glob.RaisedSignal = s;
    SetSignalPending(s);
    
}

void ClearError() {
    // this just sets things to their defaults
    SetError(ERealError);
}

void SetVariableError(char* text) {
    glob.VariableError = text;
    RaiseSignal(SigSyntax);
    glob.Error = EVariableError;
    SetReturnValue(ArgString());
}

void SetError(Errors e) {
    // we decide whether it should be ignored or not
    switch (e) {
      case ERealError :
	glob.ReasonForReturn = ReasonError; // non-maskable error
	break;
      case ESyntax :
	RaiseSignal(SigSyntax);
	break;
      case ESignal :
	RaiseSignal(SigSignal);
	break;
      case ENoValue : 
	 RaiseSignal(SigNoValue);
	break;
      case EError :
	 RaiseSignal(SigError);
	break;
      case EFailure :
	 RaiseSignal(SigFailure);
	break;
      case ENotReady :
	 RaiseSignal(SigNotReady);
	break;
      case EHalt :
	 RaiseSignal(SigHalt);
	break;
	
      case EUnknownAddress :
      case EInvalidNumericForm :
	RaiseSignal(SigSyntax);
	break;
      case EInvalidEnd:
	RaiseSignal(SigSyntax);
	break;
      case EProgramInterrupted :
	RaiseSignal(SigHalt);
	break;
      case EMachineResourcesExhausted :
	 RaiseSignal(SigSyntax);
	break;
      case EUnmatchedCommentOrQuote :
	 RaiseSignal(SigSyntax);
	break;
      case EWhenOrOtherwiseExpected :
	 RaiseSignal(SigSyntax);
	break;
      case EUnexpectedThenOrElse :
	 RaiseSignal(SigSyntax);
	break;
      case EUnexpectedWhenOrOtherwise :
	 RaiseSignal(SigSyntax);
	break;
      case EUnexpectedOrUnmatchedEnd :
	 RaiseSignal(SigSyntax);
	break;
      case EControlStackFull :
	glob.ReasonForReturn = ReasonError; //non-maskable error
	break;
      case EClauseTooLong :
	glob.ReasonForReturn = ReasonError; //non-maskable error
	break;
      case EInvalidCharacterInProgram :
	RaiseSignal(SigSyntax);
//	glob.ReasonForReturn = ReasonError; //non-maskable error
	break;
      case EIncompleteDoSelectIf :
	 RaiseSignal(SigSyntax);
	break;
      case EInvalidHexadecimalOrBinaryString :
	 RaiseSignal(SigSyntax);
	break;
      case ELabelNotFound :
	 RaiseSignal(SigSyntax);
	break;
      case EUnexpectedProcedure :
	 RaiseSignal(SigSyntax);
	break;
      case EThenExpected :
	 RaiseSignal(SigSyntax);
	break;
      case EStringOrSymbolExpected :
	 RaiseSignal(SigSyntax);
	break;
      case ESymbolExpected :
	 RaiseSignal(SigSyntax);
	break;
      case EInvalidDataOnEndOfClause :
	glob.ReasonForReturn = ReasonError; //non-maskable error
	break;
      case EInvalidCharacterString : 
	RaiseSignal(SigSyntax);
//	glob.ReasonForReturn = ReasonError; //non-maskable error
	break;
      case EInvalidDataString :
	RaiseSignal(SigSyntax);
//	glob.ReasonForReturn = ReasonError; //non-maskable error
	break;
      case EInvalidTraceRequest :
	RaiseSignal(SigSyntax);
//	glob.ReasonForReturn = ReasonError; //non-maskable error
	break;
      case EInvalidSubkeywordFound :
	 RaiseSignal(SigSyntax);
	break;
      case EInvalidWholeNumber :
	 RaiseSignal(SigSyntax);
	break;
      case EInvalidDoSyntax :
	 RaiseSignal(SigSyntax);
	break;
      case EInvalidLeaveOrIterate :
	 RaiseSignal(SigSyntax);
	break;
      case EEnvironmentNameTooLong :
	 RaiseSignal(SigSyntax);
	break;
      case ENameOrStringTooLong :
	RaiseSignal(SigSyntax);
//	glob.ReasonForReturn = ReasonError; // non-maskable error
	break;
      case ENameStartsWithNumberOrPeriod :
	 RaiseSignal(SigSyntax);
	break;
      case EInvalidExpressionResult :
	 RaiseSignal(SigSyntax);
	break;
      case ELogicalValueNotZeroOrOne :
	 RaiseSignal(SigSyntax);
	break;
      case EInvalidExpression : 
	 RaiseSignal(SigSyntax);
	break;
      case EUnmatchedLeftparenInExpression :
	 RaiseSignal(SigSyntax);
	break;
      case EUnexpectedCommaOrRightparen :
	 RaiseSignal(SigSyntax);
	break;
      case EInvalidTemplateOrPattern :
	 RaiseSignal(SigSyntax);
	break;
      case EEvaluationStackOverflow :
	glob.ReasonForReturn = ReasonError; // non-maskable error
	break;
      case EIncorrectCallToRoutine :
	 RaiseSignal(SigSyntax);
	break;
      case EBadArithmeticConversion :
	 RaiseSignal(SigSyntax);
	break;
      case EArithmeticOverflowUnderflow :
	glob.ReasonForReturn = ReasonError; // non-maskable error
	break;
      case ERoutineNotFound : 
	 RaiseSignal(SigSyntax);
	break;
      case EFunctionDidNotReturnData :
	 RaiseSignal(SigSyntax);
	break;
      case ENoDataSpecifiedOnFunctionReturn :
	 RaiseSignal(SigSyntax);
	break;
      case EInvalidVariableReference :
	 RaiseSignal(SigSyntax);
	break;
      case EFailureInSystemService :
	glob.ReasonForReturn = ReasonError; // non-maskable error
	break;
      case EInterpretationError :
	glob.ReasonForReturn = ReasonError; // non-maskable error
	break;
      default :
	fprintf(stderr, "SetError : unimplemented %s\n",ErrorText[e]);
    };
    glob.Error = e;
    SetReturnValue(ArgString());
}

void SetBreak() {
    glob.ReasonForReturn = ReasonBreak;
}

void SetIterate(ArgString a) {
    glob.ReasonForReturn = ReasonIterate;
    glob.ReturnValue = a;
}

void SetEof() {
    glob.ReasonForReturn = ReasonEof;
}

void SetLeave(ArgString a) {
    glob.ReasonForReturn = ReasonLeave;
    glob.ReturnValue = a;
}

void SetExit(ArgString v) {
//    glob.Environments.GetData()->ReturnValue = v;
//    glob.Environments.GetData()->ReasonForReturn = ReasonExit;
    glob.ReturnValue = v;
    glob.ReasonForReturn = ReasonExit;
}

void SetReturn(ArgString v) {
//    glob.Environments.GetData()->ReturnValue = v;
//    glob.Environments.GetData()->ReasonForReturn = ReasonReturn;
    glob.ReturnValue = v;
    glob.ReasonForReturn = ReasonReturn;
}

void SetReturnValue(ArgString v) {
    glob.ReturnValue = v;
}

ArgString ReturnValue() {
//    return glob.Environments.GetData()->ReturnValue;
    return glob.ReturnValue;
}

// 
void SetErrorNumberForSignal(Signals s) {
    switch (s) {
      case SigSignal :
	glob.Error = ESignal;
	break;
      case SigNoValue :
	glob.Error = ENoValue;
	break;
      case SigSyntax :
	glob.Error = ESyntax;
	break;
      case SigHalt :
	glob.Error = EProgramInterrupted;
	break;
      case SigError :
	glob.Error = EError;
	break;
      case SigFailure :
	glob.Error = EFailure;
	break;
      case SigNotReady :
	glob.Error = ENotReady;
	break;
    };
}

const char * GetErrorText(Errors e) {
    static const char nullstring[] = "";
    if ((int)e < ERealError || (int)e >= ELastError) {
	return nullstring;
    } else if (e == EVariableError) {
	return glob.VariableError;
    } else {
	return ErrorText[e];
    }
}

int ActionIsSignal(Signals s) {
    return glob.Environments.GetData()->SignalAction[s] == ASignal;
}

int ActionIsCall(Signals s) {
    return glob.Environments.GetData()->SignalAction[s] == ACall;
}

// any of these functions should ONLY be called if the return value was BAD!

int IsExit() {
//    return glob.Environments.GetData()->ReasonForReturn == ReasonExit;
    return glob.ReasonForReturn == ReasonExit;
}

int IsBreak() {
    return glob.ReasonForReturn == ReasonBreak;
}

int IsIterate() {
    return glob.ReasonForReturn == ReasonIterate;
}

int IsLeave() {
    return glob.ReasonForReturn == ReasonLeave;
}

int IsReturn() {
//    return glob.Environments.GetData()->ReasonForReturn == ReasonReturn;
    return glob.ReasonForReturn == ReasonReturn;
}

int IsSignal() {
//     return glob.Environments.GetData()->ReasonForReturn == ReasonSignal;
    return glob.ReasonForReturn == ReasonSignal;
}

int IsError() {
    return glob.ReasonForReturn == ReasonError;
}

int IsEof() {
    return glob.ReasonForReturn == ReasonEof;
}

int IsSyntaxError() {
    return glob.ReasonForReturn == ReasonSignal && glob.RaisedSignal == SigSyntax;
}

int IsHaltError() {
    return glob.ReasonForReturn == ReasonSignal && glob.RaisedSignal == SigHalt;
}
// end of support functions

